/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo;

import java.sql.SQLException;
import java.util.List;
import java.util.Map;

import kr.kangwoo.damo.engine.IExtendedExecuteService;
import kr.kangwoo.damo.engine.cache.CacheKey;
import kr.kangwoo.damo.engine.provider.QueryProvider;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.damo.engine.query.result.ListRowHandler;
import kr.kangwoo.damo.engine.query.result.MapRowHandler;
import kr.kangwoo.util.MessageUtils;

/**
 * 확장된 쿼리 실행 서비스를 구현하고 있다.
 * 
 * @author badnom
 * @author agoodguy
 *
 */
public abstract class ExtendedExecutor implements IExtendedExecuteService {

	/**
	 * <p>쿼리 제공자(QueryProvider)를 반환한다.</p>
	 * 
	 * @return 쿼리 제공자(QueryProvider)
	 */
	public abstract QueryProvider getQueryProvider();
	
	public abstract Query getQuery(String queryId) throws SQLException;
	
	public int insert(Object entityObject) throws SQLException {
		Query query = getQueryProvider().getInsertQuery(entityObject.getClass());
		return executeUpdate(query, entityObject);
	}
	
	public int insert(String queryId, Object entityObject) throws SQLException {
		Query query = getQuery(queryId);
		return executeUpdate(query, entityObject);
	}
	
	public int update(Object entityObject) throws SQLException {
		Query query = getQueryProvider().getUpdateQuery(entityObject.getClass());
		return executeUpdate(query, entityObject);
	}
	
	public int update(String queryId, Object entityObject) throws SQLException {
		Query query = getQuery(queryId);
		return executeUpdate(query, entityObject);
	}
	
	public int updateSelective(Object entityObject) throws SQLException {
		Query query = getQueryProvider().getUpdateSelectiveQuery(entityObject.getClass());
		return executeUpdate(query, entityObject);
	}
	
	public int delete(Object entityObject) throws SQLException {
		Query query = getQueryProvider().getDeleteQuery(entityObject.getClass());
		return executeUpdate(query, entityObject);
	}
	
	public int deleteByPrimaryKey(Object entityObject) throws SQLException {
		Query query = getQueryProvider().getDeleteByPrimaryKeyQuery(entityObject.getClass());
		return executeUpdate(query, entityObject);
	}
	
	public int delete(String queryId, Object entityObject) throws SQLException {
		Query query = getQuery(queryId);
		return executeUpdate(query, entityObject);
	}
	
	@SuppressWarnings("unchecked")
	public <E> E getObject(E entityObject) throws SQLException {
		Query query = getQueryProvider().getGetObjectQuery(entityObject.getClass());
		return (E)getObject(query, entityObject, entityObject.getClass());
	}
	
	@SuppressWarnings("unchecked")
	public <E> E getObject(CacheKey cacheKey) throws SQLException {
		E entityObject = (E)cacheKey.getParameterObject();
		Query query = getQueryProvider().getGetObjectQuery(entityObject.getClass());
		return (E)getObject(query, cacheKey, entityObject.getClass());
	}
	
	@SuppressWarnings("unchecked")
	public <E> E getObjectByPrimaryKey(E entityObject) throws SQLException {
		Query query = getQueryProvider().getGetObjectByPrimaryKeyQuery(entityObject.getClass());
		return (E)getObject(query, entityObject, entityObject.getClass());
	}
	
	@SuppressWarnings("unchecked")
	public <E> E getObjectByPrimaryKey(CacheKey cacheKey) throws SQLException {
		E entityObject = (E)cacheKey.getParameterObject();
		Query query = getQueryProvider().getGetObjectByPrimaryKeyQuery(entityObject.getClass());
		return (E)getObject(query, cacheKey, entityObject.getClass());
	}
	
	public <E> E getObject(String queryId, Object parameterObject, Class<E> resultClass) throws SQLException {
		Query query = getQuery(queryId);
		return getObject(query, parameterObject, resultClass);
	}
	
	public <E> E getObject(Query query, Object parameterObject, Class<E> resultClass) throws SQLException {
		ListRowHandler<E> rowHandler = new ListRowHandler<E>();
		executeQuery(query, parameterObject, resultClass, rowHandler);
		if (rowHandler.getList().size() == 1) {
			return rowHandler.getList().get(0);
		}
		if (rowHandler.getList().size() > 1) {
			throw new SQLException(MessageUtils.format(Messages.getString("ExtendedExecutor.IS_NOT_ONE"), rowHandler.getList())); //$NON-NLS-1$
		}
		return null;
	}
	
	public <E, N extends Number> N getCount(E entityObject, Class<N> resultClass) throws SQLException {
		Query query = getQueryProvider().getGetCountQuery(entityObject.getClass(), resultClass);
		return getCount(query, entityObject, resultClass);
	}
	
	@SuppressWarnings("unchecked")
	public <E, N extends Number> N getCount(CacheKey cacheKey, Class<N> resultClass) throws SQLException {
		E entityObject = (E)cacheKey.getParameterObject();
		Query query = getQueryProvider().getGetCountQuery(entityObject.getClass(), resultClass);
		return getCount(query, entityObject, resultClass);
	}
	
	public <N extends Number> N getCount(Query query, Object parameterObject, Class<N> resultClass) throws SQLException {
		return getObject(query, parameterObject, resultClass);
	}
	
	public <N extends Number> N getCount(String queryId, Object parameterObject, Class<N> resultClass) throws SQLException {
		Query query = getQuery(queryId);
		return getObject(query, parameterObject, resultClass);
	}
	
	public <E> List<E> getList(E entityObject) throws SQLException {
		return getList(entityObject, null);
	}
	
	public <E> List<E> getList(CacheKey cacheKey) throws SQLException {
		return getList(cacheKey, null);
	}
	
	@SuppressWarnings("unchecked")
	public <E> List<E> getList(E entityObject, String orderBy) throws SQLException {
		Query query = getQueryProvider().getGetListQuery(entityObject.getClass(), orderBy);
		return (List<E>)getList(query, entityObject, entityObject.getClass());
	}
	
	@SuppressWarnings("unchecked")
	public <E> List<E> getList(CacheKey cacheKey, String orderBy) throws SQLException {
		E entityObject = (E)cacheKey.getParameterObject();
		Query query = getQueryProvider().getGetListQuery(entityObject.getClass(), orderBy);
		return (List<E>)getList(query, cacheKey, entityObject.getClass());
	}
	
	public <E> List<E> getList(E entityObject, int pageNo, int pageSize) throws SQLException {
		return getList(entityObject, null, pageNo, pageSize);
	}
	
	public <E> List<E> getList(CacheKey cacheKey, int pageNo, int pageSize) throws SQLException {
		return getList(cacheKey, null, pageNo, pageSize);
	}
	
	@SuppressWarnings("unchecked")
	public <E> List<E> getList(E entityObject, String orderBy, int pageNo, int pageSize) throws SQLException {
		Query query = getQueryProvider().getGetListQuery(entityObject.getClass(), orderBy);
		return (List<E>)getList(query, entityObject, entityObject.getClass(), pageNo, pageSize);
	}
	
	@SuppressWarnings("unchecked")
	public <E> List<E> getList(CacheKey cacheKey, String orderBy, int pageNo, int pageSize) throws SQLException {
		E entityObject = (E)cacheKey.getParameterObject();
		Query query = getQueryProvider().getGetListQuery(entityObject.getClass(), orderBy);
		return (List<E>)getList(query, cacheKey, entityObject.getClass(), pageNo, pageSize);
	}
	
	public <E> List<E> getList(Query query, Object parameterObject, Class<E> resultClass) throws SQLException {
		ListRowHandler<E> rowHandler = new ListRowHandler<E>();
		executeQuery(query, parameterObject, resultClass, rowHandler);
		return rowHandler.getList();
	}
	
	public <E> List<E> getList(String queryId, Object parameterObject, Class<E> resultClass) throws SQLException {
		Query query = getQuery(queryId);
		return (List<E>)getList(query, parameterObject, resultClass);
	}
	
	public <E> List<E> getList(Query query, Object parameterObject, Class<E> resultClass, int pageNo, int pageSize) throws SQLException {
		ListRowHandler<E> rowHandler = new ListRowHandler<E>();
		executeQuery(query, parameterObject, resultClass, rowHandler, pageNo, pageSize);
		return rowHandler.getList();
	}
	
	public <E> List<E> getList(String queryId, Object parameterObject, Class<E> resultClass, int pageNo, int pageSize) throws SQLException {
		Query query = getQuery(queryId);
		return (List<E>)getList(query, parameterObject, resultClass, pageNo, pageSize);
	}
	
	public <K, V> Map<K, V> getMap(V entityObject, String keyProperty) throws SQLException {
		return getMap(entityObject, null, keyProperty);
	}
	
	public <K, V> Map<K, V> getMap(CacheKey cacheKey, String keyProperty) throws SQLException {
		return getMap(cacheKey, null, keyProperty);
	}
	
	@SuppressWarnings("unchecked")
	public <K, V> Map<K, V> getMap(V entityObject, String orderBy, String keyProperty) throws SQLException {
		Query query = getQueryProvider().getGetListQuery(entityObject.getClass(), orderBy);
		return (Map<K, V>)getMap(query, entityObject, entityObject.getClass(), keyProperty);
	}
	
	@SuppressWarnings("unchecked")
	public <K, V> Map<K, V> getMap(CacheKey cacheKey, String orderBy, String keyProperty) throws SQLException {
		V entityObject = (V)cacheKey.getParameterObject();
		Query query = getQueryProvider().getGetListQuery(entityObject.getClass(), orderBy);
		return (Map<K, V>)getMap(query, entityObject, entityObject.getClass(), keyProperty);
	}

	public <K, V> Map<K, V> getMap(Query query, Object parameterObject, Class<V> resultClass, String keyProperty) throws SQLException {
		MapRowHandler<K, V> rowHandler = new MapRowHandler<K, V>(keyProperty);
		executeQuery(query, parameterObject, resultClass, rowHandler);
		return rowHandler.getMap();
	}
	
	@SuppressWarnings("unchecked")
	public <K, V> Map<K, V> getMap(String queryId, Object parameterObject, Class<V> resultClass, String keyProperty) throws SQLException {
		Query query = getQuery(queryId);
		return (Map<K, V>)getMap(query, parameterObject, resultClass, keyProperty);
	}

}
