/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo;

import java.sql.Connection;
import java.sql.SQLException;

import javax.sql.DataSource;

import kr.kangwoo.damo.config.DamoConfigLoader;
import kr.kangwoo.damo.engine.IBatchService;
import kr.kangwoo.damo.engine.IPersistenceService;
import kr.kangwoo.damo.engine.ISessionService;
import kr.kangwoo.damo.engine.ITransactionService;
import kr.kangwoo.damo.engine.cache.CacheProvider;
import kr.kangwoo.damo.engine.executor.QueryExecutor;
import kr.kangwoo.damo.engine.provider.QueryProvider;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.damo.engine.query.RunnableQuery;
import kr.kangwoo.damo.engine.transaction.Transaction;
import kr.kangwoo.damo.engine.transaction.TransactionManager;

/**
 * 영속성 관리자(PersistenceManager)이다. 
 * 세션 서비스, 트랙잭션 서비스, 실행 서비스, 배치 서비스들을 제공한다.
 * 
 * <pre>
 * PersistenceManager pm = PersistenceManager.getPersistenceManager();
 * pm.beginTransaction();
 * try {
 * 	Entity entity.setId = new Entity();
 * 	entity.setId(1004);
 * 	entity.setName("바보");
 * 	int updateRows = pm.insert(entity);
 * 	pm.commitTransaction();
 * } finally {
 * 	pm.endTransaction();
 * }
 * 
 * 
 * </pre>
 * 
 * @author badnom
 * @author agoodguy
 *
 */
public abstract class PersistenceManager extends ExtendedExecutor implements IBatchService, ISessionService,
		ITransactionService, IPersistenceService {

	private static class PersistenceManagerHolder {
		public static PersistenceManager persistenceManager = new PersistenceManagerImpl(DamoConfigLoader.getDamoConfig());
	}
	
	protected SessionProvider sessionProvider;
	protected DataSource dataSource;
	protected CacheProvider cacheProvider;
	protected QueryExecutor queryExecutor;
	protected QueryProvider queryProvider;
	protected TransactionManager transactionManager;

	protected abstract Session getLocalSession();
	
	public static PersistenceManager getPersistenceManager() {
		return PersistenceManagerHolder.persistenceManager;
	}
	
	public void shutdown() {
	}

	public Session openSession() {
		return sessionProvider.createSession(this);
	}

	public Session openSession(Connection conn) throws SQLException {
		Session session = sessionProvider.createSession(this);
		session.setUserConnection(conn);
		return session;
	}

	public void beginTransaction() throws SQLException {
		getLocalSession().beginTransaction();
	}

	public void beginTransaction(int transactionIsolationLevel)
			throws SQLException {
		getLocalSession().beginTransaction(transactionIsolationLevel);
	}

	public void commitTransaction() throws SQLException {
		getLocalSession().commitTransaction();
	}

	public void rollbackTransaction() throws SQLException {
		getLocalSession().rollbackTransaction();
	}

	public void endTransaction() throws SQLException {
		try {
			getLocalSession().endTransaction();
		} finally {
			getLocalSession().close();
		}
	}

	public Connection getCurrentConnection() throws SQLException {
		return getLocalSession().getCurrentConnection();
	}

	public void setUserConnection(Connection conn) throws SQLException {
		getLocalSession().setUserConnection(conn);
	}

	public SessionProvider getSessionProvider() {
		return sessionProvider;
	}
	
	public DataSource getDataSource() {
		return dataSource;
	}

	public void setDataSource(DataSource dataSource) {
		this.dataSource = dataSource;
	}

	public CacheProvider getCacheProvider() {
		return cacheProvider;
	}
	
	public QueryExecutor getQueryExecutor() {
		return queryExecutor;
	}

	public void setQueryExecutor(QueryExecutor queryExecutor) {
		this.queryExecutor = queryExecutor;
	}

	public QueryProvider getQueryProvider() {
		return queryProvider;
	}
	
	public void setQueryProvider(QueryProvider queryProvider) {
		this.queryProvider = queryProvider;
	}

	public TransactionManager getTransactionManager() {
		return transactionManager;
	}

	public void setTransactionManager(TransactionManager transactionManager) {
		this.transactionManager = transactionManager;
	}

	@Override
	public Query getQuery(String queryId) throws SQLException {
		return queryProvider.getQuery(queryId);
	}
	
	/**
	 * <p> 일괄 작업(Batch)을 추가한다.</p>
	 * <p>주의 : 트랜잭션을 직접 시작했을 경우만 사용할수 있다.</p>
	 * 
	 * @param query
	 * @param parameterObject
	 * @throws SQLException
	 */
	public void addBatch(RunnableQuery query, Object parameterObject)
			throws SQLException {
		Session session = getLocalSession();
		Transaction transaction = session.getTransaction();
		if (transaction == null) {
			throw new SQLException(Messages.getString("PersistenceManager.BATCH_TRAN_NOT_BEGIN")); //$NON-NLS-1$
		}
		session.addBatch(query, parameterObject);
	}

	/**
	 * <p>일괄 작업(Batch)을 실행한다.</p>
	 * <p>주의 : 트랜잭션을 직접 시작했을 경우만 사용할수 있다.</p>
	 * 
	 * @param query
	 * @return
	 * @throws SQLException
	 */
	public int[] executeBatch(RunnableQuery query) throws SQLException {
		Session session = getLocalSession();
		Transaction transaction = session.getTransaction();
		if (transaction == null) {
			throw new SQLException(Messages.getString("PersistenceManager.BATCH_TRAN_NOT_BEGIN")); //$NON-NLS-1$
		}
		return session.executeBatch(query);
	}
	

}
