/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo;

import java.sql.SQLException;
import java.util.Properties;

import javax.sql.DataSource;

import kr.kangwoo.damo.config.DamoConfig;
import kr.kangwoo.damo.config.DamoConfigLoader;
import kr.kangwoo.damo.config.IDamoConfig;
import kr.kangwoo.damo.engine.batch.BatchCallback;
import kr.kangwoo.damo.engine.batch.BatchResult;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.damo.engine.query.result.RowHandler;
import kr.kangwoo.damo.engine.transaction.BaseTransactionProvider;
import kr.kangwoo.damo.engine.transaction.Transaction;
import kr.kangwoo.damo.engine.transaction.TransactionException;
import kr.kangwoo.damo.engine.transaction.TransactionManager;
import kr.kangwoo.damo.engine.transaction.TransactionProvider;
import kr.kangwoo.damo.engine.transaction.jdbc.JdbcTransactionProvider;
import kr.kangwoo.util.logging.Logger;
import kr.kangwoo.util.logging.LoggerFactory;

/**
 * 영속성 관리자(PersistenceManager)의 최종 구현체이다.
 * 
 * @author badnom
 * @author agoodguy
 *
 */
public class PersistenceManagerImpl extends PersistenceManager {
	
	private Logger logger = LoggerFactory.getLogger(PersistenceManagerImpl.class);
	
	private final ThreadLocal<Session> localSession = new ThreadLocal<Session>();
	private TransactionProvider transactionProvider;
	
	public PersistenceManagerImpl() {
		this(new DamoConfig());
	}
	
	public PersistenceManagerImpl(IDamoConfig damoConfig) {
		if (damoConfig == null) {
			logger.warn(Messages.getString("PersistenceManagerImpl.DAMO_CONFIG_IS_NULL")); //$NON-NLS-1$
			damoConfig = DamoConfigLoader.getDamoConfig();
		}
		init(damoConfig);
	}
	
	public PersistenceManagerImpl(Properties properties) {
		this(new DamoConfig(properties));
	}
	
	protected void init(IDamoConfig damoConfig) {
		sessionProvider = new SessionProvider();
		sessionProvider.setMaxActive(damoConfig.getSessionMaxActive());
		sessionProvider.setTimeout(damoConfig.getSessionMaxWait());
		
		cacheProvider = damoConfig.getCacheProvider();
		
		queryProvider = damoConfig.getQueryProvider();
		
		queryExecutor = damoConfig.getQueryExecutor();
		
		transactionProvider = damoConfig.getTransactionProvider();
		DataSource dataSource = damoConfig.getDataSource();
		transactionManager = new TransactionManager(transactionProvider);
		transactionManager.setMaxActive(damoConfig.getTransactionMaxActive());
		transactionManager.setTimeout(damoConfig.getTransactionMaxWait());
		if (dataSource != null) {
			setDataSource(dataSource);
		}
	}
	
	@Override
	protected Session getLocalSession() {
		Session session = localSession.get();
		if (session == null || session.isClosed()) {
			session = sessionProvider.createSession(this);
			localSession.set(session);
		}
		return session;
	}
	
	@Override
	public void setDataSource(DataSource dataSource) {
		super.setDataSource(dataSource);
		if (transactionProvider == null) {
			logger.warn(Messages.getString("PersistenceManagerImpl.TRANSACTION_PROVIDER_IS_NULL")); //$NON-NLS-1$
			transactionProvider = new JdbcTransactionProvider();	
			transactionManager.setTransactionProvider(transactionProvider);
		}
		transactionProvider.setDataSource(dataSource);	
	}
	
	public void setTransactionProvider(BaseTransactionProvider transactionProvider) {
		this.transactionProvider = transactionProvider;
		if (transactionProvider != null && transactionProvider.getDataSource() == null && dataSource != null) {
			transactionProvider.setDataSource(dataSource);	
		}
		this.transactionManager.setTransactionProvider(transactionProvider);
	}
	
	public int executeUpdate(Query query, Object entityObject) throws SQLException {
		Session session = getLocalSession();
		int updateCount = 0;
		Transaction transaction = session.getTransaction();
		boolean autoBegin = (transaction == null);
		try {
			transaction = autoBeginTransaction(session, autoBegin, transaction);
			updateCount = session.executeUpdate(query, entityObject);
			autoCommitTransaction(session, autoBegin);
		} finally {
			autoEndTransaction(session, autoBegin);
		}
		return updateCount;
	}
	
	public <E> void executeQuery(Query query, Object parameterObject, Class<E> resultClass, RowHandler<E> rowHandler) throws SQLException {
		Session session = getLocalSession();
		Transaction transaction = session.getTransaction();
		boolean autoBegin = (transaction == null);
		try {
			transaction = autoBeginTransaction(session, autoBegin, transaction);
			session.executeQuery(query, parameterObject, resultClass, rowHandler);
			autoCommitTransaction(session, autoBegin);
		} finally {
			autoEndTransaction(session, autoBegin);
		}
	}
	
	public <E> E executeCallable(Query query, Object parameterObject,
			Class<E> resultClass) throws SQLException {
		E e = null;
		Session session = getLocalSession();
		Transaction transaction = session.getTransaction();
		boolean autoBegin = (transaction == null);
		try {
			transaction = autoBeginTransaction(session, autoBegin, transaction);
			e = session.executeCallable(query, parameterObject, resultClass);
			autoCommitTransaction(session, autoBegin);
		} finally {
			autoEndTransaction(session, autoBegin);
		}
		return e;
	}
	
	public <E> void executeQuery(Query query, Object parameterObject, Class<E> resultClass, RowHandler<E> rowHandler, int pageNo, int pageSize) throws SQLException {
		Session session = getLocalSession();
		Transaction transaction = session.getTransaction();
		boolean autoBegin = (transaction == null);
		try {
			transaction = autoBeginTransaction(session, autoBegin, transaction);
			session.executeQuery(query, parameterObject, resultClass, rowHandler, pageNo, pageSize);
//			autoCommitTransaction(session, autoBegin);
		} finally {
			autoEndTransaction(session, autoBegin);
		}
	}
	
	/**
	 * <p>배치를 실행한다.</p>
	 * 
	 * @param batchCallback
	 * @throws SQLException
	 */
	public BatchResult executeBatch(BatchCallback batchCallback) throws SQLException {
		BatchResult result = null;
		Session session = getLocalSession();
		Transaction transaction = session.getTransaction();
		boolean autoBegin = (transaction == null);
		try {
			transaction = autoBeginTransaction(session, autoBegin, transaction);
			result = session.executeBatch(batchCallback);
			autoCommitTransaction(session, autoBegin);
		} finally {
			autoEndTransaction(session, autoBegin);
		}
		return result;
		
	}
	/**
	 * <p>자동 시작일 경우 트랜잭션을 새로 만들어서 반환하고, 아니면 기존 트랜잭션을 반환한다.</p>
	 * 
	 * @param session
	 * @param autoBegin 자동 시작 여부
	 * @param transaction
	 * @return
	 * @throws SQLException
	 * @throws TransactionException
	 */
	private Transaction autoBeginTransaction(Session session,
			boolean autoBegin, Transaction transaction) throws SQLException, TransactionException {
		if (autoBegin) {
			session.beginTransaction();
			transaction = session.getTransaction();
		}
		return transaction;
	}
	
	/**
	 * <p>자동 시작일 경우 커밋을 한다.</p>
	 * 
	 * @param session
	 * @param autoBegin 자동 시작 여부
	 * @throws SQLException
	 * @throws TransactionException
	 */
	private void autoCommitTransaction(Session session, boolean autoBegin) throws SQLException, TransactionException {
		if (autoBegin) {
			session.commitTransaction();
		}
	}
	
	/**
	 * <p>자동 시작일 경우 트랜잭션을 종료 한다.</p>
	 * 
	 * @param session
	 * @param autoBegin
	 * @throws SQLException
	 * @throws TransactionException
	 */
	private void autoEndTransaction(Session session, boolean autoBegin) throws SQLException, TransactionException {
		if (autoBegin) {
			try {
				session.endTransaction();
			} finally {
				session.close();
			}
		}
	}

}
