/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import kr.kangwoo.damo.engine.IBatchService;
import kr.kangwoo.damo.engine.IExecuteService;
import kr.kangwoo.damo.engine.IPersistenceService;
import kr.kangwoo.damo.engine.ITransactionService;
import kr.kangwoo.damo.engine.batch.BatchCallback;
import kr.kangwoo.damo.engine.batch.BatchContext;
import kr.kangwoo.damo.engine.batch.BatchResult;
import kr.kangwoo.damo.engine.cache.CacheKey;
import kr.kangwoo.damo.engine.cache.CacheProvider;
import kr.kangwoo.damo.engine.cache.CacheStrategy;
import kr.kangwoo.damo.engine.executor.QueryExecutor;
import kr.kangwoo.damo.engine.provider.QueryProvider;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.damo.engine.query.RunnableQuery;
import kr.kangwoo.damo.engine.query.result.RowHandler;
import kr.kangwoo.damo.engine.transaction.Transaction;
import kr.kangwoo.damo.engine.transaction.TransactionContext;
import kr.kangwoo.damo.engine.transaction.TransactionException;
import kr.kangwoo.damo.engine.transaction.TransactionManager;
import kr.kangwoo.damo.engine.transaction.TransactionStatus;

public class Session extends ExtendedExecutor implements TransactionContext, BatchContext, ITransactionService, IExecuteService, IBatchService {

	private boolean closed;
	
	private SessionProvider sessionProvider;
	private TransactionManager transactionManager;
	private CacheProvider cacheProvider;
	private QueryProvider queryProvider;
	private QueryExecutor queryExecutor;
	
	private TransactionStatus transactionStatus;
	private Transaction transaction;
	private boolean commitRequired;
	private Map<String, PreparedStatement> preparedStatements;

	
	protected Session(IPersistenceService persistenceSupport) {
		this.closed = false;
		this.commitRequired = false;
		this.sessionProvider = persistenceSupport.getSessionProvider();
		this.transactionStatus = TransactionStatus.NO_TRANSACTION;
		this.transactionManager = persistenceSupport.getTransactionManager();
		this.cacheProvider = persistenceSupport.getCacheProvider();
		this.queryProvider = persistenceSupport.getQueryProvider();
		this.queryExecutor = persistenceSupport.getQueryExecutor();
		this.preparedStatements = new HashMap<String, PreparedStatement>();
	}
	
	public void beginTransaction() throws SQLException, TransactionException {
		transactionManager.begin(this);
	}
	
	public void beginTransaction(int transactionIsolationLevel) throws SQLException, TransactionException {
		transactionManager.begin(this, transactionIsolationLevel);
	}
	
	public void commitTransaction() throws SQLException, TransactionException {
		transactionManager.commit(this);
	}
	
	public void rollbackTransaction() throws SQLException, TransactionException {
		transactionManager.rollback(this);
	}
	
	public void endTransaction() throws SQLException, TransactionException {
		transactionManager.end(this);
	}
	
	public Connection getCurrentConnection() throws SQLException {
		return (transaction != null) ? transaction.getConnection() : null;
	}
	
	public void setUserConnection(Connection conn) throws SQLException {
		transactionManager.setUserProvidedConnection(this, conn);
	}
	
	public TransactionStatus getTransactionStatus() {
		return transactionStatus;
	}
	
	public void setTransactionStatus(TransactionStatus transactionStatus) {
		this.transactionStatus = transactionStatus;
	}

	public Transaction getTransaction() {
		return transaction;
	}

	public void setTransaction(Transaction transaction) {
		// Transaction이 변경되면 기존에 사용하던 Statement들을 모두 닫는다.
		if (this.transaction != null) {
			closeStatements();	
		}
		this.transaction = transaction;
	}

	public boolean isCommitRequired() {
		return commitRequired;
	}

	public void setCommitRequired(boolean commitRequired) {
		this.commitRequired = commitRequired;
	}
	
	
	public void closeStatements() {
		for (PreparedStatement pstmt : preparedStatements.values()) {
			try {
				pstmt.close();
			} catch(Exception e) {
				// 무시
			}
		}
		preparedStatements.clear();
	}

	public void close() {
		closeStatements();
		if (!closed) {
			closed = true;
			sessionProvider.closeSession(this);
		}
	}
	
	public boolean isClosed() {
		return closed;
	}
	
	public void addBatch(RunnableQuery query, Object parameterObject) throws SQLException {
		commitRequired = true;
		queryExecutor.addBatch(this, getCurrentConnection(), query, parameterObject);
	}
	
	public int[] executeBatch(RunnableQuery query) throws SQLException {
		commitRequired = true;
		return queryExecutor.executeBatch(this, getCurrentConnection(), query);
	}
	
	public BatchResult executeBatch(BatchCallback batchCallback) throws SQLException {
		commitRequired = true;
		return batchCallback.doIteration(this);
	}

	public int executeUpdate(Query query, Object parameterObject) throws SQLException {
		commitRequired = true;
		int updateCount = queryExecutor.executeUpdate(this, getCurrentConnection(), query, parameterObject);
		if (updateCount > 0) {
			if (query.getCacheName() != null && (query.getCacheStrategy() == CacheStrategy.ReadWrite)) {
				cacheProvider.clearCache(query, parameterObject);
			}
		}
		return updateCount;
	}

	public <E> void executeQuery(Query query, Object parameterObject, Class<E> resultClass, RowHandler<E> rowHandler) throws SQLException {
		if (parameterObject instanceof CacheKey) {
			List<E> list = cacheProvider.getCachedList(this, query, (CacheKey)parameterObject, resultClass);
			if (list != null) {
				for (E obj : list) {
					rowHandler.handleRow(obj);
				}
			}
		} else {
			queryExecutor.executeQuery(this, getCurrentConnection(), query, parameterObject, resultClass, rowHandler);	
		}
	}
	
	public <E> void executeQuery(Query query, Object parameterObject, Class<E> resultClass, RowHandler<E> rowHandler, int pageNo, int pageSize) throws SQLException {
		if (parameterObject instanceof CacheKey) {
			List<E> list = cacheProvider.getCachedList(this, query, (CacheKey)parameterObject, resultClass, pageNo, pageSize);
			if (list != null) {
				for (E obj : list) {
					rowHandler.handleRow(obj);
				}
			}
		} else {
			queryExecutor.executeQuery(this, getCurrentConnection(), query, parameterObject, resultClass, rowHandler, pageNo, pageSize);	
		}
	}
	
	public <E> E executeCallable(Query query, Object parameterObject, Class<E> resultClass) throws SQLException {
		return queryExecutor.executeCallable(this, getCurrentConnection(), query, parameterObject, resultClass);
	}
	
	@Override
	public QueryProvider getQueryProvider() {
		return queryProvider;
	}
	
	@Override
	public Query getQuery(String queryId) throws SQLException {
		return queryProvider.getQuery(queryId);
	}
	
	public PreparedStatement getPreparedStatement(RunnableQuery query) {
		// TODO 동일한 sql인데, resultSetType과 resultSetConcurrency을 다르게 가져가는 경우가 있을까? 나중에 고민해보자.
		return preparedStatements.get(query.getSql());
	}
	
	public void putPreparedStatement(RunnableQuery query, PreparedStatement pstmt) {
		preparedStatements.put(query.getSql(), pstmt);
	}
	
	public boolean hasPreparedStatement(RunnableQuery query) {
		return preparedStatements.containsKey(query.getSql());
	}
	
	public boolean hasPreparedStatement(PreparedStatement pstmt) {
		return preparedStatements.containsValue(pstmt);
	}

}
