/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import kr.kangwoo.damo.engine.IPersistenceService;
import kr.kangwoo.util.MessageUtils;

/**
 * 세션 제공자(SessionProvider)이다. 세션을 생성하고, 소멸시키는 기능과 현재 사용 중인 세션갯수의 정보를 저장하고 있다.
 * 
 * @author badnom
 * @author agoodguy
 *
 */
public class SessionProvider {

	private volatile int maxActive = Integer.MAX_VALUE;
	private final ReentrantLock takeLock = new ReentrantLock();
	private int usedSessions;
	private final Condition notEmpty = takeLock.newCondition();
	private long nanosTimeout = 0;

	/**
	 * 기본 생성자이다.
	 */
	public SessionProvider() {
	}
	
	/**
	 * <p>최대 활동 갯수를 반환한다.</p>
	 * @return 최대 활동 갯수
	 */
	public int getMaxActive() {
		return maxActive;
	}

	/**
	 * <p>동시에 활동 가능한, 최대 활동 갯수를 설정한다.</p>
	 * 
	 * @param maxActive 최대 활동 갯수
	 */
	public void setMaxActive(int maxActive) {
		this.maxActive = maxActive;
		if (this.maxActive < 1) {
			this.maxActive = Integer.MAX_VALUE;
		}
	}
	
	/**
	 * <p>대기 시간 값(MILLISECONDS)을 반환한다.</p>
	 * @return
	 */
	public long getTimeout() {
		return TimeUnit.MICROSECONDS.toMillis(nanosTimeout);
	}

	/**
	 * <p>대기 시간 값을 설정한다.</p>
	 * 
	 * @param timeout 대기 시간(MILLISECONDS)
	 */
	public void setTimeout(long timeout) {
		setTimeout(timeout, TimeUnit.MILLISECONDS);
	}
	
	/**
	 * <p>대기 시간 값을 설정한다.</p>
	 * 
	 * @param timeout 대기 시간
	 * @param timeUnit 시간 단위
	 */
	public void setTimeout(long timeout, TimeUnit timeUnit) {
		nanosTimeout = timeUnit.toNanos(timeout);
	}
	
	/**
	 * <p>새로운 세션을 생성한다.
	 * 현재 활동중인 세션의 갯수가 최대 활동 갯수를 넘어갔을 경우는 대기 시간만큼 대기한후, 생성이 가능하면 생성하고, 아니면 에러를 발생시킨다. </p>
	 * @param persistenceSupport
	 * @return
	 */
	public Session createSession(IPersistenceService persistenceSupport) {
		try {
			takeLock.lockInterruptibly();
			try {
				long nanosTimeout = this.nanosTimeout;
				while (usedSessions >= maxActive && nanosTimeout > 0) {
					nanosTimeout = notEmpty.awaitNanos(nanosTimeout);
				}
				
				if (usedSessions < maxActive) {
					Session session = new Session(persistenceSupport);
					++usedSessions;
					return session;
				} else {
					throw new IllegalStateException(MessageUtils.format(Messages.getString("SessionProvider.NOT_AVAILABE_SESSIONS"),  usedSessions, maxActive));	 //$NON-NLS-1$
				}
			} finally {
				takeLock.unlock();
			}
		} catch(InterruptedException ex) {
			throw new IllegalStateException(Messages.getString("SessionProvider.SESSION_INTERRUPTED"), ex);	 //$NON-NLS-1$
		}
	}

	/**
	 * <p>세션을 종료한다.</p>
	 * @param session
	 */
	public void closeSession(Session session) {
		try {
			takeLock.lockInterruptibly();
			try {
				--usedSessions;
				notEmpty.signal();
			} finally {
				takeLock.unlock();
			}
		} catch(InterruptedException ex) {
			throw new IllegalStateException(Messages.getString("SessionProvider.SESSION_INTERRUPTED"), ex);	 //$NON-NLS-1$
		}
	}
}
