/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.config;

import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Properties;

import javax.naming.InitialContext;
import javax.sql.DataSource;

import kr.kangwoo.damo.DamoException;
import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.SimpleDataSource;
import kr.kangwoo.damo.engine.cache.CacheProvider;
import kr.kangwoo.damo.engine.cache.DummyCacheProvider;
import kr.kangwoo.damo.engine.executor.QueryExecutor;
import kr.kangwoo.damo.engine.executor.SimpleQueryExecutor;
import kr.kangwoo.damo.engine.provider.QueryGenerator;
import kr.kangwoo.damo.engine.provider.QueryLoader;
import kr.kangwoo.damo.engine.provider.QueryProvider;
import kr.kangwoo.damo.engine.provider.SimpleQueryProvider;
import kr.kangwoo.damo.engine.provider.generator.SimpleQueryGenerator;
import kr.kangwoo.damo.engine.provider.loader.file.QueryFileLoader;
import kr.kangwoo.damo.engine.provider.loader.file.SimpleQueryFileLoader;
import kr.kangwoo.damo.engine.transaction.BaseTransactionProvider;
import kr.kangwoo.damo.engine.transaction.IConnectionLogManager;
import kr.kangwoo.damo.engine.transaction.TransactionProvider;
import kr.kangwoo.damo.engine.transaction.jdbc.JdbcTransactionProvider;
import kr.kangwoo.damo.external.ehcache.EhCacheProvider;
import kr.kangwoo.damo.support.dbcp.DbcpDataSource;
import kr.kangwoo.util.BeanUtils;
import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.NumberUtils;
import kr.kangwoo.util.StringUtils;
import kr.kangwoo.util.io.ClassPathResource;
import kr.kangwoo.util.io.FileSystemResource;
import kr.kangwoo.util.io.Resource;
import kr.kangwoo.util.logging.Logger;
import kr.kangwoo.util.logging.LoggerFactory;

public class DamoConfig implements IDamoConfig {

	private Logger logger = LoggerFactory.getLogger(DamoConfig.class);
	
	private Properties properties;
	
	private int sessionMaxActive;
	private long sessionMaxWait;
	private int transactionMaxActive;
	private long transactionMaxWait;
	
	private String cacheProviderClassName;
	private CacheProvider cacheProvider;
	
	private String queryLoaderClassName;
	private QueryLoader queryLoader;
	private QueryGenerator queryGenerator;
	private String queryProviderClassName;
	private QueryProvider queryProvider;
	private String queryExecutorClassName;
	private QueryExecutor queryExecutor;
	private String transactionProviderClassName;
	private TransactionProvider transactionProvider;
	
	private DataSource dataSource;
	private String dataSourceType;
	private String dataSourceName;
	private String jdbcDriver;
	private String jdbcUrl;
	private String jdbcUser;
	private String jdbcPassword;

	private String connectionLogManagerClassName;
	private IConnectionLogManager connectionLogManager;
	
	
	public DamoConfig() {
	}
	
	public DamoConfig(URL url) {
		if (url != null) {
			logger.info(MessageUtils.format(Messages.getString("DamoConfig.USE_CONFIG_FILE"), url.getFile())); //$NON-NLS-1$
			properties = new Properties();
			try {
				properties.load(url.openStream());
				init(properties);
			} catch (Exception e) {
				throw new DamoException(MessageUtils.format(Messages.getString("DamoConfig.ERROR_CONFIG_FILE"), url.getFile()), e); //$NON-NLS-1$
			}
		}
	}

	public DamoConfig(Properties properties) {
		logger.info(Messages.getString("DamoConfig.USE_PROPERTIES")); //$NON-NLS-1$
		if (properties != null) {
			logger.debug("properties : " + properties.toString());	
		}
		this.properties = properties;
		init(this.properties);
	}
	
	@SuppressWarnings("unchecked")
	protected void init(Properties properties) {
		if (properties == null) {
			logger.warn(Messages.getString("DamoConfig.PROPERTIES_IS_NULL")); //$NON-NLS-1$
			properties = new Properties();
		}
		
		sessionMaxActive = NumberUtils.toInteger(properties.getProperty("SessionMaxActive"), 0); //$NON-NLS-1$
		sessionMaxWait = NumberUtils.toLong(properties.getProperty("SessionMaxWait"), 0); //$NON-NLS-1$
		transactionMaxActive = NumberUtils.toInteger(properties.getProperty("TransactionMaxActive"), 0); //$NON-NLS-1$
		transactionMaxWait = NumberUtils.toLong(properties.getProperty("TransactionMaxWait"), 0); //$NON-NLS-1$
		
		cacheProviderClassName = properties.getProperty("CacheProvider"); //$NON-NLS-1$
		if (StringUtils.isNotBlank(cacheProviderClassName)) {
			try {
				cacheProvider = (CacheProvider)BeanUtils.newInstance(Class.forName(cacheProviderClassName));
			} catch (Exception e) {
				logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_CACH_PROVIDER"), cacheProviderClassName), e); //$NON-NLS-1$
			}
		}
		if (cacheProvider == null) {
			try {
				cacheProvider = new EhCacheProvider();
				if (logger.isDebugEnabled()) {
					logger.debug(Messages.getString("DamoConfig.USE_EH_CACHE_PROVIDER")); //$NON-NLS-1$
				}				
			} catch (Throwable ex) {
				cacheProvider = DummyCacheProvider.getInstance();
				logger.error(MessageUtils.format(Messages.getString("DamoConfig.INIT_EH_CACHE_PROVIDER"), cacheProviderClassName), ex); //$NON-NLS-1$
			}

		}
		
		queryLoaderClassName = properties.getProperty("QueryLoader"); //$NON-NLS-1$
		// QueryLoader
		if (StringUtils.isNotBlank(queryLoaderClassName)) {
			try {
				queryLoader = (QueryLoader)BeanUtils.newInstance(Class.forName(queryLoaderClassName));
			} catch (Exception e) {
				logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_QUERY_LOADER"), queryLoaderClassName), e); //$NON-NLS-1$
			}
		}
		if (queryLoader == null) {
			queryLoader = new SimpleQueryFileLoader();
			if (logger.isDebugEnabled()) {
				logger.debug(Messages.getString("DamoConfig.USE_SIMPLE_QUERY_FILE_LOADER")); //$NON-NLS-1$
			}
		}
		
		if (queryLoader instanceof QueryFileLoader) {
			List<File> fileList = new ArrayList<File>();
			Enumeration<Object> enumeration = properties.keys();
			while (enumeration.hasMoreElements()) {
				String key = (String)enumeration.nextElement();
				if (key.startsWith("QueryLoader.Files[") && key.endsWith("]")) { //$NON-NLS-1$ //$NON-NLS-2$
					String resourceName = properties.getProperty(key);
					if (StringUtils.isNotBlank(resourceName)) {
						Resource resource = null;
						if (resourceName.startsWith("classpath:")) { //$NON-NLS-1$
							resource = new ClassPathResource(StringUtils.substringAfter(resourceName, "classpath:"), DamoConfig.class); //$NON-NLS-1$
						} else if (resourceName.startsWith(":")) { //$NON-NLS-1$
							resource = new FileSystemResource(StringUtils.substringAfter(resourceName, "file:")); //$NON-NLS-1$
						} else {
							resource = new FileSystemResource(resourceName);		
						}
						try {
							fileList.add(resource.getFile());	
							if (logger.isDebugEnabled()) {
								logger.debug("QueryFileLoader.File : " + resource); //$NON-NLS-1$
							}
						} catch(Exception e) {
							logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_FILE_READ"), resource), e); //$NON-NLS-1$
						}
					}
				}
			}
			
			if (fileList.size() > 0) {
				((QueryFileLoader)queryLoader).init(fileList.toArray(new File[fileList.size()]));
			}
		}
		
		
		// QueryGenerator
		queryGenerator = new SimpleQueryGenerator();
		
		queryProviderClassName = properties.getProperty("QueryProvider"); //$NON-NLS-1$
		// QueryProvider
		if (StringUtils.isNotBlank(queryProviderClassName)) {
			try {
				queryProvider = (QueryProvider)BeanUtils.newInstance(Class.forName(queryProviderClassName));
			} catch (Exception e) {
				logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_QUERY_PROVIDER"), queryProviderClassName), e); //$NON-NLS-1$
			}
		}
		if (queryProvider == null) {
			queryProvider = new SimpleQueryProvider();
			if (logger.isDebugEnabled()) {
				logger.debug(Messages.getString("DamoConfig.USE_SIMPLE_QUERY_PROVIDER")); //$NON-NLS-1$
			}
		}
		if (queryProvider instanceof SimpleQueryProvider) {
			SimpleQueryProvider simpleQueryProvider = (SimpleQueryProvider)queryProvider;
			simpleQueryProvider.setQueryLoader(queryLoader);
			simpleQueryProvider.setQueryGenerator(queryGenerator);
		}
		
		
		queryExecutorClassName = properties.getProperty("QueryExecutor"); //$NON-NLS-1$
		// QueryExecutor
		if (StringUtils.isNotBlank(queryExecutorClassName)) {
			try {
				queryExecutor = (QueryExecutor)BeanUtils.newInstance(Class.forName(queryExecutorClassName));
			} catch (Exception e) {
				logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_QUERY_EXECUTOR"), queryExecutorClassName), e); //$NON-NLS-1$
			}
		}
		if (queryExecutor == null) {
			queryExecutor = new SimpleQueryExecutor();
			logger.debug(Messages.getString("DamoConfig.USE_SIMPLE_QUERY_EXECUTOR")); //$NON-NLS-1$
		}
		
		
		
		
		connectionLogManagerClassName = properties.getProperty("ConnectionLogManager"); //$NON-NLS-1$
		if (StringUtils.isNotBlank(connectionLogManagerClassName)) {
			try {
				Class<IConnectionLogManager> clazz = (Class<IConnectionLogManager>) Class.forName(connectionLogManagerClassName);
				connectionLogManager = BeanUtils.newInstance(clazz);
			} catch (Exception e) {
				logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_CONNECTION_LOG_MANAGER"), connectionLogManagerClassName), e); //$NON-NLS-1$
			}
		}
		
		transactionProviderClassName = properties.getProperty("TransactionProvider"); //$NON-NLS-1$
		// TransactionProvider
		if (StringUtils.isNotBlank(transactionProviderClassName)) {
			try {
				transactionProvider = (TransactionProvider)BeanUtils.newInstance(Class.forName(transactionProviderClassName));
			} catch (Exception e) {
				logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_TRANSACTION_PROVIDER"), transactionProviderClassName), e); //$NON-NLS-1$
			}
		}
		if (transactionProvider == null) {
			transactionProvider = new JdbcTransactionProvider();
			if (logger.isDebugEnabled()) {
				logger.debug(Messages.getString("DamoConfig.USE_JDBC_TRANSACTION_PROVIDER")); //$NON-NLS-1$
			}
		}
		
		if (transactionProvider instanceof BaseTransactionProvider) {
			((BaseTransactionProvider)transactionProvider).setConnectionLogManager(connectionLogManager);
		}
		
		
		dataSourceType = properties.getProperty("DataSource.Type", "class"); //$NON-NLS-1$ //$NON-NLS-2$
		dataSourceName = properties.getProperty("DataSource.Name", "SimpleDataSource"); //$NON-NLS-1$ //$NON-NLS-2$
		jdbcDriver = properties.getProperty("Jdbc.Driver"); //$NON-NLS-1$
		jdbcUrl = properties.getProperty("Jdbc.Url"); //$NON-NLS-1$
		jdbcUser = properties.getProperty("Jdbc.User"); //$NON-NLS-1$
		jdbcPassword = properties.getProperty("Jdbc.Password"); //$NON-NLS-1$
		// DataSource
		if (StringUtils.isNotBlank(dataSourceType)) {
			if ("class".equalsIgnoreCase(dataSourceType)) { //$NON-NLS-1$
				if ("SimpleDataSource".equalsIgnoreCase(dataSourceName) || "kr.kangwoo.damo.SimpleDataSource".equals(dataSourceName)) { //$NON-NLS-1$ //$NON-NLS-2$
					String driverClass = jdbcDriver;
					String url = jdbcUrl;
					String user = jdbcUser;
					String password = jdbcPassword;
					if (StringUtils.isNotBlank(driverClass) && StringUtils.isNotBlank(url)) {
						try {
							dataSource = new SimpleDataSource(driverClass, url, user, password);
							logger.info(MessageUtils.format(Messages.getString("DamoConfig.INITED_SIMPLE_DATA_SOURCE"), dataSource)); //$NON-NLS-1$
						} catch (Exception e) {
							logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_SIMPLE_DATE_SOURCE"), dataSourceName, driverClass, url, user, password), e); //$NON-NLS-1$
						}
					}
				} else if (("DbcpDataSource".equalsIgnoreCase(dataSourceName) || "kr.kangwoo.damo.support.dbcp.DbcpDataSource".equals(dataSourceName)) //$NON-NLS-1$ //$NON-NLS-2$
						&& StringUtils.isNotBlank(jdbcDriver) && StringUtils.isNotBlank(jdbcUrl)) {
					Properties dbcpProperties = new Properties();
					dbcpProperties.setProperty("driverClassName", jdbcDriver); //$NON-NLS-1$
					dbcpProperties.setProperty("url", jdbcUrl); //$NON-NLS-1$
					dbcpProperties.setProperty("username", jdbcUser); //$NON-NLS-1$
					dbcpProperties.setProperty("password", jdbcPassword); //$NON-NLS-1$

					Enumeration<Object> enumeration = properties.keys();
					while (enumeration.hasMoreElements()) {
						String key = (String)enumeration.nextElement();
						if (key.startsWith("dbcp.")) { //$NON-NLS-1$
							dbcpProperties.setProperty(StringUtils.substringAfter(key, "dbcp."), properties.getProperty(key)); //$NON-NLS-1$
						}
					}
					
					try {
						dataSource = new DbcpDataSource(dbcpProperties);
						logger.info(MessageUtils.format(Messages.getString("DamoConfig.INITED_DBCP_DATA_SOURCE"), dataSource)); //$NON-NLS-1$
					} catch (Exception e) {
						logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_DBCP_DATA_SOURCE"), dataSourceName, dbcpProperties), e); //$NON-NLS-1$
					}
				} else {
					try {
						dataSource = (DataSource)BeanUtils.newInstance(Class.forName(dataSourceName));
						logger.info(MessageUtils.format(Messages.getString("DamoConfig.INITED_DATA_SOURCE"), dataSource)); //$NON-NLS-1$
					} catch (Exception e) {
						logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_DATA_SOURCE"), dataSourceName), e); //$NON-NLS-1$
					}
				}
			} else if ("jndi".equalsIgnoreCase(dataSourceType)) { //$NON-NLS-1$
				try {
					InitialContext initCtx = new InitialContext();
					dataSource = (DataSource) initCtx.lookup(dataSourceName);
					logger.info(MessageUtils.format(Messages.getString("DamoConfig.INITED_DATA_SOURCE"), dataSource)); //$NON-NLS-1$
				} catch (Exception e) {
					logger.error(MessageUtils.format(Messages.getString("DamoConfig.ERROR_DATA_SOURCE"), dataSourceName), e); //$NON-NLS-1$
				}
			} else {
				logger.warn(MessageUtils.format(Messages.getString("DamoConfig.NOT_SUPPORTED_DATA_SOURCE_TYPE"), dataSourceType)); //$NON-NLS-1$
			}
		}
	}


	public int getSessionMaxActive() {
		return sessionMaxActive;
	}

	public long getSessionMaxWait() {
		return sessionMaxWait;
	}

	public int getTransactionMaxActive() {
		return transactionMaxActive;
	}

	public long getTransactionMaxWait() {
		return transactionMaxWait;
	}

	public CacheProvider getCacheProvider() {
		return cacheProvider;
	}

	public QueryLoader getQueryLoader() {
		return queryLoader;
	}
	
	public QueryProvider getQueryProvider() {
		return queryProvider;
	}
	
	public QueryExecutor getQueryExecutor() {
		return queryExecutor;
	}
	
	public TransactionProvider getTransactionProvider() {
		return transactionProvider;
	}
	
	public DataSource getDataSource() {
		return dataSource;
	}
}
