/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine;

import java.sql.SQLException;
import java.util.List;
import java.util.Map;

import kr.kangwoo.damo.engine.cache.CacheKey;
import kr.kangwoo.damo.engine.query.Query;

/**
 * 확장된 쿼리 실행 서비스를 정의하고 있다.
 * 
 * @author badnom
 * @author agoodguy
 *
 */
public interface IExtendedExecuteService extends IExecuteService {
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 INSERT 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 * @param entityObject 엔티티 오브젝트
	 * @return 처리된 로우(row) 수
	 * @throws SQLException
	 */
	int insert(Object entityObject) throws SQLException;
	
	/**
	 * <p>쿼리 아이디에 해당하는 INSERT 쿼리를 읽어와서, 쿼리를 실행시킨다.</p>
	 * 
	 * @param queryId 쿼리 아이디
	 * @param entityObject 엔티티 오브젝트
	 * @return 처리된 로우(row) 수
	 * @throws SQLException
	 */
	int insert(String queryId, Object entityObject) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 UPDATE 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * UPDATE 쿼리의 조건절(where clause)은 엔티티 오브젝트의 PrimaryKey로 정의되어 있다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @return 처리된 로우(row) 수
	 * @throws SQLException
	 */
	int update(Object entityObject) throws SQLException;
	
	/**
	 * <p>쿼리 아이디에 해당하는 UPDATE 쿼리를 읽어와서, 쿼리를 실행시킨다.</p>
	 * 
	 * @param queryId 쿼리 아이디
	 * @param entityObject 엔티티 오브젝트
	 * @return 처리된 로우(row) 수
	 * @throws SQLException
	 */
	int update(String queryId, Object entityObject) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 UPDATE 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * UPDATE 쿼리의 조건절(where clause)은 엔티티 오브젝트의 PrimaryKey로 정의되어 있고,
	 * 엔티티 오브젝트의 <code>null</code>인 값들은 업데이트 대상에서 제외된다.
	 * 대상 Entity Object는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @return 처리된 로우(row) 수
	 * @throws SQLException
	 */
	int updateSelective(Object entityObject) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 DELETE 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * DELETE 쿼리의 조건절(where clause)은 엔티티 오브젝트의 <code>null</code>이 아닌 값들로 정의된다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @return 처리된 로우(row) 수
	 * @throws SQLException
	 */
	int delete(Object entityObject) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 DELETE 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * DELETE 쿼리의 조건절(where clause)은 엔티티 오브젝트의 PrimaryKey로 정의된다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @return 처리된 로우(row) 수
	 * @throws SQLException
	 */
	int deleteByPrimaryKey(Object entityObject) throws SQLException;
	
	/**
	 * <p>쿼리 아이디에 해당하는 DELETE 쿼리를 읽어와서, 쿼리를 실행시킨다.</p>
	 * 
	 * @param queryId 쿼리 아이디
	 * @param entityObject 엔티티 오브젝트
	 * @return 처리된 로우(row) 수
	 * @throws SQLException
	 */
	int delete(String queryId, Object entityObject) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 SELECT 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의  <code>null</code>이 아닌 값들로 정의된다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 *  
	 * <pre>
	 * Entity param = new Entity();
	 * param.setId(1004);
	 * Entity entity = getObject(param);
	 * </pre>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @return 조건에 해당하는 엔티티 오브젝트
	 * @throws SQLException
	 */
	<E> E getObject(E entityObject) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의  <code>null</code>이 아닌 값들로 조건을 생성한 후, 캐시(cache)로 부터 데이터를 읽어온다. 
	 * 해당하는 조건의 데이터가 캐시(cache)에 없을 경우, 데이터 베이스로 부터 데이터를 읽어 온 후, 캐시(cache)에 자동 저장한 후, 그 값을 반환한다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야하고 
	 * kr.kangwoo.damo.annotation.Cache 라는 어노테이션도 정의되어 있어야한다.</p>
	 * 
	 * <pre>
	 * Entity param = new Entity();
	 * param.setId(1004);
	 * Entity entity = getObject(new CacheKey(param));
	 * </pre>
	 *  
	 * @param cacheKey 엔티티 오브젝트의 캐시키
	 * @return 조건에 해당하는 엔티티 오브젝트
	 * @throws SQLException
	 */
	<E> E getObject(CacheKey cacheKey) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 SELECT 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의 PrimaryKey로 정의된다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 *  
	 * <pre>
	 * Entity param = new Entity();
	 * param.setId(1004);
	 * Entity entity = getObject(param);
	 * </pre>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @return 조건에 해당하는 엔티티 오브젝트
	 * @throws SQLException
	 */
	<E> E getObjectByPrimaryKey(E entityObject) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의  PrimaryKey로 정의된 값들로 조건을 생성한 후, 캐시(cache)로 부터 데이터를 읽어온다. 
	 * 해당하는 조건의 데이터가 캐시(cache)에 없을 경우, 데이터 베이스로 부터 데이터를 읽어 온 후, 캐시(cache)에 자동 저장한 후, 그 값을 반환한다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야하고 
	 * kr.kangwoo.damo.annotation.Cache 라는 어노테이션도 정의되어 있어야한다.</p>
	 * 
	 * <pre>
	 * Entity param = new Entity();
	 * param.setId(1004);
	 * Entity entity = getObject(new CacheKey(param));
	 * </pre>
	 * 
	 * @param cacheKey 엔티티 오브젝트의 캐시키
	 * @return 조건에 해당하는 엔티티 오브젝트
	 * @throws SQLException
	 */
	<E> E getObjectByPrimaryKey(CacheKey cacheKey) throws SQLException;
	
	/**
	 * <p>쿼리 아이디에 해당하는 SELECT 쿼리를 읽어와서, 쿼리를 실행시킨다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.</p>
	 * 
	 * @param queryId 쿼리 아이디
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스
	 * @return
	 * @throws SQLException
	 */
	<E> E getObject(String queryId, Object parameterObject, Class<E> resultClass) throws SQLException;
	
	/**
	 * <p>SELECT 쿼리를 실행시킨다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.</p>
	 * 
	 * @param query 실행할 쿼리
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스
	 * @return
	 * @throws SQLException
	 */
	<E> E getObject(Query query, Object parameterObject, Class<E> resultClass) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 COUNT(*)가 포함된 SELECT 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의 <code>null</code>이 아닌 값들로 정의된다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @param resultClass 결과 클래스(Number형이어야함)
	 * @return
	 * @throws SQLException
	 */
	<E, N extends Number> N getCount(E entityObject, Class<N> resultClass) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의  <code>null</code>이 아닌 값들로 조건을 생성한 후, 캐시(cache)로 부터 데이터를 읽어온다. 
	 * 해당하는 조건의 데이터가 캐시(cache)에 없을 경우, 데이터 베이스로 부터 데이터를 읽어 온 후, 캐시(cache)에 자동 저장한 후, 그 값을 반환한다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야하고 
	 * kr.kangwoo.damo.annotation.Cache 라는 어노테이션도 정의되어 있어야한다.</p>
	 * 
	 * @param cacheKey 파라메터 오브젝트를 포함한 캐시키
	 * @param resultClass 결과 클래스(Number형이어야함)
	 * @return
	 * @throws SQLException
	 */
	<E, N extends Number> N getCount(CacheKey cacheKey, Class<N> resultClass) throws SQLException;
	
	/**
	 * <p>쿼리 아이디에 해당하는 SELECT 쿼리를 읽어와서, 쿼리를 실행시킨다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.</p>
	 * 
	 * @param query 실행할 쿼리
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스 (Number형이어야함)
	 * @return
	 * @throws SQLException
	 */
	<N extends Number> N getCount(Query query, Object parameterObject, Class<N> resultClass) throws SQLException;
	
	/**
	 * <p>쿼리 아이디에 해당하는 SELECT 쿼리를 읽어와서, 쿼리를 실행시킨다.
	 * 쿼리 실행의 결과는 없거나, 1개여야만 한다. 복수개의 결과가 반환될경우 에러를 발생시킨다.</p>
	 * 
	 * @param queryId 쿼리 아이디
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스 (Number형이어야함)
	 * @return
	 * @throws SQLException
	 */
	<N extends Number> N getCount(String queryId, Object parameterObject, Class<N> resultClass) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 SELECT 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의  <code>null</code>이 아닌 값들로 정의된다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 *  
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * List<Entity> entityList = getList(param);
	 * </pre>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @return 조건에 해당하는 엔티티 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(E entityObject) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의  <code>null</code>이 아닌 값들로 조건을 생성한 후, 캐시(cache)로 부터 데이터를 읽어온다. 
	 * 해당하는 조건의 데이터가 캐시(cache)에 없을 경우, 데이터 베이스로 부터 데이터를 읽어 온 후, 캐시(cache)에 자동 저장한 후, 그 값을 반환한다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야하고 
	 * kr.kangwoo.damo.annotation.Cache 라는 어노테이션도 정의되어 있어야한다.</p>
	 * 
	 * <pre>
	 * Entity param = new Entity();
	 * param.setId(1004);
	 * List<Entity> entityList = getList(new CacheKey(param));
	 * </pre>
	 *  
	 * @param cacheKey 엔티티 오브젝트의 캐시키
	 * @return 조건에 해당하는 엔티티 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(CacheKey cacheKey) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 SELECT 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의  <code>null</code>이 아닌 값들로 정의된다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 *  
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * List<Entity> entityList = getList(param, "NAME ASC");
	 * </pre>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @param orderBy 정렬문구
	 * @return 조건에 해당하는 엔티티 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(E entityObject, String orderBy) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의  <code>null</code>이 아닌 값들로 조건을 생성한 후, 캐시(cache)로 부터 데이터를 읽어온다. 
	 * 해당하는 조건의 데이터가 캐시(cache)에 없을 경우, 데이터 베이스로 부터 데이터를 읽어 온 후, 캐시(cache)에 자동 저장한 후, 그 값을 반환한다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야하고 
	 * kr.kangwoo.damo.annotation.Cache 라는 어노테이션도 정의되어 있어야한다.</p>
	 * 
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * List<Entity> entityList = getList(new CacheKey(param), "NAME ASC");
	 * </pre>
	 *  
	 * @param cacheKey 엔티티 오브젝트의 캐시키
	 * @param orderBy 정렬문구
	 * @return 조건에 해당하는 엔티티 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(CacheKey cacheKey, String orderBy) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 SELECT 쿼리를 생성한 후, 페이징(Paging) 처리하여 쿼리를 실행시킨다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의  <code>null</code>이 아닌 값들로 정의된다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 *  
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * List<Entity> entityList = getList(param, 1, 10);
	 * </pre>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @param pageNo 페이지 번호(1부터 시작한다.)
	 * @param pageSize 페이지 크기(0보다 커야한다.)
	 * @return 조건에 해당하는 엔티티 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(E entityObject, int pageNo, int pageSize) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의  <code>null</code>이 아닌 값들로 조건을 생성한 후, 캐시(cache)로 부터 데이터를 읽어온다. 
	 * 해당하는 조건의 데이터가 캐시(cache)에 없을 경우, 데이터 베이스로 부터 데이터를 읽어 온 후, 캐시(cache)에 자동 저장한 후, 그 값을 반환한다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야하고 
	 * kr.kangwoo.damo.annotation.Cache 라는 어노테이션도 정의되어 있어야한다.</p>
	 * 
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * List<Entity> entityList = getList(new CacheKey(param), 1, 10);
	 * </pre>
	 *  
	 * @param cacheKey 엔티티 오브젝트의 캐시키
	 * @param pageNo 페이지 번호(1부터 시작한다.)
	 * @param pageSize 페이지 크기(0보다 커야한다.)
	 * @return 조건에 해당하는 엔티티 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(CacheKey cacheKey, int pageNo, int pageSize) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 SELECT 쿼리를 생성한 후, 페이징(Paging) 처리하여 쿼리를 실행시킨다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의  <code>null</code>이 아닌 값들로 정의된다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 *  
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * List<Entity> entityList = getList(param, "NAME ASC", 1, 10);
	 * </pre>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @param orderBy 정렬문구
	 * @param pageNo 페이지 번호(1부터 시작한다.)
	 * @param pageSize 페이지 크기(0보다 커야한다.)
	 * @return 조건에 해당하는 엔티티 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(E entityObject, String orderBy, int pageNo, int pageSize) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의  <code>null</code>이 아닌 값들로 조건을 생성한 후, 캐시(cache)로 부터 데이터를 읽어온다. 
	 * 해당하는 조건의 데이터가 캐시(cache)에 없을 경우, 데이터 베이스로 부터 데이터를 읽어 온 후, 캐시(cache)에 자동 저장한 후, 그 값을 반환한다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야하고 
	 * kr.kangwoo.damo.annotation.Cache 라는 어노테이션도 정의되어 있어야한다.</p>
	 * 
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * List<Entity> entityList = getList(new CacheKey(param, "NAME ASC"));
	 * </pre>
	 *  
	 * @param cacheKey 엔티티 오브젝트의 캐시키
	 * @param orderBy 정렬문구
	 * @param pageNo 페이지 번호(1부터 시작한다.)
	 * @param pageSize 페이지 크기(0보다 커야한다.)
	 * @return 조건에 해당하는 엔티티 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(CacheKey cacheKey, String orderBy, int pageNo, int pageSize) throws SQLException;
	
	/**
	 * <p>SELECT 쿼리를 실행시킨다.</p>
	 * 
	 * @param query 실행할 쿼리
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스
	 * @return 조건에 해당하는 결과 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(Query query, Object parameterObject, Class<E> resultClass) throws SQLException;
	
	/**
	 * <p>쿼리 아이디에 해당하는 SELECT 쿼리를 읽어와서, 쿼리를 실행시킨다.</p>
	 * 
	 * @param queryId 쿼리 아이디
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스
	 * @return 조건에 해당하는 결과 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(String queryId, Object parameterObject, Class<E> resultClass) throws SQLException;
	
	/**
	 * <p>SELECT 쿼리를 실행시킨다.</p>
	 * 
	 * @param query 실행할 쿼리
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스
	 * @param pageNo 페이지 번호(1부터 시작한다.)
	 * @param pageSize 페이지 크기(0보다 커야한다.)
	 * @return 조건에 해당하는 결과 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(Query query, Object parameterObject, Class<E> resultClass, int pageNo, int pageSize) throws SQLException;
	
	/**
	 * <p>쿼리 아이디에 해당하는 SELECT 쿼리를 읽어와서, 쿼리를 실행시킨다.</p>
	 * 
	 * @param queryId 쿼리 아이디
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스
	 * @param pageNo 페이지 번호(1부터 시작한다.)
	 * @param pageSize 페이지 크기(0보다 커야한다.)
	 * @return 조건에 해당하는 결과 오브젝트들의 목록
	 * @throws SQLException
	 */
	<E> List<E> getList(String queryId, Object parameterObject, Class<E> resultClass, int pageNo, int pageSize) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 SELECT 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의  <code>null</code>이 아닌 값들로 정의된다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 *  
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * Map<Integer, Entity> entityMap = getMap(param, "id");
	 * </pre>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @param keyProperty 맵(Map)의 키(key)에 사용할 프로퍼티(property)명
	 * @return 조건에 해당하는 엔티티 오브젝트들의 집합
	 * @throws SQLException
	 */
	<K, V> Map<K, V> getMap(V entityObject, String keyProperty) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의  <code>null</code>이 아닌 값들로 조건을 생성한 후, 캐시(cache)로 부터 데이터를 읽어온다. 
	 * 해당하는 조건의 데이터가 캐시(cache)에 없을 경우, 데이터 베이스로 부터 데이터를 읽어 온 후, 캐시(cache)에 자동 저장한 후, 그 값을 반환한다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야하고 
	 * kr.kangwoo.damo.annotation.Cache 라는 어노테이션도 정의되어 있어야한다.</p>
	 * 
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * Map<Integer, Entity> entityMap = getMap(new CacheKey(param), "id");
	 * </pre>
	 *  
	 * @param cacheKey 엔티티 오브젝트의 캐시키
	 * @param keyProperty 맵(Map)의 키(key)에 사용할 프로퍼티(property)명
	 * @return 조건에 해당하는 엔티티 오브젝트들의 집합
	 * @throws SQLException
	 */
	<K, V> Map<K, V> getMap(CacheKey cacheKey, String keyProperty) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의 정보를 가지고, 자동으로 SELECT 쿼리를 생성한 후, 쿼리를 실행시킨다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의  <code>null</code>이 아닌 값들로 정의된다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.</p>
	 *  
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * Map<Integer, Entity> entityMap = getMap(param, "NAME ASC", "id");
	 * </pre>
	 * 
	 * @param entityObject 엔티티 오브젝트
	 * @param orderBy 정렬문구
	 * @param keyProperty 맵(Map)의 키(key)에 사용할 프로퍼티(property)명
	 * @return 조건에 해당하는 엔티티 오브젝트들의 집합
	 * @throws SQLException
	 */
	<K, V> Map<K, V> getMap(V entityObject, String orderBy, String keyProperty) throws SQLException;
	
	/**
	 * <p>엔티티 오브젝트의  <code>null</code>이 아닌 값들로 조건을 생성한 후, 캐시(cache)로 부터 데이터를 읽어온다. 
	 * 해당하는 조건의 데이터가 캐시(cache)에 없을 경우, 데이터 베이스로 부터 데이터를 읽어 온 후, 캐시(cache)에 자동 저장한 후, 그 값을 반환한다.
	 * 대상 엔티티 오브젝트는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야하고 
	 * kr.kangwoo.damo.annotation.Cache 라는 어노테이션도 정의되어 있어야한다.</p>
	 * 
	 * <pre>
	 * Entity param = new Entity();
	 * param.setName("바보");
	 * Map<Integer, Entity> entityMap = getMap(new CacheKey(param), "NAME ASC", "id");
	 * </pre>
	 *  
	 * @param cacheKey 엔티티 오브젝트의 캐시키
	 * @param orderBy 정렬문구
	 * @param keyProperty 맵(Map)의 키(key)에 사용할 프로퍼티(property)명
	 * @return 조건에 해당하는 엔티티 오브젝트들의 집합
	 * @throws SQLException
	 */
	<K, V> Map<K, V> getMap(CacheKey cacheKey, String orderBy, String keyProperty) throws SQLException;

	/**
	 * <p>SELECT 쿼리를 실행시킨다.</p>
	 * 
	 * @param query 실행할 쿼리
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스
	 * @param keyProperty 맵(Map)의 키(key)에 사용할 프로퍼티(property)명
	 * @return 조건에 해당하는 결과 오브젝트들의 집합
	 * @throws SQLException
	 */
	<K, V> Map<K, V> getMap(Query query, Object parameterObject, Class<V> resultClass, String keyProperty) throws SQLException;
	
	/**
	 * <p>쿼리 아이디에 해당하는 SELECT 쿼리를 읽어와서, 쿼리를 실행시킨다.</p>
	 * 
	 * @param query 실행할 쿼리
	 * @param parameterObject 파라메터 오브젝트 또는 파라메터 오브젝트를 포함한 캐시키(CacheKey)
	 * @param resultClass 결과 클래스
	 * @param keyProperty 맵(Map)의 키(key)에 사용할 프로퍼티(property)명
	 * @return 조건에 해당하는 결과 오브젝트들의 집합
	 * @throws SQLException
	 */
	<K, V> Map<K, V> getMap(String queryId, Object parameterObject, Class<V> resultClass, String keyProperty) throws SQLException;
}
