/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.cache;

import java.io.Serializable;

import kr.kangwoo.damo.DamoException;
import kr.kangwoo.damo.Messages;
import kr.kangwoo.util.ArrayUtils;
import kr.kangwoo.util.BeanUtils;
import kr.kangwoo.util.ObjectUtils;
import kr.kangwoo.util.bean.ClassInfo;
import kr.kangwoo.util.bean.Invoker;

/**
 * 캐시를 식별할 고유키이다.
 * 
 * @author badnom
 * @author agoodguy
 *
 */
public class CacheKey implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = -8051690270985717062L;
	private Object parameterObject;
	private Object[] keyAttributes;
	
	/**
	 * 생성자
	 * 
	 * @param parameterObject 파라메터 객체
	 */
	public CacheKey(Object parameterObject) {
		this.parameterObject = parameterObject;
		this.keyAttributes = null;
	}
	
	/**
	 * 생성자
	 * @param cacheKey 캐시키
	 * @param keyAttributes 키 속성들
	 */
	public CacheKey(CacheKey cacheKey, Object... keyAttributes) {
		this.parameterObject = cacheKey.getParameterObject();
		this.keyAttributes = ArrayUtils.concat(cacheKey.keyAttributes, keyAttributes);
	}

	/**
	 * <p>파라메터 객체를 반환한다.</p>
	 * 
	 * @return 파라메터 객체
	 */
	public Object getParameterObject() {
		return parameterObject;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof CacheKey) {
			try {
				CacheKey another = (CacheKey)obj;
				if (parameterObject.getClass() == another.parameterObject.getClass()) {
					ClassInfo<?> classInfo = ClassInfo.getInstance(parameterObject.getClass());
					Invoker[] invokers = classInfo.getGetInvokers();
					if (invokers != null) {
						for (Invoker invoker : invokers) {
							if (ObjectUtils.equals(invoker.invoke(parameterObject, BeanUtils.NO_ARGUMENTS), invoker.invoke(another.parameterObject, BeanUtils.NO_ARGUMENTS)) == false) {
								return false;
							}
						}
					}
				}
				return ArrayUtils.equals(keyAttributes, another.keyAttributes);
			} catch (Exception e) {
				throw new DamoException(Messages.getString("CacheKey.EQUALS_ERROR"), e); //$NON-NLS-1$
			}
		}
		return false;
	}

	@Override
	public int hashCode() {
		int h = 0;
		try {
			ClassInfo<?> classInfo = ClassInfo.getInstance(parameterObject.getClass());
			Invoker[] invokers = classInfo.getGetInvokers();
			if (invokers != null) {
				for (Invoker invoker : invokers) {
					h += ObjectUtils.hashCode(invoker.invoke(parameterObject, BeanUtils.NO_ARGUMENTS));
				}
			}
			if (keyAttributes != null) {
				for (Object obj : keyAttributes) {
					h += ObjectUtils.hashCode(obj);
				}
			}

		} catch (Exception e) {
			throw new DamoException(Messages.getString("CacheKey.HASH_CODE_ERROR"), e); //$NON-NLS-1$
		}
		return h;
	}
	
}
