/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.executor;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.concurrent.ConcurrentHashMap;

import kr.kangwoo.damo.Session;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.damo.engine.query.RunnableQuery;
import kr.kangwoo.damo.engine.query.StaticQuery;
import kr.kangwoo.damo.engine.query.parameter.ParameterInfo;
import kr.kangwoo.damo.engine.query.result.RowHandler;
import kr.kangwoo.util.SystemUtils;

/**
 * 오라클용 쿼리 실행기이다.
 * 
 * @author badnom
 *
 */
public class OracleQueryExecutor extends SimpleQueryExecutor {

	private ConcurrentHashMap<String, RunnableQuery> pagingQueryMap = new ConcurrentHashMap<String, RunnableQuery>();
	
	public OracleQueryExecutor() {
	}

	private RunnableQuery getPagingQuery(RunnableQuery runnableQuery) throws SQLException {
		RunnableQuery pagingQuery = pagingQueryMap.get(runnableQuery.getSql());
		if (pagingQuery == null) {
			StringBuilder sql = new StringBuilder();
			sql.append("SELECT * FROM (").append(SystemUtils.LINE_SEPARATOR); //$NON-NLS-1$
			sql.append("\tSELECT ROWNUM T_RNUM, T_PAGING.* FROM (").append(SystemUtils.LINE_SEPARATOR); //$NON-NLS-1$
			sql.append(runnableQuery.getSql()).append(SystemUtils.LINE_SEPARATOR);
			sql.append("\t) T_PAGING ").append(SystemUtils.LINE_SEPARATOR); //$NON-NLS-1$
			sql.append("\tWHERE ROWNUM <= ? ").append(SystemUtils.LINE_SEPARATOR); //$NON-NLS-1$
			sql.append(") WHERE T_RNUM > ?").append(SystemUtils.LINE_SEPARATOR); //$NON-NLS-1$
			StaticQuery query = new StaticQuery(runnableQuery.getCommandType(), sql.toString(), runnableQuery.getParameterInfos(), runnableQuery.getDescription());
			query.setFetchSize(runnableQuery.getFetchSize());
			query.setResultSetType(runnableQuery.getResultSetType());
			query.setResultSetConcurrency(runnableQuery.getResultSetConcurrency());
			pagingQuery = query;
			
			RunnableQuery temp = pagingQueryMap.putIfAbsent(runnableQuery.getSql(), pagingQuery);
			if (temp != null) {
				pagingQuery = temp; 
			}
		}
		return pagingQuery;
	}
	
	private void setParameter(Session session, PreparedStatement pstmt,
			RunnableQuery query, Object parameterObject, Integer pageNo, Integer pageSize) throws SQLException {
		super.bindParameters(session, pstmt, query, parameterObject);
		if (pageNo != null && pageSize != null) {
			ParameterInfo<?>[] inputParameterInfos = getInputParameterInfos(query);
			int index = (inputParameterInfos != null) ? inputParameterInfos.length + 1 : 1;
			int endRow = ((pageNo - 1) * pageSize);
			int beginRow = endRow - pageSize;
			
			pstmt.setInt(index++, endRow);
			pstmt.setInt(index++, beginRow);
		}
	}
	/**
	 * <p>결과값을 돌려주는 SQL을 실행한다.(자동적으로 페이징 처리를 한후, 해당 페이지만 돌려준다.)</p>
	 * <p>(여기서는 서브쿼리를 이용해서 페이지 이동을 하였다.)</p>
	 * 
	 * @param <E>
	 * @param session
	 * @param conn
	 * @param query
	 * @param parameterObject 파라메터 객체
	 * @param resultClass
	 * @param rowHandler
	 * @param pageNo 페이지번호
	 * @param pageSize 페이지크기
	 * @throws SQLException
	 */
	@Override
	public <E> void executeQuery(Session session, Connection conn,
			Query query, Object parameterObject, Class<E> resultClass, RowHandler<E> rowHandler, int pageNo, int pageSize) throws SQLException {
		PreparedStatement pstmt = null;
		ResultSet rs = null;
		try {
			RunnableQuery runnableQuery = query.getRunnableQuery(parameterObject);
			runnableQuery = getPagingQuery(runnableQuery);
			// ResultSetType
			pstmt = prepareStatement(session, conn, runnableQuery);
			// FetchSize
			if (runnableQuery.getFetchSize() != null) {
				pstmt.setFetchSize(runnableQuery.getFetchSize());
			}
			setParameter(session, pstmt, runnableQuery, parameterObject, pageNo, pageSize);
			pstmt.execute();
			rs = handleResult(session, conn, pstmt, runnableQuery, parameterObject, resultClass, rowHandler, null, null);
		} finally {
			closeResultSet(rs);
			closeStatement(session, pstmt);
		}
	}

}
