/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.executor;

import java.sql.Connection;
import java.sql.SQLException;

import kr.kangwoo.damo.Session;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.damo.engine.query.RunnableQuery;
import kr.kangwoo.damo.engine.query.result.RowHandler;

/**
 * 쿼리 실행기이다.
 * 
 * @author badnom
 *
 */
public interface QueryExecutor {

	/**
	 * <p>INSERT, UPDATE, DELETE SQL을 실행하거나, 결과값이 없는 SQL을 실행한다.</p>
	 * 
	 * @param session
	 * @param conn
	 * @param query 쿼리
	 * @param parameterObject 파라메터 객체
	 * @return
	 * @throws SQLException
	 */
	public abstract int executeUpdate(Session session, Connection conn,
			Query query, Object parameterObject) throws SQLException;

	/**
	 * <p>결과값을 돌려주는 SQL을 실행한다.</p>
	 * 
	 * @param session
	 * @param conn
	 * @param query 쿼리
	 * @param parameterObject 파라메터 객체
	 * @param resultClass
	 * @param rowHandler
	 * @throws SQLException
	 */
	public abstract <E> void executeQuery(Session session, Connection conn,
			Query query, Object parameterObject, Class<E> resultClass,
			RowHandler<E> rowHandler) throws SQLException;
	
	/**
	 * <p>결과값을 돌려주는 SQL을 실행한다.(자동적으로 페이징 처리를 한후, 해당 페이지만 돌려준다.)</p>
	 * 
	 * @param session
	 * @param conn
	 * @param query
	 * @param parameterObject 파라메터 객체
	 * @param resultClass
	 * @param rowHandler
	 * @param pageNo 페이지번호 (1부터 시작한다.)
	 * @param pageSize 페이지크기 (0보다 커야한다)
	 * @throws SQLException
	 */
	public abstract <E> void executeQuery(Session session, Connection conn,
			Query query, Object parameterObject, Class<E> resultClass, RowHandler<E> rowHandler, int pageNo, int pageSize) throws SQLException;

	/**
	 * <p>프로시저(Procedure)를 실행한다.</p>
	 * 
	 * @param session
	 * @param conn
	 * @param query
	 * @param parameterObject
	 * @param resultClass
	 * @return
	 * @throws SQLException
	 */
	public abstract <E> E executeCallable(Session session, Connection conn, Query query, Object parameterObject, Class<E> resultClass) throws SQLException;
	
	/**
	 * <p>배치를 추가한다.</p>
	 * 
	 * @param session
	 * @param conn
	 * @param query
	 * @param parameterObject
	 * @throws SQLException
	 */
	public abstract void addBatch(Session session, Connection conn, RunnableQuery query, Object parameterObject) throws SQLException;
	
	/**
	 * <p>배치를 실행한다.</p>
	 * 
	 * @param session
	 * @param conn
	 * @param query
	 * @return
	 * @throws SQLException
	 */
	public abstract int[] executeBatch(Session session, Connection conn, RunnableQuery query) throws SQLException;
	
	
}
