/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.executor;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import kr.kangwoo.damo.DamoException;
import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.Session;
import kr.kangwoo.damo.engine.executor.interceptor.ExecuteInterceptor;
import kr.kangwoo.damo.engine.executor.interceptor.HandleRowInterceptor;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.damo.engine.query.RunnableQuery;
import kr.kangwoo.damo.engine.query.parameter.ParameterInfo;
import kr.kangwoo.damo.engine.query.parameter.type.TypeHandlerFactory;
import kr.kangwoo.damo.engine.query.result.RowHandler;
import kr.kangwoo.util.BeanUtils;
import kr.kangwoo.util.ClassUtils;
import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.StringUtils;
import kr.kangwoo.util.SystemUtils;
import kr.kangwoo.util.bean.Invoker;

/**
 * 간단한 쿼리 실행기이다.
 * 
 * @author badnom
 *
 */
public class SimpleQueryExecutor implements QueryExecutor {

	protected TypeHandlerFactory typeHandlerFactory = TypeHandlerFactory.getInstance();
	protected boolean printDescription = true;
	
	public SimpleQueryExecutor() {
	}

	public boolean isPrintDescription() {
		return printDescription;
	}

	public void setPrintDescription(boolean printDescription) {
		this.printDescription = printDescription;
	}

	/**
	 * <p>INSERT, UPDATE, DELETE SQL을 실행하거나, 결과값이 없는 SQL을 실행한다.</p>
	 * 
	 * @param session
	 * @param conn
	 * @param query 쿼리
	 * @param parameterObject 파라메터 객체
	 * @return
	 * @throws SQLException
	 * @see kr.kangwoo.damo.engine.executor.QueryExecutor#executeUpdate(kr.kangwoo.damo.Session, java.sql.Connection, kr.kangwoo.damo.engine.query.Query, java.lang.Object)
	 */
	public int executeUpdate(Session session, Connection conn, Query query, Object parameterObject) throws SQLException {
		int updateCount = 0;
		PreparedStatement pstmt = null;
		try {
			final RunnableQuery runnableQuery = query.getRunnableQuery(parameterObject);
			ExecuteInterceptor[] interceptors = query.getExecuteInterceptors();
			boolean hasInterceptors = (interceptors != null) && interceptors.length > 0;
			if (hasInterceptors) {
				for (ExecuteInterceptor interceptor : interceptors) {
					interceptor.beforeExecute(session, conn, runnableQuery, parameterObject);	
				}
			}
			pstmt = prepareStatement(session, conn, runnableQuery);
			bindParameters(session, pstmt, runnableQuery, parameterObject);
			pstmt.execute();
			updateCount = pstmt.getUpdateCount();
			if (hasInterceptors) {
				for (ExecuteInterceptor interceptor : interceptors) {
					interceptor.afterExecute(session, conn, runnableQuery, parameterObject, updateCount);	
				}
			}
		} finally {
			closeStatement(session, pstmt);
		}
		return updateCount;
	}
	
	/**
	 * <p>결과값을 돌려주는 SQL을 실행한다.</p>
	 * 
	 * @param <E>
	 * @param session
	 * @param conn
	 * @param query 쿼리
	 * @param parameterObject 파라메터 객체
	 * @param resultClass
	 * @param rowHandler
	 * @throws SQLException
	 * @see kr.kangwoo.damo.engine.executor.QueryExecutor#executeQuery(kr.kangwoo.damo.Session, java.sql.Connection, kr.kangwoo.damo.engine.query.Query, java.lang.Object, java.lang.Class, kr.kangwoo.damo.engine.query.result.RowHandler)
	 */
	public <E> void executeQuery(Session session, Connection conn,
			Query query, Object parameterObject, Class<E> resultClass, RowHandler<E> rowHandler) throws SQLException {
		PreparedStatement pstmt = null;
		ResultSet rs = null;
		try {
			final RunnableQuery runnableQuery = query.getRunnableQuery(parameterObject);
			ExecuteInterceptor[] interceptors = query.getExecuteInterceptors();
			boolean hasInterceptors = (interceptors != null) && interceptors.length > 0;
			if (hasInterceptors) {
				for (ExecuteInterceptor interceptor : interceptors) {
					interceptor.beforeExecute(session, conn, runnableQuery, parameterObject);	
				}
			}
			pstmt = prepareStatement(session, conn, runnableQuery);
			bindParameters(session, pstmt, runnableQuery, parameterObject);
			pstmt.execute();
			rs = handleResult(session, conn, pstmt, runnableQuery, parameterObject, resultClass, rowHandler, null, null);
			if (hasInterceptors) {
				for (ExecuteInterceptor interceptor : interceptors) {
					interceptor.afterExecute(session, conn, runnableQuery, parameterObject, rowHandler);	
				}
			}
		} finally {
			closeResultSet(rs);
			closeStatement(session, pstmt);
		}
	}
	
	/**
	 * <p>결과값을 돌려주는 SQL을 실행한다.(자동적으로 페이징 처리를 한후, 해당 페이지만 돌려준다.)</p>
	 * <p>(여기서는 커서를 이용해서 페이지 이동을 하였다.)</p>
	 * @param <E>
	 * @param session
	 * @param conn
	 * @param query
	 * @param parameterObject 파라메터 객체
	 * @param resultClass
	 * @param rowHandler
	 * @param pageNo 페이지번호
	 * @param pageSize 페이지크기
	 * @throws SQLException
	 */
	public <E> void executeQuery(Session session, Connection conn,
			Query query, Object parameterObject, Class<E> resultClass, RowHandler<E> rowHandler, int pageNo, int pageSize) throws SQLException {
		PreparedStatement pstmt = null;
		ResultSet rs = null;
		try {
			final RunnableQuery runnableQuery = query.getRunnableQuery(parameterObject);
			ExecuteInterceptor[] interceptors = query.getExecuteInterceptors();
			boolean hasInterceptors = (interceptors != null) && interceptors.length > 0;
			if (hasInterceptors) {
				for (ExecuteInterceptor interceptor : interceptors) {
					interceptor.beforeExecute(session, conn, runnableQuery, parameterObject);	
				}
			}
			pstmt = prepareStatement(session, conn, runnableQuery);
			bindParameters(session, pstmt, runnableQuery, parameterObject);
			pstmt.execute();
			rs = handleResult(session, conn, pstmt, runnableQuery, parameterObject, resultClass, rowHandler, pageNo, pageSize);
			if (hasInterceptors) {
				for (ExecuteInterceptor interceptor : interceptors) {
					interceptor.afterExecute(session, conn, runnableQuery, parameterObject, rowHandler);	
				}
			}
		} finally {
			closeResultSet(rs);
			closeStatement(session, pstmt);
		}
	}
	
	/**
	 * <p>프로시저(Procedure)를 실행한다.</p>
	 * 
	 * @param <E>
	 * @param session
	 * @param conn
	 * @param query
	 * @param parameterObject
	 * @param resultClass
	 * @return
	 * @throws SQLException
	 */
	public <E> E executeCallable(Session session, Connection conn, Query query, Object parameterObject, Class<E> resultClass) throws SQLException {
		E result = null;
		CallableStatement cstmt = null;
		try {
			RunnableQuery runnableQuery = query.getRunnableQuery(parameterObject);
			cstmt = prepareCall(session, conn, runnableQuery);
			bindParameters(session, cstmt, runnableQuery, parameterObject);
			cstmt.execute();
			if (resultClass != null) {
				result = handleResult(cstmt, runnableQuery, resultClass);	
			}
		} finally {
			closeStatement(session, cstmt);
		}
		
		return result;
	}
	
	/**
	 * <p>배치를 추가한다.</p>
	 * 
	 * @param session
	 * @param conn
	 * @param query
	 * @param parameterObject
	 * @throws SQLException
	 */
	public void addBatch(Session session, Connection conn, RunnableQuery query, Object parameterObject) throws SQLException {
		PreparedStatement pstmt = prepareStatement(session, conn, query);
		bindParameters(session, pstmt, query, parameterObject);
		pstmt.addBatch();
	}
	
	/**
	 * <p>배치를 실행한다.</p>
	 * 
	 * @param session
	 * @param conn
	 * @param query
	 * @return
	 * @throws SQLException
	 */
	public int[] executeBatch(Session session, Connection conn, RunnableQuery query) throws SQLException {
		PreparedStatement pstmt = prepareStatement(session, conn, query);
		return pstmt.executeBatch();
	}
	
	/**
	 * <p>쿼리에서 입력 파라메터들을 가져온다.</p>
	 * 
	 * @param query
	 * @return
	 * @throws SQLException
	 */
	protected ParameterInfo<?>[] getInputParameterInfos(RunnableQuery query) throws SQLException {
		ParameterInfo<?>[] inputParameterInfos = null;
		if (query.getParameterInfos() != null) {
			List<ParameterInfo<?>> parameterInfoList = new ArrayList<ParameterInfo<?>>();
			for (ParameterInfo<?> parameterInfo : query.getParameterInfos()) {
				if (parameterInfo.isInputParameter()) {
					parameterInfoList.add(parameterInfo);
				}
			}
			inputParameterInfos = ParameterInfo.toArray(parameterInfoList);
		}
		return inputParameterInfos;
	}
	
	/**
	 * <p>PreparedStatement에 파라메터값들을 설정한다.</p>
	 * 
	 * @param pstmt
	 * @param query
	 * @param parameterObject
	 * @throws SQLException
	 */
	protected void bindParameters(Session session, PreparedStatement pstmt,
			RunnableQuery query, Object parameterObject) throws SQLException {
		if (parameterObject != null) {
			ParameterInfo<?>[] inputParameterInfos = (ParameterInfo<?>[]) getInputParameterInfos(query);
			if (inputParameterInfos != null) {
				if (typeHandlerFactory.hasTypeHandler(parameterObject.getClass())) {
					if (inputParameterInfos.length > 0) {
						inputParameterInfos[0].bindParaemter(pstmt, 1, parameterObject);
					}
				} else {
					int index = 0;
					for (ParameterInfo<?> parameterInfo : inputParameterInfos) {
						if (parameterInfo.isInputParameter()) {
							index++;
							Object value = parameterInfo.invoke(parameterObject);
							parameterInfo.bindParaemter(pstmt, index, value);
						}
					}

				}
			}	
		}

	}
	
	protected void bindParameters(Session session, CallableStatement cstmt,
			RunnableQuery query, Object parameterObject) throws SQLException {
		if (parameterObject != null) {
			ParameterInfo<?>[] parameterInfos = query.getParameterInfos();
			if (parameterInfos != null) {
				if (typeHandlerFactory.hasTypeHandler(parameterObject.getClass())) {
					if (parameterInfos.length > 0) {
						if (parameterInfos[0].isInputParameter()) {
							parameterInfos[0].bindParaemter(cstmt, 1, parameterObject);	
						} else {
							parameterInfos[0].registerOutParameter(cstmt, 1);
						}
					}
				} else {
					int index = 0;
					for (ParameterInfo<?> parameterInfo : parameterInfos) {
						if (parameterInfo.isInputParameter()) {
							index++;
							Object value = parameterInfo.invoke(parameterObject);
							parameterInfo.bindParaemter(cstmt, index, value);
						} else if (parameterInfo.isInputParameter()) {
							index++;
							parameterInfo.registerOutParameter(cstmt, index);
						}
					}
				}
			}
		}
	}
	
	/**
	 * <p>출력 파라메터 정보들을 가져온다. 
	 * 만약 Query에 출력 파라메터가 정의되어 있다면 정의된것을 사용하고, 
	 * 정의되어 있지 않다면 ResultSet을 분석하여 출력 파라메터를 생성한다.
	 * </p>
	 * 
	 * @param rs
	 * @param query
	 * @param resultClass
	 * @return
	 * @throws SQLException
	 */
	@SuppressWarnings("unchecked")
	protected ParameterInfo<?>[] getOutputParameterInfos(ResultSet rs, RunnableQuery query, Class<?> resultClass) throws SQLException {
		ParameterInfo<?>[] outputParameterInfos = null;
		
		// TODO CACHE 추가할것.
		
		// 쿼리에서 파라메터 정보가져오기
		if (query.getParameterInfos() != null) {
			List<ParameterInfo<?>> parameterInfoList = new ArrayList<ParameterInfo<?>>();
			for (ParameterInfo<?> parameterInfo : query.getParameterInfos()) {
				if (parameterInfo.isOutputParameter()) {
					parameterInfoList.add(parameterInfo);
				}
			}
			outputParameterInfos = ParameterInfo.toArray(parameterInfoList);
		}
		
		// 쿼리에 정보가 존재하지 않을 경우 ResultSetMetaData을 분석해서 사용한다.
		if (outputParameterInfos == null || outputParameterInfos.length == 0) {
			List<ParameterInfo<?>> parameterInfoList = new ArrayList<ParameterInfo<?>>();
			ResultSetMetaData rsmd = rs.getMetaData();
			int columnCount = rsmd.getColumnCount();
			if (typeHandlerFactory.hasTypeHandler(resultClass)) {
				// resultClass을 처리할 수 있는 TypeHandler가 존재할경우 첫번째 컬럼만 사용한다.
				if (columnCount > 0) {
					ParameterInfo parameterInfo = new ParameterInfo(ParameterInfo.MODE.OUT);
					parameterInfo.setName(StringUtils.camelString(rsmd.getColumnName(1), false));
					parameterInfo.setColumnName(rsmd.getColumnName(1));
					parameterInfo.setPrecision(rsmd.getPrecision(1));
					parameterInfo.setScale(rsmd.getScale(1));
					parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(resultClass, rsmd, 1));

					parameterInfoList.add(parameterInfo);
				}
			} else {
				if (ClassUtils.isImplementsInterface(resultClass, Map.class)) {
					for (int i = 1; i <= columnCount; i++) {
						ParameterInfo parameterInfo = new ParameterInfo(ParameterInfo.MODE.OUT);
						parameterInfo.setName(StringUtils.camelString(rsmd.getColumnName(i), false));
						parameterInfo.setColumnName(rsmd.getColumnName(i));
						parameterInfo.setPrecision(rsmd.getPrecision(i));
						parameterInfo.setScale(rsmd.getScale(i));
						parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(rsmd, i));

						parameterInfoList.add(parameterInfo);
					}
				} else {
					for (int i = 1; i <= columnCount; i++) {
						String name = StringUtils.camelString(rsmd.getColumnName(i), false);
						Invoker invoker = getSetInvoker(resultClass, name);
						if (invoker != null) {
							ParameterInfo parameterInfo = new ParameterInfo(ParameterInfo.MODE.OUT);
							parameterInfo.setName(name);
							parameterInfo.setColumnName(rsmd.getColumnName(i));
							parameterInfo.setPrecision(rsmd.getPrecision(i));
							parameterInfo.setScale(rsmd.getScale(i));
							parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(invoker.getType(), rsmd, i));

							parameterInfoList.add(parameterInfo);
						}
					}
				}

			}
			outputParameterInfos = ParameterInfo.toArray(parameterInfoList);
		}
		
		return outputParameterInfos;
	}
	
	/**
	 * <p>쿼리 실행 결과를 처리한다.</p>
	 * 
	 * @param <E>
	 * @param session
	 * @param conn
	 * @param pstmt
	 * @param runnableQuery
	 * @param parameterObject
	 * @param resultClass
	 * @param rowHandler
	 * @return
	 * @throws SQLException
	 */
	@SuppressWarnings("unchecked")
	protected <E> ResultSet handleResult(Session session, Connection conn, PreparedStatement pstmt, RunnableQuery runnableQuery, Object parameterObject, Class<E> resultClass, RowHandler<E> rowHandler, Integer pageNo, Integer pageSize) throws SQLException {
		ResultSet rs = pstmt.getResultSet();
		ParameterInfo<?>[] outputParameterInfo = getOutputParameterInfos(rs, runnableQuery, resultClass);
		rowHandler.init(outputParameterInfo);
		
		HandleRowInterceptor[] interceptors = runnableQuery.getHandleRowInterceptors();
		boolean hasInterceptors = (interceptors != null) && (interceptors.length > 0);
		
		// 페이지 이동 처리
		if (pageNo != null && pageSize != null) {
			int skipRow = ((pageNo - 1) * pageSize);
			if (rs.getType() != ResultSet.TYPE_FORWARD_ONLY) {
				rs.absolute(skipRow);	
			} else {
				for (int i = 0; i < skipRow; i++) {
					if (!rs.next()) {
						return rs;
					}
				}
			}
		}

		int rowCount = 0;
		if (typeHandlerFactory.hasTypeHandler(resultClass)) {
			// 결과 클래스가 기본 타입일 경우
			while((pageSize == null || rowCount < pageSize) && rs.next()) {
				for (ParameterInfo<?> parameterInfo : outputParameterInfo) {
					if (parameterInfo.isOutputParameter()) {
						Object value = parameterInfo.getResult(rs);
						if (hasInterceptors) {
							for (HandleRowInterceptor interceptor : interceptors) {
								interceptor.beforeHandleRow(session, conn, runnableQuery, parameterObject, value);
							}
						}
						rowHandler.handleRow((E)value);
						if (hasInterceptors) {
							for (HandleRowInterceptor interceptor : interceptors) {
								interceptor.afterHandleRow(session, conn, runnableQuery, parameterObject, value);
							}
						}
						break;
					}
				}
				rowCount++;
			}
		} else {
			// 결과 클래스가 ValueObject/Collection 일 경우
			while((pageSize == null || rowCount < pageSize) && rs.next()) {
				E valueObject = newInstance(resultClass);
				for (ParameterInfo<?> parameterInfo : outputParameterInfo) {
					if (parameterInfo.isOutputParameter()) {
						Object value = parameterInfo.getResult(rs);
						setProperty(valueObject, parameterInfo.getName(), value);
					}
				}
				if (hasInterceptors) {
					for (HandleRowInterceptor interceptor : interceptors) {
						interceptor.beforeHandleRow(session, conn, runnableQuery, parameterObject, valueObject);
					}
				}
				rowHandler.handleRow(valueObject);
				if (hasInterceptors) {
					for (HandleRowInterceptor interceptor : interceptors) {
						interceptor.afterHandleRow(session, conn, runnableQuery, parameterObject, valueObject);
					}
				}
				rowCount++;
			}
		}
		return rs;
	}
	
	@SuppressWarnings("unchecked")
	protected <E> E handleResult(CallableStatement cstmt, RunnableQuery query, Class<E> resultClass) throws SQLException {
		ParameterInfo<?>[] parameterInfos = query.getParameterInfos();

		int index = 0;
		if (typeHandlerFactory.hasTypeHandler(resultClass)) {
			// 결과 클래스가 기본 타입일 경우
			for (ParameterInfo<?> parameterInfo : parameterInfos) {
				if (parameterInfo.isOutputParameter()) {
					index++;
				} else if (parameterInfo.isOutputParameter()) {
					index++;
					Object value = parameterInfo.getResult(cstmt, index);
					return (E)value;
				}
			}
		} else {
			// 결과 클래스가 ValueObject/Collection 일 경우
			E valueObject = newInstance(resultClass);
			for (ParameterInfo<?> parameterInfo : parameterInfos) {
				if (parameterInfo.isOutputParameter()) {
					index++;
				} else if (parameterInfo.isOutputParameter()) {
					index++;
					Object value = parameterInfo.getResult(cstmt, index);
					setProperty(valueObject, parameterInfo.getName(), value);
				}
			}
			return valueObject;
		}
		return null;
	}
	
	
	/**
	 * <p>대상 객체의 해당하는 속성에 값을 부여한다.</p>
	 * 
	 * @param target 대상 객체
	 * @param name 속성 이름
	 * @param value 속성 값
	 */
	protected void setProperty(Object target, String name, Object value) {
		try {
			BeanUtils.setProperty(target, name, value);
		} catch (Exception e) {
			throw new DamoException(MessageUtils.format(Messages.getString("SimpleQueryExecutor.SET_PROPERTY_ERROR"), target, name, value), e); //$NON-NLS-1$
		}
	}
	
	/**
	 * <p>새로운 인스턴스를 생성한다.</p>
	 * 
	 * @param <T>
	 * @param clazz
	 * @return
	 */
	protected <T> T newInstance(Class<T> clazz) {
		try {
			return BeanUtils.newInstance(clazz);
		} catch (Exception e) {
			throw new DamoException(MessageUtils.format(Messages.getString("SimpleQueryExecutor.INSTANTIATION_EXCEPTION"), clazz), e); //$NON-NLS-1$
		}
	}
	
	/**
	 * <p>대상 클래스에, 해당하는 이름의 값을 설정할 수 있는 호출자를 가져온다.<p>
	 * 
	 * @param resultClass 대상 클래스
	 * @param name 이름
	 * @return
	 */
	protected Invoker getSetInvoker(Class<?> resultClass, String name) {
		try {
			return BeanUtils.getSetInvoker(resultClass, name);	
		} catch (Exception e) {
			// 무시
		}
		return null;
	}
	
	protected String getSqlWithDescription(RunnableQuery query) {
		String sql = query.getSql();
		if (printDescription && StringUtils.isNotBlank(query.getDescription())) {
			StringBuilder sqlWithDesc = new StringBuilder();
			sqlWithDesc.append("/* "); //$NON-NLS-1$
			sqlWithDesc.append(query.getDescription());
			sqlWithDesc.append(" */"); //$NON-NLS-1$
			sqlWithDesc.append(SystemUtils.LINE_SEPARATOR);
			sqlWithDesc.append(sql);
			sql = sqlWithDesc.toString();
		}
		return sql;
	}
	
	/**
	 * <p>Connection으로부터 PreparedStatement을 생성한다. 만약 세션에 동일한 sql에 해당하는 PreparedStatement가 있을 경우 재사용한다.</p>
	 * 
	 * @param session
	 * @param conn
	 * @param sql
	 * @return
	 * @throws SQLException
	 */
	protected PreparedStatement prepareStatement(Session session, Connection conn, RunnableQuery query) throws SQLException {
		return prepareStatement(session, conn, query, true);
	}
	
	protected PreparedStatement prepareStatement(Session session, Connection conn, RunnableQuery runnableQuery, boolean resuse) throws SQLException {
		if (resuse && session.hasPreparedStatement(runnableQuery)) {
			return session.getPreparedStatement(runnableQuery);
		} else {
			String sql = getSqlWithDescription(runnableQuery);
			PreparedStatement pstmt;
			if (runnableQuery.getResultSetType() != null && runnableQuery.getResultSetConcurrency() != null) {
				pstmt = conn.prepareStatement(sql, runnableQuery.getResultSetType(), runnableQuery.getResultSetConcurrency());
				// FetchSize
				if (runnableQuery.getFetchSize() != null) {
					pstmt.setFetchSize(runnableQuery.getFetchSize());
				}
			} else if (runnableQuery.getResultSetType() != null) {
				pstmt = conn.prepareStatement(sql, runnableQuery.getResultSetType(), ResultSet.CONCUR_READ_ONLY);
				// FetchSize
				if (runnableQuery.getFetchSize() != null) {
					pstmt.setFetchSize(runnableQuery.getFetchSize());
				}
			} else {
				pstmt = conn.prepareStatement(sql);
				// FetchSize
				if (runnableQuery.getFetchSize() != null) {
					pstmt.setFetchSize(runnableQuery.getFetchSize());
				}
			}
			if (resuse) {
				session.putPreparedStatement(runnableQuery, pstmt);	
			}
			return pstmt;
		}
	}
	
	protected CallableStatement prepareCall(Session session, Connection conn, RunnableQuery query) throws SQLException {
		if (session.hasPreparedStatement(query)) {
			return (CallableStatement)session.getPreparedStatement(query);
		} else {
			String sql = getSqlWithDescription(query);
			CallableStatement cstmt;
			if (query.getResultSetType() != null && query.getResultSetConcurrency() != null) {
				cstmt = conn.prepareCall(sql, query.getResultSetType(), query.getResultSetConcurrency());
			} else if (query.getResultSetType() != null) {
				cstmt = conn.prepareCall(sql, query.getResultSetType(), ResultSet.CONCUR_READ_ONLY);
			} else {
				cstmt = conn.prepareCall(sql);	
			}
			session.putPreparedStatement(query, cstmt);
			return cstmt;
		}
	}
	
	/**
	 * <p>ResultSet을 닫는다.</p>
	 * @param rs
	 */
	protected void closeResultSet(ResultSet rs) {
		if (rs != null) {
			try {
				rs.close();	
			} catch(SQLException e) {
				// 무시
			}
		}
	}
	
	/**
	 * <p>PreparedStatement을 닫는다. 만약 세션에 존재한다면 닫지 않는다.</p>
	 * 
	 * @param session
	 * @param pstmt
	 */
	protected void closeStatement(Session session, PreparedStatement pstmt) {
		if (pstmt != null) {
			if (!session.hasPreparedStatement(pstmt)) {
				try {
					pstmt.close();	
				} catch(SQLException e) {
					// 무시
				}
			}
		}
	}

}
