/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.provider;

import java.sql.SQLException;

import kr.kangwoo.damo.engine.query.Query;

/**
 * 쿼리 생성기이다.
 * 엔티티 클래스를 분석해서 INSERT/UPDATE/DELETE/SELECT 쿼리를 생성하는 역할을 정의하고 있다.
 * 대상이 되는 엔티티 클래스는 kr.kangwoo.damo.annotation.Table 이라는 어노테이션이 정의되어 있어야한다.
 * 
 * @author badnom
 *
 */
public interface QueryGenerator {
	
	/**
	 * <p>엔티티 클래스를 분석하여, INSERT 쿼리를 생성한다.</p>
	 * 
	 * @param entityClass 엔티티 클래스
	 * @return INSERT 쿼리
	 * @throws SQLException
	 */
	Query getInsertQuery(Class<?> entityClass) throws SQLException;
	
	/**
	 * <p>엔티티 클래스를 분석하여, UPDATE 쿼리를 생성한다.
	 * UPDATE 쿼리의 조건절(where clause)은 엔티티 오브젝트의 PrimaryKey로 정의되어 있다.
	 * </p>
	 * 
	 * @param entityClass 엔티티 클래스
	 * @return
	 * @throws SQLException
	 */
	Query getUpdateQuery(Class<?> entityClass) throws SQLException;
	
	/**
	 * <p>엔티티 클래스를 분석하여, UPDATE 쿼리를 생성한다.
	 * UPDATE 쿼리의 조건절(where clause)은 엔티티 오브젝트의 PrimaryKey로 정의되어 있고,
	 * 엔티티 오브젝트의 <code>null</code>인 값들은 업데이트 대상에서 제외된다.
	 * </p>
	 * 
	 * @param entityClass 엔티티 클래스
	 * @return
	 * @throws SQLException
	 */
	Query getUpdateSelectiveQuery(Class<?> entityClass) throws SQLException;
	
	/**
	 * <p>엔티티 클래스를 분석하여, DELETE 쿼리를 생성한다.
 	 * DELETE 쿼리의 조건절(where clause)은 엔티티 오브젝트의 <code>null</code>이 아닌 값들로 정의된다.
	 * </p> 
	 * 
	 * @param entityClass 엔티티 클래스
	 * @return
	 * @throws SQLException
	 */
	Query getDeleteQuery(Class<?> entityClass) throws SQLException;
	
	/**
	 * <p>엔티티 클래스를 분석하여, DELETE 쿼리를 생성한다.
	 * DELETE 쿼리의 조건절(where clause)은 엔티티 오브젝트의 PrimaryKey로 정의된다.
	 * </p>
	 * 
	 * @param entityClass 엔티티 클래스
	 * @return
	 * @throws SQLException
	 */
	Query getDeleteByPrimaryKeyQuery(Class<?> entityClass) throws SQLException;
	
	/**
	 * <p>엔티티 클래스를 분석하여, SELECT 쿼리를 생성한다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의  <code>null</code>이 아닌 값들로 정의된다.
	 * </p>
	 * 
	 * @param entityClass 엔티티 클래스
	 * @return
	 * @throws SQLException
	 */
	Query getGetObjectQuery(Class<?> entityClass) throws SQLException;
	
	/**
	 * <p>엔티티 클래스를 분석하여, INSERT 쿼리를 생성한다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의 PrimaryKey로 정의된다.
	 * </p>
	 * 
	 * @param entityClass 엔티티 클래스
	 * @return
	 * @throws SQLException
	 */
	Query getGetObjectByPrimaryKeyQuery(Class<?> entityClass) throws SQLException;

	/**
	 * <p>엔티티 클래스를 분석하여, SELECT 쿼리를 생성한다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의 <code>null</code>이 아닌 값들로 정의된다.
	 * </p>
	 * 
	 * @param entityClass 엔티티 클래스
	 * @param resultClass
	 * @return
	 * @throws SQLException
	 */
	<N extends Number> Query getGetCountQuery(Class<?> entityClass, Class<N> resultClass) throws SQLException;
	
	/**
	 * <p>엔티티 클래스를 분석하여, SELECT 쿼리를 생성한다.
	 * SELECT 쿼리의 조건절(where clause)은 엔티티 오브젝트의 <code>null</code>이 아닌 값들로 정의된다.
	 * </p>
	 * 
	 * @param entityClass 엔티티 클래스
	 * @param orderBy 정렬문구
	 * @return
	 * @throws SQLException
	 */
	Query getGetListQuery(Class<?> entityClass, String orderBy) throws SQLException;
	
}
