/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.provider;

import java.sql.SQLException;

import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.engine.provider.generator.SimpleQueryGenerator;
import kr.kangwoo.damo.engine.provider.loader.DummyQueryLoader;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.util.MessageUtils;

/**
 * 간단한 쿼리 제공자(QueryProvider)이다. 
 * 기본적으로는 "SimpleQueryGenerator"라는 쿼리 생성기(Query Generator)를 사용하고, 쿼리 적재기(Query Loader)는 사용을 하지 않는다.
 * 
 * @author badnom
 *
 */
public class SimpleQueryProvider implements QueryProvider {
	
	private QueryLoader queryLoader;
	private QueryGenerator queryGenerator;
	
	public SimpleQueryProvider() {
	}
	
	/**
	 * <p>쿼리 적재기를 가져온다.</p>
	 * 
	 * @return 쿼리 적재기
	 */
	public QueryLoader getQueryLoader() {
		return queryLoader;
	}
	
	/**
	 * <p>쿼리 적재기를 설정한다.</p>
	 * 
	 * @param queryLoader 쿼리 적재기
	 */
	public void setQueryLoader(QueryLoader queryLoader) {
		this.queryLoader = queryLoader;
		if (this.queryLoader == null) {
			this.queryLoader = DummyQueryLoader.getDummyQueryLoader();
		}
	}

	/**
	 * <p>쿼리 생성기를 가져온다.</p>
	 * 
	 * @return 쿼리 생성기
	 */
	public QueryGenerator getQueryGenerator() {
		return queryGenerator;
	}

	/**
	 * <p>쿼리 생성기를 설정한다.</p>
	 * 
	 * @param queryGenerator 쿼리 생성기
	 */
	public void setQueryGenerator(QueryGenerator queryGenerator) {
		this.queryGenerator = queryGenerator;
		if (this.queryGenerator == null) {
			queryGenerator = new SimpleQueryGenerator();
		}
	}

	public Query getQuery(String queryId) throws SQLException {
		Query query = queryLoader.getQuery(queryId);
		if (query == null) {
			throw new SQLException(MessageUtils.format(Messages.getString("SimpleQueryProvider.NOT_FOUND_QUERY"), queryId)); //$NON-NLS-1$
		}
		return queryLoader.getQuery(queryId);
	}
	
	public Query getInsertQuery(Class<?> entityClass) throws SQLException {
		return queryGenerator.getInsertQuery(entityClass);
	}
	
	public Query getUpdateQuery(Class<?> entityClass) throws SQLException {
		return queryGenerator.getUpdateQuery(entityClass);
	}
	
	public Query getUpdateSelectiveQuery(Class<?> entityClass) throws SQLException {
		return queryGenerator.getUpdateSelectiveQuery(entityClass);
	}
		
	public Query getDeleteQuery(Class<?> entityClass) throws SQLException {
		return queryGenerator.getDeleteQuery(entityClass);
	}
	
	public Query getDeleteByPrimaryKeyQuery(Class<?> entityClass) throws SQLException {
		return queryGenerator.getDeleteByPrimaryKeyQuery(entityClass);
	}
	
	public Query getGetObjectQuery(Class<?> entityClass) throws SQLException {
		return queryGenerator.getGetObjectQuery(entityClass);
	}

	public Query getGetObjectByPrimaryKeyQuery(Class<?> entityClass) throws SQLException {
		return queryGenerator.getGetObjectByPrimaryKeyQuery(entityClass);
	}
	
	public <N extends Number> Query getGetCountQuery(Class<?> entityClass, Class<N> resultClass) throws SQLException {
		return queryGenerator.getGetCountQuery(entityClass, resultClass);
	}
	
	public Query getGetListQuery(Class<?> entityClass, String orderBy) throws SQLException {
		return queryGenerator.getGetListQuery(entityClass, orderBy);
	}

}
