/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.provider.generator;

import java.lang.reflect.Array;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.Session;
import kr.kangwoo.damo.annotation.Cache;
import kr.kangwoo.damo.annotation.SelectKey;
import kr.kangwoo.damo.engine.cache.CacheKey;
import kr.kangwoo.damo.engine.executor.interceptor.ExecuteInterceptor;
import kr.kangwoo.damo.engine.executor.interceptor.ExecuteInterceptorAdapter;
import kr.kangwoo.damo.engine.executor.interceptor.HandleRowInterceptor;
import kr.kangwoo.damo.engine.executor.interceptor.HandleRowInterceptorAdapter;
import kr.kangwoo.damo.engine.provider.QueryGenerator;
import kr.kangwoo.damo.engine.provider.generator.model.EntityDefinition;
import kr.kangwoo.damo.engine.provider.generator.model.EntityDefinitionProvider;
import kr.kangwoo.damo.engine.provider.generator.model.FieldDefinition;
import kr.kangwoo.damo.engine.provider.generator.model.JoinColumnDefinition;
import kr.kangwoo.damo.engine.query.BaseQuery;
import kr.kangwoo.damo.engine.query.CommandType;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.damo.engine.query.QueryHelper;
import kr.kangwoo.damo.engine.query.RunnableQuery;
import kr.kangwoo.damo.engine.query.parameter.ParameterInfo;
import kr.kangwoo.damo.engine.query.parameter.type.TypeHandlerFactory;
import kr.kangwoo.damo.engine.query.token.DynamicToken;
import kr.kangwoo.damo.engine.query.token.IfToken;
import kr.kangwoo.damo.engine.query.token.StaticToken;
import kr.kangwoo.damo.engine.query.token.Token;
import kr.kangwoo.util.BeanUtils;
import kr.kangwoo.util.ClassUtils;
import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.StringUtils;
import kr.kangwoo.util.logging.Logger;
import kr.kangwoo.util.logging.LoggerFactory;

/**
 * 간단한 쿼리 생성기이다.
 * 쿼리 요청시, 기존에 만들어진 쿼리가 있을 경우 기존 쿼리를 사용하고 없을 경우, 새로 만들어서 사용한다.
 * 
 * @author badnom
 *
 */
public class SimpleQueryGenerator implements QueryGenerator {
	
	private static final ConcurrentHashMap<Class<?>, ConcurrentHashMap<String, Query>> QUERY_PROVIDER_CACHE_MAP = new ConcurrentHashMap<Class<?>, ConcurrentHashMap<String, Query>>();
	
	private static final String INSERT_QUERY_ID = "Insert"; //$NON-NLS-1$
	private static final String UPDATE_QUERY_ID = "Update"; //$NON-NLS-1$
	private static final String UPDATE_SELECTIVE_QUERY_ID = "UpdateSelectvice"; //$NON-NLS-1$
	private static final String DELETE_QUERY_ID = "Delete"; //$NON-NLS-1$
	private static final String DELETE_BY_PRIMARY_KEY_QUERY_ID = "DeleteByPrimaryKey"; //$NON-NLS-1$
	private static final String GET_OBJECT_QUERY_ID = "getObject"; //$NON-NLS-1$
	private static final String GET_OBJECT_BY_PRIMARY_KEY_QUERY_ID = "getObjectByPrimaryKey"; //$NON-NLS-1$
	private static final String GET_COUNT_QUERY_ID = "getCount"; //$NON-NLS-1$
	private static final String GET_LIST_QUERY_ID = "getList"; //$NON-NLS-1$
	
	private Logger logger = LoggerFactory.getLogger(getClass());
	
	private EntityDefinitionProvider entityDefinitionManager;
	private TypeHandlerFactory typeHandlerFactory;
	
	public SimpleQueryGenerator() {
		this.entityDefinitionManager = new EntityDefinitionProvider();
		this.typeHandlerFactory = TypeHandlerFactory.getInstance();
	}
	
	protected String makeDescription(Class<?> entityClass, String queryId) {
		StringBuilder sb = new StringBuilder();
		sb.append(entityClass.getName());
		sb.append("."); //$NON-NLS-1$
		sb.append(queryId);
		sb.append(" generated by "); //$NON-NLS-1$
		sb.append(getClass().getName());
		return sb.toString();
	}

	public Query getInsertQuery(Class<?> entityClass) throws SQLException {
		ConcurrentHashMap<String, Query> cacheMap = QUERY_PROVIDER_CACHE_MAP.get(entityClass);
		if (cacheMap == null) {
			cacheMap = new ConcurrentHashMap<String, Query>(); 
			ConcurrentHashMap<String, Query> temp = QUERY_PROVIDER_CACHE_MAP.putIfAbsent(entityClass, cacheMap);
			if (temp != null) {
				cacheMap = temp;
			}
		}
		
		Query query = cacheMap.get(INSERT_QUERY_ID);
		if (query == null) {
			query = makeInsertQuery(entityClass);
			Query temp = cacheMap.putIfAbsent(INSERT_QUERY_ID, query);
			if (temp != null) {
				query = temp;
			}
		}
		
		return query;
	}
	
	@SuppressWarnings("unchecked")
	protected Query makeInsertQuery(Class<?> entityClass) throws SQLException {
		EntityDefinition entityDef = entityDefinitionManager.getEntityDefinition(entityClass);
		
		ExecuteInterceptor executeInterceptor = null;
		
		List<Token> tokenList = new ArrayList<Token>();
		
		tokenList.add(new StaticToken("INSERT INTO ")); //$NON-NLS-1$
		tokenList.addAll(QueryHelper.newTokenList(entityDef.getTableName().toString(), entityClass));
		
		StringBuilder columnClause = new StringBuilder();
		columnClause.append("("); //$NON-NLS-1$
		
		StringBuilder valuesClause = new StringBuilder();
		valuesClause.append(" VALUES ("); //$NON-NLS-1$
		
		List<ParameterInfo<?>> parameterInfoList = new ArrayList<ParameterInfo<?>>();
		
		boolean isFirst = true;
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable()) {
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));
				
				if (isFirst) {
					isFirst = false;
				} else {
					columnClause.append(", "); //$NON-NLS-1$
					valuesClause.append(", "); //$NON-NLS-1$
				}
				columnClause.append(columnDef.getNameForQuery());
				valuesClause.append("?"); //$NON-NLS-1$
				parameterInfoList.add(parameterInfo);
				
				executeInterceptor = newExecuteInterceptorForSelectKey(columnDef, entityClass);
			}
		}
		
		columnClause.append(")"); //$NON-NLS-1$
		valuesClause.append(")"); //$NON-NLS-1$
		
		String sql = columnClause.append(valuesClause).toString();
		tokenList.add(new StaticToken(sql, ParameterInfo.toArray(parameterInfoList)));
		
		BaseQuery query = QueryHelper.optimizedQuery(CommandType.INSERT, tokenList.toArray(new Token[tokenList.size()]), makeDescription(entityClass, INSERT_QUERY_ID));
		if (executeInterceptor != null) {
			query.addExecuteInterceptor(executeInterceptor);	
		}
		Cache cache = entityDef.getCache();
		if (cache != null) {
			query.setCacheName(cache.cacheName());
			query.setCacheStrategy(cache.cacheStrategy());
		}
		return query;
	}

	protected ExecuteInterceptor newExecuteInterceptorForSelectKey(
			FieldDefinition columnDef, Class<?> entityClass) throws SQLException {
		ExecuteInterceptor executeInterceptor = null;
		if (columnDef.getSelectKey() != null) {
			SelectKey selectKey = columnDef.getSelectKey();
			if (StringUtils.isNotBlank(selectKey.sql())) {
				final Query query = QueryHelper.newSelectQuery(selectKey.sql(), entityClass);
				final String pName = columnDef.getName();
				final Class<?> pType = columnDef.getJavaType();
				if ("pre".equalsIgnoreCase(selectKey.type()) || "before".equalsIgnoreCase(selectKey.type())) { //$NON-NLS-1$ //$NON-NLS-2$
					executeInterceptor = new ExecuteInterceptorAdapter() {
						private Query selectQuery = query;
						
						@Override
						public void beforeExecute(Session session, Connection conn, RunnableQuery runnableQuery,
								Object parameterObject)
								throws SQLException {
							try {
								if (BeanUtils.getProperty(parameterObject, pName) == null) {
									Object value = session.getObject(selectQuery, parameterObject, pType);
									try {
										BeanUtils.setProperty(parameterObject, pName, value);
									} catch (Exception e) {
										throw new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.SET_PROPERTY_ERROR"), ClassUtils.getClassName(parameterObject), pName, value)); //$NON-NLS-1$
									}
								}
							} catch (Exception e) {
								throw new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.GET_PROPERTY_ERROR"), ClassUtils.getClassName(parameterObject), pName)); //$NON-NLS-1$
							}
						}
					};	
				} else if ("post".equalsIgnoreCase(selectKey.type()) || "after".equalsIgnoreCase(selectKey.type())) { //$NON-NLS-1$ //$NON-NLS-2$
					executeInterceptor = new ExecuteInterceptorAdapter() {
						private Query selectQuery = query;
						@Override
						public void afterExecute(Session session, Connection conn, RunnableQuery runnableQuery,
								Object parameterObject, Object resultObject)
								throws SQLException {
							try {
								if (BeanUtils.getProperty(parameterObject, pName) == null) {
									Object value = session.getObject(selectQuery, parameterObject, pType);
									try {
										BeanUtils.setProperty(parameterObject, pName, value);
									} catch (Exception e) {
										throw new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.SET_PROPERTY_ERROR"), ClassUtils.getClassName(parameterObject), pName, value)); //$NON-NLS-1$
									}
								}
							} catch (Exception e) {
								throw new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.GET_PROPERTY_ERROR"), ClassUtils.getClassName(parameterObject), pName)); //$NON-NLS-1$
							}
						}
					};
				}
			} else if (StringUtils.isNotBlank(selectKey.id())) {
				final String queryId = selectKey.id();
				final String pName = columnDef.getName();
				final Class<?> pType = columnDef.getJavaType();
				if ("pre".equalsIgnoreCase(selectKey.type()) || "before".equalsIgnoreCase(selectKey.type())) { //$NON-NLS-1$ //$NON-NLS-2$
					executeInterceptor = new ExecuteInterceptorAdapter() {
						@Override
						public void beforeExecute(Session session, Connection conn, RunnableQuery runnableQuery,
								Object parameterObject)
								throws SQLException {
							try {
								Query selectQuery = session.getQueryProvider().getQuery(queryId);
								if (BeanUtils.getProperty(parameterObject, pName) == null) {
									Object value = session.getObject(selectQuery, parameterObject, pType);
									try {
										BeanUtils.setProperty(parameterObject, pName, value);
									} catch (Exception e) {
										throw new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.SET_PROPERTY_ERROR"), ClassUtils.getClassName(parameterObject), pName, value)); //$NON-NLS-1$
									}
								}
							} catch (Exception e) {
								throw new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.GET_PROPERTY_ERROR"), ClassUtils.getClassName(parameterObject), pName)); //$NON-NLS-1$
							}
						}
					};	
				} else if ("post".equalsIgnoreCase(selectKey.type()) || "after".equalsIgnoreCase(selectKey.type())) { //$NON-NLS-1$ //$NON-NLS-2$
					executeInterceptor = new ExecuteInterceptorAdapter() {
						@Override
						public void afterExecute(Session session, Connection conn, RunnableQuery runnableQuery,
								Object parameterObject, Object resultObject)
								throws SQLException {
							try {
								Query selectQuery = session.getQueryProvider().getQuery(queryId);
								if (BeanUtils.getProperty(parameterObject, pName) == null) {
									Object value = session.getObject(selectQuery, parameterObject, pType);
									try {
										BeanUtils.setProperty(parameterObject, pName, value);
									} catch (Exception e) {
										throw new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.SET_PROPERTY_ERROR"), ClassUtils.getClassName(parameterObject), pName, value)); //$NON-NLS-1$
									}
								}
							} catch (Exception e) {
								throw new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.GET_PROPERTY_ERROR"), ClassUtils.getClassName(parameterObject), pName)); //$NON-NLS-1$
							}
						}
					};
				}
			} else {
				logger.warn(MessageUtils.format(Messages.getString("SimpleQueryGenerator.SELECT_KEY_ERROR"), ClassUtils.getClassName(entityClass))); //$NON-NLS-1$
			}

		}
		return executeInterceptor;
	}
	
	public Query getUpdateQuery(Class<?> entityClass) throws SQLException {
		ConcurrentHashMap<String, Query> cacheMap = QUERY_PROVIDER_CACHE_MAP.get(entityClass);
		if (cacheMap == null) {
			cacheMap = new ConcurrentHashMap<String, Query>();
			ConcurrentHashMap<String, Query> temp = QUERY_PROVIDER_CACHE_MAP.putIfAbsent(entityClass, cacheMap);
			if (temp != null) {
				cacheMap = temp;
			}
		}
		
		Query query = cacheMap.get(UPDATE_QUERY_ID);
		if (query == null) {
			query = makeUpdateQuery(entityClass);
			Query temp = cacheMap.putIfAbsent(UPDATE_QUERY_ID, query);
			if (temp != null) {
				query = temp;
			}
		}
		
		return query;
	}
	
	@SuppressWarnings("unchecked")
	public Query makeUpdateQuery(Class<?> entityClass) throws SQLException {
		EntityDefinition entityDef = entityDefinitionManager.getEntityDefinition(entityClass);
		
		List<Token> tokenList = new ArrayList<Token>();
		
		tokenList.add(new StaticToken("UPDATE ")); //$NON-NLS-1$
		tokenList.addAll(QueryHelper.newTokenList(entityDef.getTableName().toString(), entityClass));
		
		StringBuilder setClause = new StringBuilder();
		setClause.append(" SET "); //$NON-NLS-1$
		
		StringBuilder whereClause = new StringBuilder();
		whereClause.append(" WHERE "); //$NON-NLS-1$
		
		List<ParameterInfo<?>> parameterInfoList = new ArrayList<ParameterInfo<?>>();
		
		// SET
		boolean isFirst = true;
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable() && !columnDef.isPrimaryKey()) {
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));
				
				if (isFirst) {
					isFirst = false;
				} else {
					setClause.append(", "); //$NON-NLS-1$
				}
				setClause.append(columnDef.getNameForQuery()).append(" = ?"); //$NON-NLS-1$
				parameterInfoList.add(parameterInfo);
			}
		}
		tokenList.add(new StaticToken(setClause.toString(), ParameterInfo.toArray(parameterInfoList)));
		parameterInfoList.clear();
		
		// WHERE
		isFirst = true;
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable() && columnDef.isPrimaryKey()) {
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));
				
				if (isFirst) {
					isFirst = false;
				} else {
					whereClause.append(" AND "); //$NON-NLS-1$
				}
				whereClause.append(columnDef.getNameForQuery()).append(" = ?"); //$NON-NLS-1$
				parameterInfoList.add(parameterInfo);
			}
		}
		
		tokenList.add(new StaticToken(whereClause.toString(), ParameterInfo.toArray(parameterInfoList)));
		
		BaseQuery query = QueryHelper.optimizedQuery(CommandType.UPDATE, tokenList.toArray(new Token[tokenList.size()]), makeDescription(entityClass, UPDATE_QUERY_ID));
		Cache cache = entityDef.getCache();
		if (cache != null) {
			query.setCacheName(cache.cacheName());
			query.setCacheStrategy(cache.cacheStrategy());
		}
		return query;
	}
	
	
	public Query getUpdateSelectiveQuery(Class<?> entityClass) throws SQLException {
		ConcurrentHashMap<String, Query> cacheMap = QUERY_PROVIDER_CACHE_MAP.get(entityClass);
		if (cacheMap == null) {
			cacheMap = new ConcurrentHashMap<String, Query>();
			ConcurrentHashMap<String, Query> temp = QUERY_PROVIDER_CACHE_MAP.putIfAbsent(entityClass, cacheMap);
			if (temp != null) {
				cacheMap = temp;
			}
		}
		
		Query query = cacheMap.get(UPDATE_SELECTIVE_QUERY_ID);
		if (query == null) {
			query = makeUpdateSelectiveQuery(entityClass);
			Query temp = cacheMap.putIfAbsent(UPDATE_SELECTIVE_QUERY_ID, query);
			if (temp != null) {
				query = temp;
			}
		}
		
		return query;
	}
	
	@SuppressWarnings("unchecked")
	protected Query makeUpdateSelectiveQuery(Class<?> entityClass) throws SQLException {
		EntityDefinition entityDef = entityDefinitionManager.getEntityDefinition(entityClass);
		
		List<Token> tokenList = new ArrayList<Token>();
		
		// update
		tokenList.add(new StaticToken("UPDATE ")); //$NON-NLS-1$
		tokenList.addAll(QueryHelper.newTokenList(entityDef.getTableName().toString(), entityClass));
		
		StringBuilder whereClause = new StringBuilder();
		List<ParameterInfo<?>> whereClauseparameterInfoList = new ArrayList<ParameterInfo<?>>();

		// set
		DynamicToken dynamicToken = new DynamicToken(" SET "); //$NON-NLS-1$
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable()) {
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));			
				
				if (columnDef.isPrimaryKey()) {
					if (whereClause.length() > 0) {
						whereClause.append(" AND "); //$NON-NLS-1$
					}
					whereClause.append(columnDef.getNameForQuery()).append(" = ?"); //$NON-NLS-1$
					whereClauseparameterInfoList.add(parameterInfo);
				} else {
					IfToken child = new IfToken("${not empty " + columnDef.getName() + "}", ", ", columnDef.getNameForQuery() + " = ?", new ParameterInfo[] {parameterInfo}); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
					dynamicToken.add(child);				
				}
			}
		}
		tokenList.add(dynamicToken);

		// where
		if (whereClause.length() > 0) {
			whereClause.insert(0, " WHERE "); //$NON-NLS-1$
			tokenList.add(new StaticToken(whereClause.toString(), ParameterInfo.toArray(whereClauseparameterInfoList)));
		}
		
		
		BaseQuery query = QueryHelper.optimizedQuery(CommandType.UPDATE, tokenList.toArray(new Token[tokenList.size()]), makeDescription(entityClass, UPDATE_SELECTIVE_QUERY_ID));
		Cache cache = entityDef.getCache();
		if (cache != null) {
			query.setCacheName(cache.cacheName());
			query.setCacheStrategy(cache.cacheStrategy());
		}
		return query;
	}
	
	
	public Query getDeleteQuery(Class<?> entityClass) throws SQLException {
		ConcurrentHashMap<String, Query> cacheMap = QUERY_PROVIDER_CACHE_MAP.get(entityClass);
		if (cacheMap == null) {
			cacheMap = new ConcurrentHashMap<String, Query>();
			ConcurrentHashMap<String, Query> temp = QUERY_PROVIDER_CACHE_MAP.putIfAbsent(entityClass, cacheMap);
			if (temp != null) {
				cacheMap = temp;
			}
		}
		
		Query query = cacheMap.get(DELETE_QUERY_ID);
		if (query == null) {
			query = makeDeleteQuery(entityClass);
			Query temp = cacheMap.putIfAbsent(DELETE_QUERY_ID, query);
			if (temp != null) {
				query = temp;
			}
		}
		
		return query;
	}
	
	@SuppressWarnings("unchecked")
	protected Query makeDeleteQuery(Class<?> entityClass) throws SQLException {
		EntityDefinition entityDef = entityDefinitionManager.getEntityDefinition(entityClass);
		
		List<Token> tokenList = new ArrayList<Token>();
		tokenList.add(new StaticToken("DELETE FROM ")); //$NON-NLS-1$
		tokenList.addAll(QueryHelper.newTokenList(entityDef.getTableName().toString(), entityClass));
		
		DynamicToken dynamicToken = new DynamicToken(" WHERE "); //$NON-NLS-1$
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable()) {
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));			
				
				IfToken child = new IfToken("${not empty " + columnDef.getName() + "}", " AND ", columnDef.getNameForQuery() + " = ?", new ParameterInfo[] {parameterInfo}); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				dynamicToken.add(child);	
			}
		}
		tokenList.add(dynamicToken);
		
		BaseQuery query = QueryHelper.optimizedQuery(CommandType.SELECT, tokenList.toArray(new Token[tokenList.size()]), makeDescription(entityClass, DELETE_QUERY_ID));
		Cache cache = entityDef.getCache();
		if (cache != null) {
			query.setCacheName(cache.cacheName());
			query.setCacheStrategy(cache.cacheStrategy());
		}
		return query;
	}
	
	public Query getDeleteByPrimaryKeyQuery(Class<?> entityClass) throws SQLException {
		ConcurrentHashMap<String, Query> cacheMap = QUERY_PROVIDER_CACHE_MAP.get(entityClass);
		if (cacheMap == null) {
			cacheMap = new ConcurrentHashMap<String, Query>();
			ConcurrentHashMap<String, Query> temp = QUERY_PROVIDER_CACHE_MAP.putIfAbsent(entityClass, cacheMap);
			if (temp != null) {
				cacheMap = temp;
			}
		}
		
		Query query = cacheMap.get(DELETE_BY_PRIMARY_KEY_QUERY_ID);
		if (query == null) {
			query = makeDeleteByPrimaryKeyQuery(entityClass);
			Query temp = cacheMap.putIfAbsent(DELETE_BY_PRIMARY_KEY_QUERY_ID, query);
			if (temp != null) {
				query = temp;
			}
		}
		
		return query;
	}
	
	@SuppressWarnings("unchecked")
	protected Query makeDeleteByPrimaryKeyQuery(Class<?> entityClass) throws SQLException {
		EntityDefinition entityDef = entityDefinitionManager.getEntityDefinition(entityClass);
		
		List<Token> tokenList = new ArrayList<Token>();
		
		tokenList.add(new StaticToken("DELETE FROM ")); //$NON-NLS-1$
		tokenList.addAll(QueryHelper.newTokenList(entityDef.getTableName().toString(), entityClass));
		
		StringBuilder whereClause = new StringBuilder();

		List<ParameterInfo<?>> parameterInfoList = new ArrayList<ParameterInfo<?>>();
		boolean isFirst = true;
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable() && columnDef.isPrimaryKey()) {
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));
				
				if (isFirst) {
					isFirst = false;
				} else {
					whereClause.append(" AND "); //$NON-NLS-1$
				}
				whereClause.append(columnDef.getNameForQuery()).append(" = ?"); //$NON-NLS-1$
				parameterInfoList.add(parameterInfo);
			}
		}
		
		if (whereClause.length() > 0) {
			whereClause.insert(0, " WHERE "); //$NON-NLS-1$
		}

		String sql = whereClause.toString();
		tokenList.add(new StaticToken(sql, ParameterInfo.toArray(parameterInfoList)));
		
		
		BaseQuery query = QueryHelper.optimizedQuery(CommandType.DELETE, tokenList.toArray(new Token[tokenList.size()]), makeDescription(entityClass, DELETE_BY_PRIMARY_KEY_QUERY_ID));
		Cache cache = entityDef.getCache();
		if (cache != null) {
			query.setCacheName(cache.cacheName());
			query.setCacheStrategy(cache.cacheStrategy());
		}
		return query;
	}
	
	public Query getGetObjectQuery(Class<?> entityClass) throws SQLException {
		ConcurrentHashMap<String, Query> cacheMap = QUERY_PROVIDER_CACHE_MAP.get(entityClass);
		if (cacheMap == null) {
			cacheMap = new ConcurrentHashMap<String, Query>();
			ConcurrentHashMap<String, Query> temp = QUERY_PROVIDER_CACHE_MAP.putIfAbsent(entityClass, cacheMap);
			if (temp != null) {
				cacheMap = temp;
			}
		}
		
		Query query = cacheMap.get(GET_OBJECT_QUERY_ID);
		if (query == null) {
			query = makeGetObjectQuery(entityClass);
			Query temp = cacheMap.putIfAbsent(GET_OBJECT_QUERY_ID, query);
			if (temp != null) {
				query = temp;
			}
		}
		
		return query;
	}
	
	@SuppressWarnings("unchecked")
	protected Query makeGetObjectQuery(Class<?> entityClass) throws SQLException {
		EntityDefinition entityDef = entityDefinitionManager.getEntityDefinition(entityClass);
		
		List<Token> tokenList = new ArrayList<Token>();
		
		StringBuilder selectClause = new StringBuilder();
		selectClause.append("SELECT "); //$NON-NLS-1$
		List<ParameterInfo<?>> outputparameterInfoList = new ArrayList<ParameterInfo<?>>();
	
		// Column
		boolean isFirst = true;
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable()) {
				if (isFirst) {
					isFirst = false;
				} else {
					selectClause.append(", "); //$NON-NLS-1$
				}
				selectClause.append(columnDef.getNameForQuery());
				
				ParameterInfo parameterInfo = new ParameterInfo(ParameterInfo.MODE.OUT);
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));		
				outputparameterInfoList.add(parameterInfo);
			}
		}
		
		selectClause.append(" FROM "); //$NON-NLS-1$
		tokenList.add(new StaticToken(selectClause.toString(), ParameterInfo.toArray(outputparameterInfoList)));
		tokenList.addAll(QueryHelper.newTokenList(entityDef.getTableName().toString(), entityClass));
		
		DynamicToken dynamicToken = new DynamicToken(" WHERE "); //$NON-NLS-1$
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable()) {
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));			
				
				IfToken child = new IfToken("${not empty " + columnDef.getName() + "}", " AND ", columnDef.getNameForQuery() + " = ?", new ParameterInfo[] {parameterInfo}); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				dynamicToken.add(child);	
			}
		}
		tokenList.add(dynamicToken);
		
		List<HandleRowInterceptor> handleRowInterceptorList = new ArrayList<HandleRowInterceptor>();
		if (entityDef.getJoinColumntCount() > 0) {
			for (int i = 0; i < entityDef.getJoinColumntCount(); i++) {
				JoinColumnDefinition joinColumnDef = entityDef.getJoinColumn(i);
				HandleRowInterceptor interceptor = newHandleRowInterceptorForJoinColumn(joinColumnDef, entityClass);
				if (interceptor != null) {
					handleRowInterceptorList.add(interceptor);
				}
			}
		}
		
		BaseQuery query = QueryHelper.optimizedQuery(CommandType.SELECT, tokenList.toArray(new Token[tokenList.size()]), makeDescription(entityClass, GET_OBJECT_QUERY_ID), null, handleRowInterceptorList.toArray(new HandleRowInterceptor[handleRowInterceptorList.size()]));
		Cache cache = entityDef.getCache();
		if (cache != null) {
			query.setCacheName(cache.cacheName());
			query.setCacheStrategy(cache.cacheStrategy());
		}
		return query;
	}
	
	public Query getGetObjectByPrimaryKeyQuery(Class<?> entityClass) throws SQLException {
		ConcurrentHashMap<String, Query> cacheMap = QUERY_PROVIDER_CACHE_MAP.get(entityClass);
		if (cacheMap == null) {
			cacheMap = new ConcurrentHashMap<String, Query>();
			ConcurrentHashMap<String, Query> temp = QUERY_PROVIDER_CACHE_MAP.putIfAbsent(entityClass, cacheMap);
			if (temp != null) {
				cacheMap = temp;
			}
		}
		
		Query query = cacheMap.get(GET_OBJECT_BY_PRIMARY_KEY_QUERY_ID);
		if (query == null) {
			query = makeGetObjectByPrimaryKeyQuery(entityClass);
			Query temp = cacheMap.putIfAbsent(GET_OBJECT_BY_PRIMARY_KEY_QUERY_ID, query);
			if (temp != null) {
				query = temp;
			}
		}
		
		return query;
	}
	
	@SuppressWarnings("unchecked")
	protected Query makeGetObjectByPrimaryKeyQuery(Class<?> entityClass) throws SQLException {
		EntityDefinition entityDef = entityDefinitionManager.getEntityDefinition(entityClass);
		
		List<Token> tokenList = new ArrayList<Token>();
		
		StringBuilder selectClause = new StringBuilder();
		selectClause.append("SELECT "); //$NON-NLS-1$
		
		List<ParameterInfo<?>> parameterInfoList = new ArrayList<ParameterInfo<?>>();
		
		// Column
		boolean isFirst = true;
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable()) {
				if (isFirst) {
					isFirst = false;
				} else {
					selectClause.append(", "); //$NON-NLS-1$
				}
				selectClause.append(columnDef.getNameForQuery());
				
				ParameterInfo parameterInfo = new ParameterInfo(ParameterInfo.MODE.OUT);
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));
				parameterInfoList.add(parameterInfo);
			}
		}
		
		selectClause.append(" FROM "); //$NON-NLS-1$
		tokenList.add(new StaticToken(selectClause.toString(), ParameterInfo.toArray(parameterInfoList)));
		tokenList.addAll(QueryHelper.newTokenList(entityDef.getTableName().toString(), entityClass));
		parameterInfoList.clear();
		
		StringBuilder whereClause = new StringBuilder();
		whereClause.append(" WHERE "); //$NON-NLS-1$
		
		// WHERE
		isFirst = true;
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable() && columnDef.isPrimaryKey()) {
				if (isFirst) {
					isFirst = false;
				} else {
					whereClause.append(" AND "); //$NON-NLS-1$
				}
				whereClause.append(columnDef.getNameForQuery()).append(" = ?"); //$NON-NLS-1$
				
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));
				parameterInfoList.add(parameterInfo);
			}
		}

		tokenList.add(new StaticToken(whereClause.toString(), ParameterInfo.toArray(parameterInfoList)));
		
		List<HandleRowInterceptor> handleRowInterceptorList = new ArrayList<HandleRowInterceptor>();
		if (entityDef.getJoinColumntCount() > 0) {
			for (int i = 0; i < entityDef.getJoinColumntCount(); i++) {
				JoinColumnDefinition joinColumnDef = entityDef.getJoinColumn(i);
				HandleRowInterceptor interceptor = newHandleRowInterceptorForJoinColumn(joinColumnDef, entityClass);
				if (interceptor != null) {
					handleRowInterceptorList.add(interceptor);
				}
			}
		}
		
		BaseQuery query = QueryHelper.optimizedQuery(CommandType.SELECT, tokenList.toArray(new Token[tokenList.size()]), makeDescription(entityClass, GET_OBJECT_BY_PRIMARY_KEY_QUERY_ID), null, handleRowInterceptorList.toArray(new HandleRowInterceptor[handleRowInterceptorList.size()]));
		Cache cache = entityDef.getCache();
		if (cache != null) {
			query.setCacheName(cache.cacheName());
			query.setCacheStrategy(cache.cacheStrategy());
		}
		return query;
	}
	
	public <N extends Number> Query getGetCountQuery(Class<?> entityClass, Class<N> resultClass) throws SQLException {
		ConcurrentHashMap<String, Query> cacheMap = QUERY_PROVIDER_CACHE_MAP.get(entityClass);
		if (cacheMap == null) {
			cacheMap = new ConcurrentHashMap<String, Query>();
			ConcurrentHashMap<String, Query> temp = QUERY_PROVIDER_CACHE_MAP.putIfAbsent(entityClass, cacheMap);
			if (temp != null) {
				cacheMap = temp;
			}
		}
		
		Query query = cacheMap.get(GET_COUNT_QUERY_ID);
		if (query == null) {
			query = makeGetCountQuery(entityClass, resultClass);
			Query temp = cacheMap.putIfAbsent(GET_COUNT_QUERY_ID, query);
			if (temp != null) {
				query = temp;
			}
		}
		
		return query;
	}
	
	@SuppressWarnings("unchecked")
	protected <N extends Number> Query makeGetCountQuery(Class<?> entityClass, Class<N> resultClass) throws SQLException {
		EntityDefinition entityDef = entityDefinitionManager.getEntityDefinition(entityClass);
		
		List<Token> tokenList = new ArrayList<Token>();

		{
			List<ParameterInfo<?>> parameterInfoList = new ArrayList<ParameterInfo<?>>();
			
			ParameterInfo parameterInfo = new ParameterInfo(ParameterInfo.MODE.OUT);
			parameterInfo.setName("cnt"); //$NON-NLS-1$
			parameterInfo.setColumnName("CNT"); //$NON-NLS-1$
			parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(resultClass));
			parameterInfoList.add(parameterInfo);

			tokenList.add(new StaticToken("SELECT COUNT(*) CNT FROM ", ParameterInfo.toArray(parameterInfoList))); //$NON-NLS-1$
			tokenList.addAll(QueryHelper.newTokenList(entityDef.getTableName().toString(), entityClass));
		}
		
		DynamicToken dynamicToken = new DynamicToken(" WHERE "); //$NON-NLS-1$
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable()) {
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));			
				
				IfToken child = new IfToken("${not empty " + columnDef.getName() + "}", " AND ", columnDef.getNameForQuery() + " = ?", new ParameterInfo[] {parameterInfo}); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				dynamicToken.add(child);	
			}
		}
		tokenList.add(dynamicToken);
		
		BaseQuery query = QueryHelper.optimizedQuery(CommandType.SELECT, tokenList.toArray(new Token[tokenList.size()]), makeDescription(entityClass, GET_COUNT_QUERY_ID));
		Cache cache = entityDef.getCache();
		if (cache != null) {
			query.setCacheName(cache.cacheName());
			query.setCacheStrategy(cache.cacheStrategy());
		}
		return query;
	}


	public Query getGetListQuery(Class<?> entityClass, String orderBy) throws SQLException {
		ConcurrentHashMap<String, Query> cacheMap = QUERY_PROVIDER_CACHE_MAP.get(entityClass);
		if (cacheMap == null) {
			cacheMap = new ConcurrentHashMap<String, Query>();
			ConcurrentHashMap<String, Query> temp = QUERY_PROVIDER_CACHE_MAP.putIfAbsent(entityClass, cacheMap);
			if (temp != null) {
				cacheMap = temp;
			}
		}
		String id = GET_LIST_QUERY_ID + StringUtils.defaultIfNotBlank(orderBy, ".", null); //$NON-NLS-1$
		Query query = cacheMap.get(id);
		if (query == null) {
			query = makeGetListQuery(entityClass, orderBy);
			Query temp = cacheMap.putIfAbsent(id, query);
			if (temp != null) {
				query = temp;
			}
		}
		
		return query;
	}
	
	@SuppressWarnings("unchecked")
	protected Query makeGetListQuery(Class<?> entityClass, String orderBy) throws SQLException {
		EntityDefinition entityDef = entityDefinitionManager.getEntityDefinition(entityClass);
	
		List<Token> tokenList = new ArrayList<Token>();
		
		StringBuilder selectClause = new StringBuilder();
		selectClause.append("SELECT "); //$NON-NLS-1$
		List<ParameterInfo<?>> outputparameterInfoList = new ArrayList<ParameterInfo<?>>();
	
		// Column
		boolean isFirst = true;
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable()) {
				if (isFirst) {
					isFirst = false;
				} else {
					selectClause.append(", "); //$NON-NLS-1$
				}
				selectClause.append(columnDef.getNameForQuery());
				
				ParameterInfo parameterInfo = new ParameterInfo(ParameterInfo.MODE.OUT);
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));		
				outputparameterInfoList.add(parameterInfo);
			}
		}
		selectClause.append(" FROM "); //$NON-NLS-1$
		tokenList.add(new StaticToken(selectClause.toString(), ParameterInfo.toArray(outputparameterInfoList)));
		tokenList.addAll(QueryHelper.newTokenList(entityDef.getTableName().toString(), entityClass));
		
		DynamicToken dynamicToken = new DynamicToken(" WHERE "); //$NON-NLS-1$
		for (int i = 0; i < entityDef.getFieldCount(); i++) {
			FieldDefinition columnDef = entityDef.getField(i);
			if (columnDef.isQueriable()) {
				ParameterInfo parameterInfo = new ParameterInfo();
				parameterInfo.setName(columnDef.getName());
				parameterInfo.setColumnName(columnDef.getNameForQuery());
				parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(columnDef.getJavaType(), columnDef.getJdbcType()));			
				
				IfToken child = new IfToken("${not empty " + columnDef.getName() + "}", " AND ", columnDef.getNameForQuery() + " = ?", new ParameterInfo[] {parameterInfo}); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				dynamicToken.add(child);	
			}
		}
		tokenList.add(dynamicToken);
		
		if (StringUtils.isNotBlank(orderBy)) {
			tokenList.add(new StaticToken(" ORDER BY " + orderBy)); //$NON-NLS-1$
		}
		
		List<HandleRowInterceptor> handleRowInterceptorList = new ArrayList<HandleRowInterceptor>();
		if (entityDef.getJoinColumntCount() > 0) {
			for (int i = 0; i < entityDef.getJoinColumntCount(); i++) {
				JoinColumnDefinition joinColumnDef = entityDef.getJoinColumn(i);
				HandleRowInterceptor interceptor = newHandleRowInterceptorForJoinColumn(joinColumnDef, entityClass);
				if (interceptor != null) {
					handleRowInterceptorList.add(interceptor);
				}
			}
		}
		
		BaseQuery query = QueryHelper.optimizedQuery(CommandType.SELECT, tokenList.toArray(new Token[tokenList.size()]), makeDescription(entityClass, GET_LIST_QUERY_ID), null, handleRowInterceptorList.toArray(new HandleRowInterceptor[handleRowInterceptorList.size()]));
		Cache cache = entityDef.getCache();
		if (cache != null) {
			query.setCacheName(cache.cacheName());
			query.setCacheStrategy(cache.cacheStrategy());
		}
		return query;
	}
	
	protected HandleRowInterceptor newHandleRowInterceptorForJoinColumn(
			JoinColumnDefinition joinDef, Class<?> entityClass) throws SQLException {
		HandleRowInterceptor interceptor = null;
		if (joinDef.getTargetEntity() != null && joinDef.getKeys() != null && joinDef.getKeys().length > 0) {
			final String name = joinDef.getName();
			final Class<?> type = joinDef.getType();
			final Class<?> targetEntity = joinDef.getTargetEntity();
			final String targetProperty = joinDef.getTargetProperty();
			final String[] keys = joinDef.getKeys();
			final boolean useCache = joinDef.isUseCache();
			if (joinDef.isMany()) {
				interceptor = new HandleRowInterceptorAdapter() {
					@SuppressWarnings("unchecked")
					@Override
					public void beforeHandleRow(Session session,
							Connection conn, RunnableQuery runnableQuery,
							Object parameterObject, Object resultObject)
							throws SQLException {
						try {
							Object param = BeanUtils.newInstance(targetEntity);
							for (String key : keys) {
								BeanUtils.setProperty(param, key, BeanUtils.getProperty(resultObject, key));
							}
							List<?> list = useCache ? session.getList(new CacheKey(param)) : session.getList(param);
							if (list != null && list.size() > 0) {
								if (type.isArray()) {
									Object array = Array.newInstance(targetEntity, list.size());
									for (int i = 0; i < list.size(); i++) {
										Array.set(array, i, list.get(i));
									}
									BeanUtils.setProperty(resultObject, name, array);
								} else {
									Collection collection = (Collection)BeanUtils.newInstance(type);
									collection.addAll(list);
									BeanUtils.setProperty(resultObject, name, collection);
								}
							}
						} catch (Exception e) {
							throw (SQLException)new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.SET_COLLECTION_PROPERTY_ERROR"), ClassUtils.getClassName(resultObject), name)).initCause(e); //$NON-NLS-1$
						}
					}
				};
			} else {
				if (joinDef.getTargetEntity() == joinDef.getType()) {
					interceptor = new HandleRowInterceptorAdapter() {
						@Override
						public void beforeHandleRow(Session session,
								Connection conn, RunnableQuery runnableQuery,
								Object parameterObject, Object resultObject)
								throws SQLException {
							try {
								Object param = BeanUtils.newInstance(targetEntity);
								for (String key : keys) {
									BeanUtils.setProperty(param, key, BeanUtils.getProperty(resultObject, key));
								}
								Object value = useCache ? session.getObject(new CacheKey(param)) : session.getObject(param);
								BeanUtils.setProperty(resultObject, name, value);
							} catch (Exception e) {
								throw (SQLException)new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.SET_COLLECTION_PROPERTY_ERROR"), ClassUtils.getClassName(resultObject), name)).initCause(e); //$NON-NLS-1$
							}
						}
						
					};
				} else {
					interceptor = new HandleRowInterceptorAdapter() {
						@Override
						public void beforeHandleRow(Session session,
								Connection conn, RunnableQuery runnableQuery,
								Object parameterObject, Object resultObject)
								throws SQLException {
							try {
								Object param = BeanUtils.newInstance(targetEntity);
								for (String key : keys) {
									BeanUtils.setProperty(param, key, BeanUtils.getProperty(resultObject, key));
								}
								Object obj = useCache ? session.getObject(new CacheKey(param)) : session.getObject(param);
								Object value = BeanUtils.getProperty(obj, StringUtils.defaultIfBlank(targetProperty, name));
								BeanUtils.setProperty(resultObject, name, value);
							} catch (Exception e) {
								throw (SQLException)new SQLException(MessageUtils.format(Messages.getString("SimpleQueryGenerator.SET_COLLECTION_PROPERTY_ERROR"), ClassUtils.getClassName(resultObject), name)).initCause(e); //$NON-NLS-1$
							}
						}
					};
				}
			}

		}
		return interceptor;
	}

}
