/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.provider.generator.model;

import java.util.ArrayList;
import java.util.List;

import kr.kangwoo.damo.annotation.Cache;
import kr.kangwoo.util.ObjectUtils;

/**
 * 테이블 정보를 저장하고 있다.
 * 
 * @author badnom
 */
public class EntityDefinition {

	private TableName tableName;
	private List<FieldDefinition> fields;
	private List<FieldDefinition> primaryKeyFields;
	
	private List<FieldDefinition> baseFields;
	private List<FieldDefinition> lobFields;
	
	private Cache cache;
	
	private List<JoinColumnDefinition> joinColumns;
	
	public EntityDefinition(TableName tableName) {
		this.tableName = tableName;
		this.fields = new ArrayList<FieldDefinition>();
		this.primaryKeyFields = new ArrayList<FieldDefinition>();
		
		this.baseFields = new ArrayList<FieldDefinition>();
		this.lobFields = new ArrayList<FieldDefinition>();
		
		this.joinColumns = new ArrayList<JoinColumnDefinition>();
	}
	
	/**
	 * <p>테이블명을 가져온다.</p>
	 * 
	 * @return 테이블명
	 */
	public TableName getTableName() {
		return tableName;
	}

	/**
	 * <p>해당 컬럼을 추가한다.</p>
	 * 
	 * @param field 추가할 컬럼
	 */
	public void addField(FieldDefinition field) {
		fields.add(field);
		if (field.isPrimaryKey()) {
			primaryKeyFields.add(field);
		}
		if (field.isLob()) {
			lobFields.add(field);
		} else {
			baseFields.add(field);
		}
	}
	
	/**
	 * <p>컬럼의 갯수를 가져온다.</p>
	 * 
	 * @return 컬럼의 갯수
	 */
	public int getFieldCount() {
		return fields.size();
	}

	/**
	 * <p>인덱스에 해당하는 컬럼을 가져온다.</p>
	 * 
	 * @param index 인덱스
	 * @return
	 */
	public FieldDefinition getField(int index) {
		return fields.get(index);
	}

	/**
	 * <p>해당 컬럼을 삭제한다.</p>
	 * 
	 * @param field 삭제할 컬럼
	 * @return
	 */
	public boolean removeField(FieldDefinition field) {
		return fields.remove(field);
	}
	
	public Cache getCache() {
		return cache;
	}

	public void setCache(Cache cache) {
		this.cache = cache;
	}

	/**
	 * <p>컬럼의 갯수를 가져온다.</p>
	 * 
	 * @return 컬럼의 갯수
	 */
	public int getJoinColumntCount() {
		return joinColumns.size();
	}
	
	/**
	 * <p>해당 컬럼을 추가한다.</p>
	 * 
	 * @param column 추가할 컬럼
	 */
	public void addJoinColumn(JoinColumnDefinition column) {
		joinColumns.add(column);
	}
	
	/**
	 * <p>인덱스에 해당하는 컬럼을 가져온다.</p>
	 * 
	 * @param index 인덱스
	 * @return
	 */
	public JoinColumnDefinition getJoinColumn(int index) {
		return (JoinColumnDefinition)joinColumns.get(index);
	}

	/**
	 * <p>해당 컬럼을 삭제한다.</p>
	 * 
	 * @param column 삭제할 컬럼
	 * @return
	 */
	public boolean removeJoinColumn(JoinColumnDefinition column) {
		return joinColumns.remove(column);
	}

	@Override
	public String toString() {
		return ObjectUtils.reflectionToString(this);
	}

}
