/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.provider.generator.model;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;

import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.annotation.Cache;
import kr.kangwoo.damo.annotation.Column;
import kr.kangwoo.damo.annotation.JoinColumn;
import kr.kangwoo.damo.annotation.PrimaryKey;
import kr.kangwoo.damo.annotation.SelectKey;
import kr.kangwoo.damo.annotation.Table;
import kr.kangwoo.util.ClassUtils;
import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.StringUtils;
/**
 * 엔티티 클래스를 분석하여 테이블 정보와, 컬럼 정보를 생성해낸다.
 * 
 * @author badnom
 *
 */
public class EntityDefinitionProvider {

	public EntityDefinitionProvider() {
	}
	
	
	/**
	 * <p>대상 클래스에 선언한 필드들을 가져온다. 
	 * 테이블 클래스가 <code>null</code>일경우에는 대상 클래스의 필드들만 가져오고, 
	 * <code>null</code>이 아닐 경우는 대상 클래스의 부모 클래스를 검색한후 테이블 클래스까지의 모든 필드들을 가져온다.</p>
	 * 
	 * @param clazz 대상 클래스
	 * @param tableClass 테이블 정보가 선언된 클래스
	 * @return
	 */
	private Field[] getDeclaredFields(Class<?> clazz, Class<?> tableClass) {
		if (tableClass == null) {
			return clazz.getDeclaredFields();
		}
		
		List<Field> fieldList = new ArrayList<Field>();
		Class<?> target = clazz;
		
		do {
			Field[] fields = target.getDeclaredFields();
			if (fields.length > 0) {
				for (Field field : fields) {
					fieldList.add(field);
				}
			}
		} while((target != tableClass) && ((target = target.getSuperclass()) != null));
		return fieldList.toArray(new Field[fieldList.size()]);
	}
	
	private ConcurrentHashMap<Class<?>, EntityDefinition> entityDefinitionMap = new ConcurrentHashMap<Class<?>, EntityDefinition>();
	
	public EntityDefinition getEntityDefinition(Class<?> clazz) throws SQLException {
		EntityDefinition entityDef = 	entityDefinitionMap.get(clazz);
		if (entityDef == null) {
			entityDef = newEntityDefinition(clazz);
			EntityDefinition temp = entityDefinitionMap.putIfAbsent(clazz, entityDef);
			if (temp != null) {
				entityDef = temp;
			}
		}
		return entityDef;
	}
	
	protected EntityDefinition newEntityDefinition(Class<?> clazz) throws SQLException {
		Table table = clazz.getAnnotation(Table.class);
		Class<?> tableClass = null;
		if (table == null) {
			tableClass = clazz;
			while ((table == null) && (tableClass = tableClass.getSuperclass()) != null) {
				table = tableClass.getAnnotation(Table.class);
			}
		}
		
		if (table == null) {
			throw new SQLException(MessageUtils.format(Messages.getString("EntityDefinitionProvider.TABLE_NOT_DEFINE"), clazz)); //$NON-NLS-1$
		}
		
		String schema = table.schema();
		String tableName = table.tableName();
		boolean useActualColumnNames = table.useActualColumnNames();
		
		EntityDefinition entityDef = new EntityDefinition(new TableName(null, schema, tableName));
		
		Field[] fields = getDeclaredFields(clazz, tableClass);

		for (Field field : fields) {
			if (!Modifier.isStatic(field.getModifiers())) {
				JoinColumn joinColumn = field.getAnnotation(JoinColumn.class);
				Class<?> type = field.getType();
				if (joinColumn != null) {
					boolean isMany = false;
					String name = field.getName();
					Class<?> targetEntity = joinColumn.targetEntity();
					String targetProperty = joinColumn.targetProperty();
					boolean useCache = joinColumn.useCache();
					if (ClassUtils.isImplementsInterface(type, java.util.Collection.class) || type.isArray()) {
						isMany = true;
						// Collection, Array일 경우
						if (targetEntity == JoinColumn.UnknownEntityTarget.class) {
							targetEntity = null;
							if (ClassUtils.isImplementsInterface(type, java.util.Collection.class)) {
								if (field.getGenericType() instanceof ParameterizedType) {
									if (Modifier.isInterface(type.getModifiers()) ||  Modifier.isAbstract(type.getModifiers())) {
										// 추상일경우 대체한다.
										if (ClassUtils.isImplementsInterface(type, java.util.List.class)) {
											type = ArrayList.class;
										} else if (ClassUtils.isImplementsInterface(type, java.util.SortedSet.class)) { 
											type = TreeSet.class;
										} else if (ClassUtils.isImplementsInterface(type, java.util.Set.class)) {
											type = LinkedHashSet.class;
										}
									}
									ParameterizedType parameterizedType = (ParameterizedType)field.getGenericType();
									targetEntity = (Class<?>)parameterizedType.getActualTypeArguments()[0];
								}
							} else if(type.isArray()) {
								targetEntity = type.getComponentType();
							}
						}
					} else {
						isMany = false;
						// 필드타입 자체가 타켓엔티티가 된다.
						if (targetEntity == JoinColumn.UnknownEntityTarget.class) {
							targetEntity = type;
						}
					}
					
					if (targetEntity != null && targetEntity != JoinColumn.UnknownEntityTarget.class && StringUtils.isNotBlank(joinColumn.keys())) {
						JoinColumnDefinition joinColumnDef = new JoinColumnDefinition();
						joinColumnDef.setName(name);
						joinColumnDef.setType(type);
						joinColumnDef.setTargetEntity(targetEntity);
						joinColumnDef.setTargetProperty(targetProperty);
						joinColumnDef.setKeys(joinColumn.keys());
						joinColumnDef.setUseCache(useCache);
						joinColumnDef.setMany(isMany);
						entityDef.addJoinColumn(joinColumnDef);
					}
				} else {
					if (ClassUtils.isImplementsInterface(type, java.util.Collection.class) || type.isArray()) {
						// JoinColumn이 정의되지 않은 Collection, Array 일 경우는 처리를 하지 않는다.		
					} else {
						Column column = field.getAnnotation(Column.class);
						if (column == null || column.isQueriable()) {
							FieldDefinition columnDef = new FieldDefinition();
							columnDef.setName(field.getName());
							columnDef.setNameForQuery(getColumnNameForQuery(field, column, useActualColumnNames));
							columnDef.setPrimaryKey(field.getAnnotation(PrimaryKey.class) != null);
							columnDef.setJavaType(field.getType());
							if (column != null) {
								if (StringUtils.isNotBlank(column.jdbcType())) {
									columnDef.setJdbcType(column.jdbcType());	
								}
							}
							if (field.getAnnotation(SelectKey.class) != null) {
								SelectKey selectKey = field.getAnnotation(SelectKey.class);
								columnDef.setSelectKey(selectKey);
							}
							entityDef.addField(columnDef);
						}
					}
				}
			}
		}
		
		// Cache
		Cache cache = clazz.getAnnotation(Cache.class);
		if (cache != null) {
			entityDef.setCache(cache);
		}
		return entityDef;
	}
	


	public String getColumnNameForQuery(Field field, Column column, boolean useActualColumnNames) {
		if (column == null || StringUtils.isBlank(column.name())) {
			if (useActualColumnNames) {
				String name = field.getName();
				StringBuilder sb = new StringBuilder(name.length());
				for (int i = 0; i < name.length(); i++) {
					char ch = name.charAt(i);
					if (Character.isUpperCase(ch)) {
						sb.append('_');
					}
					sb.append(Character.toUpperCase(ch));
				}
				return sb.toString();
			} else {
				return field.getName();	
			}
		} else {
			return column.name();	
		}
	}
}
