/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.provider.loader.file;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.engine.cache.CacheStrategy;
import kr.kangwoo.damo.engine.query.BaseQuery;
import kr.kangwoo.damo.engine.query.CommandType;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.damo.engine.query.QueryHelper;
import kr.kangwoo.damo.engine.query.parameter.type.TypeHandlerFactory;
import kr.kangwoo.damo.engine.query.token.ChooseToken;
import kr.kangwoo.damo.engine.query.token.DynamicToken;
import kr.kangwoo.damo.engine.query.token.ForEachToken;
import kr.kangwoo.damo.engine.query.token.IfToken;
import kr.kangwoo.damo.engine.query.token.OtherwiseToken;
import kr.kangwoo.damo.engine.query.token.SuperToken;
import kr.kangwoo.damo.engine.query.token.Token;
import kr.kangwoo.damo.engine.query.token.WhenToken;
import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.StringUtils;
import kr.kangwoo.util.xml.sax.helper.DefaultHandler;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class QueryMapParser extends DefaultHandler {
	
	private TypeHandlerFactory typeHandlerFactory = TypeHandlerFactory.getInstance();
	
	private Map<String, Query> queryMap = new HashMap<String, Query>();
	
	private String namespace;
	private String queryId;
	private Class<?> parameterClass;
	private String description;
	private String cacheName;
	private CacheStrategy cacheStrategy;
	private List<Token> tokenList = new ArrayList<Token>();

	
	private Stack<Token> tokenStack = new Stack<Token>();
	
	private Map<String, List<Token>> unknownQueryMap = new HashMap<String, List<Token>>();
	
	@Override
	public Object getResult() {
		return queryMap;
	}
	
	public InputSource resolveEntity(String publicId, String systemId)
			throws IOException, SAXException {
		String dtdURL = null;
		if (publicId != null) {
			dtdURL = dtds.get(publicId);
		}
		if (dtdURL == null) {
			return new InputSource(QueryMapParser.class.getResourceAsStream("queryMap.dtd")); //$NON-NLS-1$
		} else {
			try {
				URL url = new URL(dtdURL);
				InputStream stream = url.openStream();
				return new InputSource(stream);
			} catch (Exception e) {
				throw new SAXException(e);
			}
		}
	}


	@Override
	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
		String before = bodyText.toString();
	    super.startElement(uri, localName, qName, attributes);
	    
	    if (match.equals("queryMap")) { //$NON-NLS-1$
	    	namespace = StringUtils.defaultIfBlank(attributes.getValue("namespace")); //$NON-NLS-1$
	    } else if (match.equals("queryMap/query") || match.equals("queryMap/select") || match.equals("queryMap/insert") || match.equals("queryMap/update") || match.equals("queryMap/delete")) { //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	    	tokenList.clear();
	    	queryId = attributes.getValue("id"); //$NON-NLS-1$
	    	if (StringUtils.isBlank(queryId)) {
	    		throw new SAXException(MessageUtils.format(Messages.getString("QueryMapParser.INPUT_QUERY_ID"), locator.getLineNumber())); //$NON-NLS-1$
	    	}
	    	queryId = StringUtils.toDelimitedStringIfNotEmpty(new String[] {namespace, queryId}, "."); //$NON-NLS-1$
	    	if (queryMap.containsKey(queryId)) {
	    		throw new SAXException(MessageUtils.format(Messages.getString("QueryMapParser.DUPLICATED_QUERY_ID"), queryId, locator.getLineNumber())); //$NON-NLS-1$
	    	}
	    	String parameterClassName = attributes.getValue("parameterClass"); //$NON-NLS-1$
	    	if (StringUtils.isNotBlank(parameterClassName)) {
		    	try {
					parameterClass = Class.forName(typeHandlerFactory.resolveTypeAlias(parameterClassName));
				} catch (ClassNotFoundException e) {
					throw new SAXException(MessageUtils.format(Messages.getString("QueryMapParser.INVAILED_PARAMETER_CLASS"), parameterClassName)); //$NON-NLS-1$
				}
	    	}
	    	description = attributes.getValue("description"); //$NON-NLS-1$
	    	if (StringUtils.isBlank(description)) {
	    		description = queryId;
	    	}
	    	cacheName = attributes.getValue("cacheName"); //$NON-NLS-1$
	    	if (StringUtils.isNotBlank(cacheName)) {
	    		if ("ReadWrite".equals(attributes.getValue("cacheStrategy"))) { //$NON-NLS-1$ //$NON-NLS-2$
	    			cacheStrategy = CacheStrategy.ReadWrite;
	    		} else {
	    			cacheStrategy = CacheStrategy.ReadOnly;
	    		}
	    	}
	    } else if (match.startsWith("queryMap/query/") || match.startsWith("queryMap/select/") || match.startsWith("queryMap/insert/") || match.startsWith("queryMap/update/") || match.startsWith("queryMap/delete/")) { //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	    	if (StringUtils.isNotBlank(before)) {
	    		List<Token> list = QueryHelper.newTokenList(before, parameterClass);
	    		if (list != null) {
		    		if (tokenStack.isEmpty()) {
		    			tokenList.addAll(list);	
		    		} else {
		    			Token parent = tokenStack.peek();
		    			if (parent instanceof SuperToken) {
		    				((SuperToken)parent).addAll(list);
		    			} else if (parent instanceof ChooseToken) {
		    				//
		    				logger.warn(MessageUtils.format(Messages.getString("QueryMapParser.NOT_PROCESSED"), list)); //$NON-NLS-1$
		    			} else {
		    				logger.warn(MessageUtils.format(Messages.getString("QueryMapParser.NOT_PROCESSED"), list)); //$NON-NLS-1$
		    			}
		    		}
	    		}
	    	}
	    	
	    	if (match.endsWith("/dynamic")) { //$NON-NLS-1$
	    		String prepend = attributes.getValue("prepend"); //$NON-NLS-1$
	    		Token token = new DynamicToken(prepend);
	    		tokenStack.push(token);
	    	} else if (match.endsWith("/if")) { //$NON-NLS-1$
	    		String test = attributes.getValue("test"); //$NON-NLS-1$
	    		String prepend = attributes.getValue("prepend"); //$NON-NLS-1$
	    		Token token = new IfToken(test, prepend);
	    		tokenStack.push(token);
	    	} else if (match.endsWith("/choose")) { //$NON-NLS-1$
	    		String prepend = attributes.getValue("prepend"); //$NON-NLS-1$
	    		Token token = new ChooseToken(prepend);
	    		tokenStack.push(token);
	    	} else if (match.endsWith("/choose/when")) { //$NON-NLS-1$
	    		String test = attributes.getValue("test"); //$NON-NLS-1$
	    		String prepend = attributes.getValue("prepend"); //$NON-NLS-1$
	    		Token token = new WhenToken(test, prepend);
	    		tokenStack.push(token);
	    	} else if (match.endsWith("/choose/otherwise")) { //$NON-NLS-1$
	    		String prepend = attributes.getValue("prepend"); //$NON-NLS-1$
	    		Token token = new OtherwiseToken(prepend);
	    		tokenStack.push(token);
	    	} else if (match.endsWith("/forEach")) { //$NON-NLS-1$
	    		String var = attributes.getValue("var"); //$NON-NLS-1$
	    		String items = attributes.getValue("items"); //$NON-NLS-1$
	    		String prepend = attributes.getValue("prepend"); //$NON-NLS-1$
	    		String open = attributes.getValue("open"); //$NON-NLS-1$
	    		String close = attributes.getValue("close"); //$NON-NLS-1$
	    		String conjunction = attributes.getValue("conjunction"); //$NON-NLS-1$
	    		Token token = new ForEachToken(var, items, prepend, open, close, conjunction);
	    		tokenStack.push(token);
	    	} else if (match.endsWith("/include")) { //$NON-NLS-1$
	    		String id = attributes.getValue("id"); //$NON-NLS-1$
	    		List<Token> includeTokenList = unknownQueryMap.get(id);
	    		if (includeTokenList == null) {
	    			includeTokenList = unknownQueryMap.get(namespace + id);
	    		}
	    		if (includeTokenList == null) {
	    			throw new SAXException(MessageUtils.format(Messages.getString("QueryMapParser.QUERY_NOT_EXIST_BY_ID"), id)); //$NON-NLS-1$
	    		}
	    		if (tokenStack.isEmpty()) {
	    			tokenList.addAll(includeTokenList);	
	    		} else {
	    			Token parent = tokenStack.peek();
	    			if (parent instanceof SuperToken) {
	    				((SuperToken)parent).addAll(includeTokenList);
	    			} else {
	    				throw new SAXException(MessageUtils.format(Messages.getString("QueryMapParser.UKNOWN_TEXT"), locator.getLineNumber())); //$NON-NLS-1$
	    			}
	    		}
	    	}
	    }
	}
	
	@Override
	public void endElement(String uri, String localName, String qName) throws SAXException {
	    String text = bodyText.toString();
	    if (match.equals("queryMap/query") || match.equals("queryMap/select") || match.equals("queryMap/insert") || match.equals("queryMap/update") || match.equals("queryMap/delete")) { //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	    	CommandType type = null;
	    	if (match.equals("queryMap/select")) { //$NON-NLS-1$
	    		type = CommandType.SELECT;
	    	} else if (match.equals("queryMap/insert")) { //$NON-NLS-1$
	    		type = CommandType.INSERT;
			} else if (match.equals("queryMap/update")) { //$NON-NLS-1$
				type = CommandType.UPDATE;
			} else if (match.equals("queryMap/delete")) { //$NON-NLS-1$
				type = CommandType.DELETE;
			} else if (match.equals("queryMap/query")) { //$NON-NLS-1$
				type = CommandType.UNKNOWN;
			}

	    	if (StringUtils.isNotBlank(text)) {
	    		List<Token> list = QueryHelper.newTokenList(text, parameterClass);
	    		if (list != null) {
	    			tokenList.addAll(list);
	    		}
	    	}

	    	BaseQuery query = QueryHelper.optimizedQuery(type, tokenList.toArray(new Token[tokenList.size()]), description);
	    	if (StringUtils.isNotBlank(cacheName)) {
	    		query.setCacheName(cacheName);
	    		query.setCacheStrategy(cacheStrategy);
	    	}
	    	queryMap.put(queryId, query);
	    	
	    	if (type == CommandType.UNKNOWN) {
	    		List<Token> newTokenList = new ArrayList<Token>(tokenList);
	    		unknownQueryMap.put(queryId, newTokenList);
	    	}
	    	
	    	queryId = null;
	    	parameterClass = null;
	    	description = null;
	    	cacheName = null;
	    	cacheStrategy = null;
	    	tokenList.clear();
	    } else if (match.startsWith("queryMap/query/") || match.startsWith("queryMap/select/") || match.startsWith("queryMap/insert/") || match.startsWith("queryMap/update/") || match.startsWith("queryMap/delete/")) { //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	    	Token token = null;
	    	if (match.endsWith("/dynamic")) { //$NON-NLS-1$
	    		token = tokenStack.pop();
	    	} else if (match.endsWith("/if")) { //$NON-NLS-1$
	    		token = tokenStack.pop();
	    		Token body = QueryHelper.newToken(text, parameterClass);
	    		if (body != null) {
	    			((IfToken)token).add(body);
	    		}
	    	} else if (match.endsWith("/choose")) { //$NON-NLS-1$
	    		token = tokenStack.pop();
	    	} else if (match.endsWith("/choose/when")) { //$NON-NLS-1$
	    		token = tokenStack.pop();
	    		Token body = QueryHelper.newToken(text, parameterClass);
	    		if (body != null) {
		    		((WhenToken)token).add(body);
	    		}
	    	} else if (match.endsWith("/choose/otherwise")) { //$NON-NLS-1$
	    		token = tokenStack.pop();
	    		Token body = QueryHelper.newToken(text, parameterClass);
	    		if (body != null) {
		    		((OtherwiseToken)token).add(body);
	    		}
	    	} else if (match.endsWith("/forEach")) { //$NON-NLS-1$
	    		token = tokenStack.pop();
	    		Token body = QueryHelper.newToken(text, parameterClass);
	    		if (body != null) {
		    		((ForEachToken)token).add(body);
	    		}
	    	}
	    	
	    	if (token != null) {
	    		if (tokenStack.isEmpty()) {
	    			tokenList.add(token);	
	    		} else {
	    			Token parent = tokenStack.peek();
	    			if (parent instanceof SuperToken) {
	    				((SuperToken)parent).add(token);
	    			} else if (parent instanceof ChooseToken) {
	    				((ChooseToken)parent).addWhenToken((WhenToken)token);
	    			} else {
	    				throw new SAXException(MessageUtils.format(Messages.getString("QueryMapParser.NOT_SUPPORTED_TOKEN"), token, locator.getLineNumber())); //$NON-NLS-1$
	    			}
	    		}
	    	} else {
	    		logger.warn(MessageUtils.format(Messages.getString("QueryMapParser.NOT_PROCESSED"), match)); //$NON-NLS-1$
	    	}

	    }
	    
	    super.endElement(uri, localName, qName);
	}

}
