/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.query;

import java.util.ArrayList;
import java.util.List;

import kr.kangwoo.damo.DamoException;
import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.bundled.org.apache.commons.el.NamedValue;
import kr.kangwoo.damo.engine.executor.interceptor.ExecuteInterceptor;
import kr.kangwoo.damo.engine.executor.interceptor.HandleRowInterceptor;
import kr.kangwoo.damo.engine.query.parameter.ParameterInfo;
import kr.kangwoo.damo.engine.query.parameter.ParameterInfo.MODE;
import kr.kangwoo.damo.engine.query.parameter.type.TypeHandlerFactory;
import kr.kangwoo.damo.engine.query.token.DynamicToken;
import kr.kangwoo.damo.engine.query.token.ReplaceToken;
import kr.kangwoo.damo.engine.query.token.StaticToken;
import kr.kangwoo.damo.engine.query.token.Token;
import kr.kangwoo.damo.external.el.ExpressionEvaluator;
import kr.kangwoo.util.BeanUtils;
import kr.kangwoo.util.ClassUtils;
import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.bean.Invoker;

public class QueryHelper {

	private static TypeHandlerFactory typeHandlerFactory = TypeHandlerFactory.getInstance();

	public static BaseQuery optimizedQuery(CommandType queryType, Token[] tokens) {
		return optimizedQuery(queryType, tokens, null, null, null);
	}
	
	public static BaseQuery optimizedQuery(CommandType queryType, Token[] tokens, String description) {
		return optimizedQuery(queryType, tokens, description, null, null);
	}
	
	public static BaseQuery optimizedQuery(CommandType queryType, Token[] tokens, String description, ExecuteInterceptor[] executeInterceptors, HandleRowInterceptor[] handleRowInterceptors) {
		List<Token> newTokenList = new ArrayList<Token>();
		boolean isStaticQuery = true;
		Token prevToken = null;
		for (Token token : tokens) {
			if (token instanceof StaticToken) {
				if (prevToken == null) {
					prevToken = token;
				} else if (prevToken instanceof StaticToken) {
					StaticToken prevQuery = prevToken.getStaticToken(null);
					StaticToken currQuery = token.getStaticToken(null);
					String sql = prevQuery.getSql() + currQuery.getSql();
					ParameterInfo<?>[] parameterInfos = ParameterInfo.merge(prevQuery.getParameterInfos(), currQuery.getParameterInfos());
					prevToken = new StaticToken(sql, parameterInfos);
				}
			} else {
				isStaticQuery = false;
				if (prevToken != null) {
					newTokenList.add(prevToken);
					prevToken = null;
				}
				newTokenList.add(token);
			}
		}
		
		if (prevToken != null) {
			newTokenList.add(prevToken);
		}
		
		if (isStaticQuery && newTokenList.size() == 1) {
			StaticToken staticToken = newTokenList.get(0).getStaticToken(null);
			return new StaticQuery(queryType, staticToken.getSql(), staticToken.getParameterInfos(), description, executeInterceptors, handleRowInterceptors);
		} else {
			return new DynamicQuery(queryType, newTokenList.toArray(new Token[newTokenList.size()]), description, executeInterceptors, handleRowInterceptors);
		}
	}
	
	@SuppressWarnings("unchecked")
	public static ParameterInfo getParameterInfo(String propertyName,
			Class<?> entityClass) throws DamoException {
		Object expression = ExpressionEvaluator.parseExpressionString(propertyName);
		ParameterInfo parameterInfo = new ParameterInfo(MODE.IN);
		parameterInfo.setName(propertyName);
		
		Class<?> propertyType = null;
		if (expression instanceof NamedValue) {
			String name = ((NamedValue)expression).getName();
			parameterInfo.setColumnName(name);	
			if (entityClass != null && !ClassUtils.isImplementsInterface(entityClass, java.util.Map.class)) {
				Invoker invoker = BeanUtils.getGetInvoker(entityClass, name);
				if (invoker != null) {
					propertyType = invoker.getType();
				} else {
					throw new DamoException(MessageUtils.format(Messages.getString("QueryHelper.GET_PROPERTY_ERROR"), entityClass, propertyName));	 //$NON-NLS-1$
				}
			}
		}

		if (propertyType == null) {
			parameterInfo.setTypeHandler(typeHandlerFactory.getUnknownTypeHandler());
		} else {
			parameterInfo.setTypeHandler(typeHandlerFactory.getTypeHandler(propertyType));
		}
		return parameterInfo;
	}
	
	
	public static List<Token> newTokenList(String sql, Class<?> entityClass) {
		List<Token> tokenList = new ArrayList<Token>();
		boolean inProperty = false;
		char mark = '.';
		int length = sql.length();
		int beginIndex = 0;
		for (int index = 0; index < length; index++) {
			char ch = sql.charAt(index);
			if (inProperty) {
				if (ch == '}') {
					if (mark == '$') {
						String tmp = sql.substring(beginIndex, index + 1);
						ParameterInfo<?> parameterInfo = getParameterInfo(tmp, entityClass);
						tokenList.add(new StaticToken("?", new ParameterInfo[] {parameterInfo})); //$NON-NLS-1$
						beginIndex = index + 1;
						inProperty = false;
					} else if (mark == '@') {
						tokenList.add(new ReplaceToken(sql.substring(beginIndex, index + 1)));
						beginIndex = index + 1;
						inProperty = false;
					}
				}
			} else {
				if (ch == '@' && (index + 1) < length) {
					if (sql.charAt(index + 1) == '{') {
						mark = '@';
						tokenList.add(new StaticToken(sql.substring(beginIndex, index)));
						beginIndex = index;
						inProperty = true;
						index++;
					}
				} else if (ch == '$' && (index + 1) < length) {
					if (sql.charAt(index + 1) == '{') {
						mark = '$';
						tokenList.add(new StaticToken(sql.substring(beginIndex, index)));
						beginIndex = index;
						inProperty = true;
						index++;
					}
				}
			}
		}
		if (beginIndex < length) {
			tokenList.add(new StaticToken(sql.substring(beginIndex, length)));
		}
		return tokenList;
	}
	
	public static Token newToken(String sql, Class<?> entityClass) {
		return newToken(newTokenList(sql, entityClass));
	}
	
	public static Token newToken(List<Token> tokenList) {
		if (tokenList == null || tokenList.size() == 0) {
			return null;
		} else if (tokenList.size() == 1) {
			return tokenList.get(0);
		} else {
			return new DynamicToken(tokenList);
		}
	}
	
	/**
	 * <p>입력 받은 SQL로 쿼리 객체를 생성한다.</p>
	 * 
	 * @param commandType 명령어 형태
	 * @param sql
	 * @param entityClass 엔티티 클래스
	 * @return
	 */
	public static Query newQuery(CommandType commandType, String sql, Class<?> entityClass) {
		List<Token> tokenList = newTokenList(sql, entityClass);
		return optimizedQuery(commandType, tokenList.toArray(new Token[tokenList.size()]));
	}
	
	public static Query newInsertQuery(String sql) {
		return newInsertQuery(sql, null);
	}
	
	public static Query newInsertQuery(String sql, Class<?> entityClass) {
		return newQuery(CommandType.INSERT, sql, entityClass);
	}
	
	public static Query newUpdateQuery(String sql) {
		return newUpdateQuery(sql, null);
	}
	
	public static Query newUpdateQuery(String sql, Class<?> entityClass) {
		return newQuery(CommandType.UPDATE, sql, entityClass);
	}
	
	public static Query newDeleteQuery(String sql) {
		return newDeleteQuery(sql, null);
	}
	
	public static Query newDeleteQuery(String sql, Class<?> entityClass) {
		return newQuery(CommandType.DELETE, sql, entityClass);
	}
	
	public static Query newSelectQuery(String sql) {
		return newSelectQuery(sql, null, null, null, null);
	}
	
	public static Query newSelectQuery(String sql, Class<?> entityClass) {
		return newSelectQuery(sql, entityClass, null, null, null);
	}

	public static Query newSelectQuery(String sql, Class<?> entityClass, Integer fetchSize, Integer resultSetType, Integer resultSetConcurrency) {
		Query query = newQuery(CommandType.SELECT, sql, entityClass);
		if (query instanceof BaseQuery) {
			BaseQuery baseQuery = (BaseQuery)query;
			baseQuery.setFetchSize(fetchSize);
			baseQuery.setResultSetType(resultSetType);
			baseQuery.setResultSetConcurrency(resultSetConcurrency);
		}
		return query;
	}
}
