/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.query.parameter;

import java.sql.CallableStatement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

import kr.kangwoo.damo.DamoException;
import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.engine.query.parameter.type.TypeHandler;
import kr.kangwoo.damo.external.el.ExpressionEvaluator;
import kr.kangwoo.util.BeanUtils;
import kr.kangwoo.util.MessageUtils;

public class ParameterInfo<E> {
	
	public static enum MODE {IN, OUT, INOUT};
	
	private MODE mode;
	private String name;
	private String columnName;
	private TypeHandler<E> typeHandler;
	private int precision;
	private int scale;
	
	private boolean isInputParameter;
	private boolean isOutputParameter;
	
	private boolean isExpression;
	private ParameterExposer parameterExposer;
	
	public ParameterInfo() {
		this(MODE.IN);
	}
	
	public ParameterInfo(MODE mode) {
		setMode(mode);
	}

	public MODE getMode() {
		return mode;
	}

	public void setMode(MODE mode) {
		this.mode = mode;
		isInputParameter = ((mode == MODE.IN) || (mode == MODE.INOUT));
		isOutputParameter = ((mode == MODE.OUT) || (mode == MODE.INOUT));
	}

	public boolean isInputParameter() {
		return isInputParameter;
	}

	public boolean isOutputParameter() {
		return isOutputParameter;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
		isExpression = (name != null) && name.startsWith("${") && name.endsWith("}"); //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	public String getColumnName() {
		return columnName;
	}

	public void setColumnName(String columnName) {
		this.columnName = columnName;
	}

	public TypeHandler<E> getTypeHandler() {
		return typeHandler;
	}

	public void setTypeHandler(TypeHandler<E> typeHandler) {
		this.typeHandler = typeHandler;
	}
	
	public int getPrecision() {
		return precision;
	}

	public void setPrecision(int length) {
		this.precision = length;
	}

	public int getScale() {
		return scale;
	}

	public void setScale(int scale) {
		this.scale = scale;
	}

	public ParameterExposer getParameterExposer() {
		return parameterExposer;
	}

	public void setParameterExposer(ParameterExposer parameterExposer) {
		this.parameterExposer = parameterExposer;
	}

	/**
	 * <p>대상 객체에서 파라메터의 값을 가져온다.</p>
	 *
	 * @param target 대상 객체
	 * @return
	 */
	@SuppressWarnings("unchecked")
	public E invoke(Object target) {
		try {
			if (isExpression) {
				if (parameterExposer == null) {
					return (E)ExpressionEvaluator.evaluate("value", name, Object.class, target); //$NON-NLS-1$
				} else {
					return (E)ExpressionEvaluator.evaluate("value", name, Object.class, parameterExposer.expose(target)); //$NON-NLS-1$
				}	
			} else {
				if (parameterExposer == null) {
					return (E)BeanUtils.getProperty(target, name);
				} else {
					return (E)ExpressionEvaluator.evaluate("value", name, Object.class, parameterExposer.expose(target)); //$NON-NLS-1$
				}	
			}
		} catch (Exception e) {
			throw new DamoException(MessageUtils.format(Messages.getString("ParameterInfo.GET_PROPERTY_ERROR"), target, name), e); //$NON-NLS-1$
		}
	}
	
	@SuppressWarnings("unchecked")
	public void bindParaemter(PreparedStatement pstmt, int parameterIndex, Object parameter) throws SQLException {
		try {
			typeHandler.bindParameter(pstmt, parameterIndex, (E)parameter);
		} catch (RuntimeException ex) {
			throw new DamoException(MessageUtils.format(Messages.getString("ParameterInfo.SET_PROPERTY_ERROR"), parameterIndex, parameter), ex); //$NON-NLS-1$
		}
	}
	
	public void registerOutParameter(CallableStatement cstmt, int parameterIndex) throws SQLException {
		typeHandler.registerOutParameter(cstmt, parameterIndex);
	}
	
	public E getResult(ResultSet rs) throws SQLException {
		return typeHandler.getResult(rs, columnName);
	}
	
	public E getResult(CallableStatement cstmt, int parameterIndex) throws SQLException {
		return typeHandler.getResult(cstmt, parameterIndex);
	}

	public static ParameterInfo<?>[] toArray(List<ParameterInfo<?>> list) {
		if (list == null) {
			return null;
		}
		return list.toArray(new ParameterInfo[list.size()]);
	}
	
	public static ParameterInfo<?>[] merge(ParameterInfo<?>[] a, ParameterInfo<?>[] b) {
		if (a == null || a.length == 0) {
			return b;
		}
		if (b == null || b.length == 0) {
			return a;
		}
		ParameterInfo<?>[] c = new ParameterInfo<?>[a.length + b.length];
		System.arraycopy(a, 0, c, 0, a.length);
		System.arraycopy(b, 0, c, a.length, b.length);
		return c;
	}


	@Override
	public ParameterInfo<E> clone() {
		ParameterInfo<E> newParameterInfo = new ParameterInfo<E>();
		newParameterInfo.mode = mode;
		newParameterInfo.name = name;
		newParameterInfo.columnName = columnName;
		newParameterInfo.typeHandler = typeHandler;
		newParameterInfo.precision = precision;
		newParameterInfo.scale = scale;
		newParameterInfo.isInputParameter = isInputParameter;
		newParameterInfo.isOutputParameter = isOutputParameter;
		newParameterInfo.parameterExposer = parameterExposer;
		return newParameterInfo;
	}

}
