/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.query.parameter.type;

import java.math.BigDecimal;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import kr.kangwoo.damo.DamoException;
import kr.kangwoo.damo.Messages;
import kr.kangwoo.util.MessageUtils;

public class TypeHandlerFactory {
	
	private final Map<Class<?>, Map<String, TypeHandler<?>>> typeHandlerMap = new HashMap<Class<?>, Map<String, TypeHandler<?>>>();
	private UnknownTypeHandler unknownTypeHandler;
	
	private final Map<String, String> typeAliasMap = new HashMap<String, String>();
	
	private static final TypeHandlerFactory typeHandlerFactory = new TypeHandlerFactory();
	
	private TypeHandlerFactory() {
		
		unknownTypeHandler = new UnknownTypeHandler(this);
		
		register(Boolean.TYPE, new BooleanTypeHandler());
		register(Boolean.class, new BooleanTypeHandler());
		registerTypeAlias("boolean", Boolean.class); //$NON-NLS-1$
		
		register(Short.TYPE, new ShortTypeHandler());
		register(Short.class, new ShortTypeHandler());
		registerTypeAlias("short", Short.class); //$NON-NLS-1$

		register(Integer.TYPE, new IntegerTypeHandler());
		register(Integer.class, new IntegerTypeHandler());
		registerTypeAlias("int", Integer.class); //$NON-NLS-1$
		registerTypeAlias("integer", Integer.class); //$NON-NLS-1$

		register(Long.TYPE, new LongTypeHandler());
		register(Long.class, new LongTypeHandler());
		registerTypeAlias("long", Long.class); //$NON-NLS-1$

		register(Float.TYPE, new FloatTypeHandler());
		register(Float.class, new FloatTypeHandler());
		registerTypeAlias("float", Float.class); //$NON-NLS-1$

		register(Double.TYPE, new DoubleTypeHandler());
		register(Double.class, new DoubleTypeHandler());
		registerTypeAlias("double", Double.class); //$NON-NLS-1$
		
		register(BigDecimal.class, new BigDecimalTypeHandler());
		registerTypeAlias("BigDecimal", BigDecimal.class); //$NON-NLS-1$
		
		register(String.class, new StringTypeHandler());
		register(String.class, "CLOB", new ClobTypeHandler()); //$NON-NLS-1$
		register(String.class, "LONGVARCHAR", new ClobTypeHandler()); //$NON-NLS-1$
		registerTypeAlias("string", String.class); //$NON-NLS-1$

		register(Date.class, new DateTypeHandler());
		register(Date.class, "DATE", new DateOnlyTypeHandler()); //$NON-NLS-1$
		register(Date.class, "TIME", new TimeOnlyTypeHandler()); //$NON-NLS-1$
		registerTypeAlias("date", Date.class); //$NON-NLS-1$
		
		register(java.sql.Date.class, new SqlDateTypeHandler());
		register(java.sql.Time.class, new SqlTimeTypeHandler());
		register(java.sql.Timestamp.class, new SqlTimestampTypeHandler());
		
		register(byte.class, new ByteTypeHandler());
		register(Byte.class, new ByteTypeHandler());
		
		register(byte[].class, new ByteArrayTypeHandler());
		register(byte[].class, "BLOB", new BlobTypeHandler()); //$NON-NLS-1$
		register(byte[].class, "LONGVARBINARY", new BlobTypeHandler()); //$NON-NLS-1$
	}
	
	public static TypeHandlerFactory getInstance() {
		return typeHandlerFactory;
	}
	
	public void registerTypeAlias(String alias, Class<?> clazz) {
		typeAliasMap.put(alias.toLowerCase(), clazz.getName());
	}
	
	public String resolveTypeAlias(String alias) {
		String typeName;
		typeName = typeAliasMap.get(alias.toLowerCase());
		if (typeName == null) {
			typeName = alias;
		}
		return typeName;
	}


	public UnknownTypeHandler getUnknownTypeHandler() {
		return unknownTypeHandler;
	}

	public <E> void register(Class<E> type, TypeHandler<E> typeHandler) {
		register(type, null, typeHandler);
	}
	
	public <E> void register(Class<E> type, String jdbcType, TypeHandler<E> typeHandler) {
		Map<String, TypeHandler<?>> map = typeHandlerMap.get(type);
		if (map == null) {
			map = new HashMap<String, TypeHandler<?>>();
			typeHandlerMap.put(type, map);
		}
		map.put(jdbcType, typeHandler);
	}
	
	public TypeHandler<?> getTypeHandler(Class<?> type) throws DamoException {
		return getTypeHandler(type, null);
	}
	
	public <E> TypeHandler<?> getTypeHandler(Class<?> type, String jdbcType) throws DamoException {
		TypeHandler<?> typeHandler = null;;
		Map<String, TypeHandler<?>> map = typeHandlerMap.get(type);
		if (map != null) {
			typeHandler = map.get(jdbcType);
		}
		if (typeHandler == null) {
			throw new DamoException(MessageUtils.format(Messages.getString("TypeHandlerFactory.NOT_EXIST_TYPE_HANDLER"), type, jdbcType)); //$NON-NLS-1$
		}
		return typeHandler;
	}

	public boolean hasTypeHandler(Class<?> type) {
		Map<String, TypeHandler<?>> map = typeHandlerMap.get(type);
		return (map != null && map.size() > 0);
	}
	
	/**
	 * <p>메타데이터(ResultSetMetaData) 타입의 핸들러를 반환한다.</p>
	 * 
	 * @param rsmd
	 * @param column
	 * @return
	 * @throws SQLException
	 */
	public TypeHandler<?> getTypeHandler(ResultSetMetaData rsmd, int column) throws SQLException {
		return getTypeHandler(null, rsmd, column);
	}

	/**
	 * <p>결과클래스(ResultClass)와 메타데이터(ResultSetMetaData)의 타입이 일치하면, 메타데이터 타입의 핸들러를 반환하고, 일치하지 않으면 결과클래스 타입의 핸들러를 반환한다.</p>
	 * 
	 * @param resultClass 결과클래스
	 * @param rsmd 메타데이터
	 * @param column 컬럼번호
	 * @return
	 * @throws SQLException
	 */
	public TypeHandler<?> getTypeHandler(Class<?> resultClass,
			ResultSetMetaData rsmd, int column) throws SQLException {
		String jdbcType = null;
		Class<?> javaType = null;
		
		switch (rsmd.getColumnType(column)) {
			case Types.ARRAY:
//				jdbcType = "ARRAY";
				javaType = Object.class;
				break;
	
			case Types.BIGINT:
//				jdbcType = "BIGINT";
				javaType = Long.class;
				break;
	
			case Types.BINARY:
//				jdbcType = "BINARY";
				javaType = byte[].class;
				break;
		
			case Types.BIT:
//				jdbcType = "BIT";
				javaType = Boolean.class;
				break;
	
			case Types.BLOB:
				jdbcType = "BLOB"; //$NON-NLS-1$
				javaType = byte[].class;
				break;
	
			case Types.BOOLEAN:
//				jdbcType = "BOOLEAN";
				javaType = Boolean.class;
				break;
	
			case Types.CHAR:
//				jdbcType = "CHAR";
				javaType = String.class;
				break;
	
			case Types.CLOB:
				jdbcType = "CLOB"; //$NON-NLS-1$
				javaType = String.class;
				break;
	
			case Types.DATALINK:
//				jdbcType = "DATALINK";
				javaType = Object.class;
				break;
	
			case Types.DATE:
//				jdbcType = "DATE";
				javaType = Date.class;
				break;
	
			case Types.DECIMAL:
//				jdbcType = "DECIMAL";
				if (rsmd.getScale(column) != 0 || rsmd.getPrecision(column) > 18) {
					javaType = BigDecimal.class;
				} else if (rsmd.getPrecision(column) > 9) {
					javaType = Long.class;
				} else if (rsmd.getPrecision(column) > 4) {
					javaType = Integer.class;
				} else if (rsmd.getPrecision(column) > 0) {
					javaType = Short.class;
				} else {
					javaType = Long.class;
				}
				break;
	
			case Types.DISTINCT:
//				jdbcType = "DISTINCT";
				javaType = Object.class;
				break;
	
			case Types.DOUBLE:
//				jdbcType = "DOUBLE";
				javaType = Double.class;
				break;
	
			case Types.FLOAT:
//				jdbcType = "FLOAT";
				javaType = Double.class;
				break;
	
			case Types.INTEGER:
//				jdbcType = "INTEGER";
				javaType = Integer.class;
				break;
	
			case Types.JAVA_OBJECT:
//				jdbcType = "JAVA_OBJECT";
				javaType = Object.class;
				break;
	
			case Types.LONGVARBINARY:
				jdbcType = "LONGVARBINARY"; //$NON-NLS-1$
				javaType = byte[].class;
				break;
	
			case Types.LONGVARCHAR:
				jdbcType = "LONGVARCHAR"; //$NON-NLS-1$
				javaType = String.class;
				break;
	
			case Types.NULL:
//				jdbcType = "NULL";
				javaType = Object.class;
				break;
	
			case Types.NUMERIC:
//				jdbcType = "NUMERIC";
				if (rsmd.getScale(column) != 0 || rsmd.getPrecision(column) > 18) {
					javaType = BigDecimal.class;
				} else if (rsmd.getPrecision(column) > 9) {
					javaType = Long.class;
				} else if (rsmd.getPrecision(column) > 4) {
					javaType = Integer.class;
				} else if (rsmd.getPrecision(column) > 0) {
					javaType = Short.class;
				} else {
					javaType = Long.class;
				}
				break;
	
			case Types.OTHER:
//				jdbcType = "OTHER";
				javaType = Object.class;
				break;
	
			case Types.REAL:
//				jdbcType = "REAL";
				javaType = Float.class;
				break;
	
			case Types.REF:
//				jdbcType = "REF";
				javaType = Object.class;
				break;
	
			case Types.SMALLINT:
//				jdbcType = "SMALLINT";
				javaType = Short.class;
				break;
	
			case Types.STRUCT:
//				jdbcType = "STRUCT";
				javaType = Object.class;
				break;
	
			case Types.TIME:
//				jdbcType = "TIME";
				javaType = Date.class;
				break;
	
			case Types.TIMESTAMP:
//				jdbcType = "TIMESTAMP";
				javaType = Date.class;
				break;
	
			case Types.TINYINT:
//				jdbcType = "TINYINT";
				javaType = Byte.class;
				break;
	
			case Types.VARBINARY:
//				jdbcType = "VARBINARY";
				javaType = byte[].class;
				break;
	
			case Types.VARCHAR:
//				jdbcType = "VARCHAR";
				javaType = String.class;
				break;
	
			default:
				if (resultClass != null) {
					javaType = resultClass;	
				} else {
					throw new SQLException(MessageUtils.format(Messages.getString("TypeHandlerFactory.NOT_SUPPORTED_TYPE"), rsmd.getColumnName(column), rsmd.getColumnType(column))); //$NON-NLS-1$
				}
				break;
		}
		
		if (resultClass == null || resultClass == javaType) {
			return getTypeHandler(javaType, jdbcType);
		} else {
			return getTypeHandler(resultClass, null);
		}
	}
}
