/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.query.token;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import kr.kangwoo.damo.DamoException;
import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.engine.query.parameter.ParameterExposer;
import kr.kangwoo.damo.engine.query.parameter.ParameterInfo;
import kr.kangwoo.damo.external.el.ExpressionEvaluator;
import kr.kangwoo.util.ArrayUtils;
import kr.kangwoo.util.BeanUtils;
import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.StringUtils;

public class ForEachToken extends SuperToken {
	
	private String var;
	private String items;
	private String prepend;
	private String open;
	private String close;
	private String conjunction;

	public ForEachToken(String var, String items, String prepend, String open, String close, String conjunction) {
		this.var = var;
		this.items = items;
		this.prepend = StringUtils.defaultIfNull(prepend);
		this.open = StringUtils.defaultIfNull(open);
		this.close = StringUtils.defaultIfNull(close);
		this.conjunction = StringUtils.defaultIfNull(conjunction);
	}
	
	public String getPrepend() {
		return prepend;
	}
	
	public StaticToken getStaticToken(Object parameterObject) {
		Object items = ExpressionEvaluator.evaluate("items", this.items, Object.class, parameterObject); //$NON-NLS-1$
		if (items == null) {
			return null;
		}
		final String itemId = this.items;
		StringBuilder sql = new StringBuilder();
		List<ParameterInfo<?>> parameterInfoList = new ArrayList<ParameterInfo<?>>();
		Object[] array = toArray(items);
		boolean isFirst = true;
		StaticToken staticToken = null;
		for (int i = 0; i < array.length; i++) {
			Map<String, Object> attrMap = new HashMap<String, Object>();
			attrMap.put(var, array[i]);
			attrMap.put("parent", parameterObject); //$NON-NLS-1$
			staticToken = super.getStaticToken(attrMap);
			if (staticToken != null) {
				if (isFirst) {
					sql.append(open);
					isFirst = false;
				} else {
					sql.append(conjunction);
				}
				sql.append(staticToken.getSql());
				if (staticToken.getParameterInfos() != null) {
					final int index = i;
					for (ParameterInfo<?> pm : staticToken.getParameterInfos()) {
						ParameterInfo<?> newPm = pm.clone();
						newPm.setParameterExposer(new ParameterExposer() {
							public Object expose(Object value) {
								Object items = ExpressionEvaluator.evaluate("items", itemId, Object.class, value); //$NON-NLS-1$
								if (items == null) {
									return null;
								}
								ParameterContext<String, Object> context = new ParameterContext<String, Object>(value);
								context.put(var, BeanUtils.getIndexedProperty(items, index));
								context.put("this", value); //$NON-NLS-1$
								return context;
							}
						});
						parameterInfoList.add(newPm);
					}
				}
			}
		}
		if (sql.length() > 0) {
			if (isFirst == false) {
				sql.insert(0, prepend);
				sql.append(close);
			}
			return new StaticToken(sql.toString(), ParameterInfo.toArray(parameterInfoList));
		} else {
			return null;
		}
	}
	class ParameterContext<K,V> implements Map<K, V> {
		private Object parameterObject;
		private Map<K, V> attrMap = new HashMap<K, V>(3);
		
		ParameterContext(Object parameterObject) {
			this.parameterObject = parameterObject;
		}

		public void clear() {
			attrMap.clear();
		}

		public boolean containsKey(Object key) {
			return attrMap.containsKey(key);
		}

		public boolean containsValue(Object value) {
			return attrMap.containsValue(value);
		}

		public Set<Map.Entry<K, V>> entrySet() {
			return attrMap.entrySet();
		}

		@SuppressWarnings("unchecked")
		public V get(Object key) {
			V value = attrMap.get(key);
			if (value == null) {
				try {
					value = (V)BeanUtils.getProperty(parameterObject, (String)key);
				} catch (Exception ex) {
					throw new DamoException(MessageUtils.format(Messages.getString("ForEachToken.GET_PROPERTY_ERROR"), parameterObject, key), ex); //$NON-NLS-1$
				}
			}
			return value;
		}

		public boolean isEmpty() {
			return attrMap.isEmpty();
		}

		public Set<K> keySet() {
			return attrMap.keySet();
		}

		public V put(K key, V value) {
			return attrMap.put(key, value);
		}

		public void putAll(Map<? extends K, ? extends V> t) {
			attrMap.putAll(t);
		}

		public V remove(Object key) {
			return attrMap.remove(key);
		}

		public int size() {
			return attrMap.size();
		}

		public Collection<V> values() {
			return attrMap.values();
		}
	}

	protected Object[] toArray(Object obj) {
		Object[] array = null;
		if (obj instanceof Object[]) {
			array = ((Object[]) obj);
		} else if (obj instanceof boolean[]) {
			array = ArrayUtils.toArray((boolean[]) obj);
		} else if (obj instanceof byte[]) {
			array = ArrayUtils.toArray((byte[]) obj);
		} else if (obj instanceof char[]) {
			array = ArrayUtils.toArray((char[]) obj);
		} else if (obj instanceof short[]) {
			array = ArrayUtils.toArray((short[]) obj);
		} else if (obj instanceof int[]) {
			array = ArrayUtils.toArray((int[]) obj);
		} else if (obj instanceof long[]) {
			array = ArrayUtils.toArray((long[]) obj);
		} else if (obj instanceof float[]) {
			array = ArrayUtils.toArray((float[]) obj);
		} else if (obj instanceof double[]) {
			array = ArrayUtils.toArray((double[]) obj);
		} else if (obj instanceof List<?>) {
			array = ((List<?>) obj).toArray();
		} else if (obj instanceof Collection<?>) {
			array = ((Collection<?>)obj).toArray();
		} else if (obj instanceof String) { 
			array = StringUtils.toArray((String)obj, ","); //$NON-NLS-1$
		} else {
			throw new DamoException(MessageUtils.format(Messages.getString("ForEachToken.NOT_SUPPORTED_TARGET"), obj)); //$NON-NLS-1$
		}
		return array;
	}



	
}
