/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.transaction;

import java.sql.Connection;
import java.sql.SQLException;

/**
 * 트랜잭션 격리 수준(isolation level)은 트랜잭션 동시성 보장과 관련된 용어로서, 
 * 격리 수준이 높을수록 동시에 수행되는 트랜잭션 간의 간섭 정도가 작다는 의미이며, 
 * 트랜잭션이 직렬적(SERIALIZABLE)으로 수행된다는 의미이다. 
 * 반면에, 격리 수준이 낮을수록 트랜잭션 간 간섭 정도가 크며, 
 * 트랜잭션이 병렬적으로 수행되므로 트랜잭션 동시성은 크지만, 
 * 데이터 일관성은 떨어진다는 의미이다. 
 * <pre>
 * public interface Connection {
 *     public static final int  TRANSACTION_NONE             = 0;
 *     public static final int  TRANSACTION_READ_UNCOMMITTED = 1;
 *     public static final int  TRANSACTION_READ_COMMITTED   = 2;
 *     public static final int  TRANSACTION_REPEATABLE_READ  = 4;
 *     public static final int  TRANSACTION_SERIALIZABLE     = 8;
 *
 *     int getTransactionIsolation();
 *     void setTransactionIsolation(int isolationlevelconstant);
 * }
 * </pre>
 * <table>
 * <tr>
 * 	<td>Transaction Level</td>
 * 	<td>Dirty reads</td>
 * 	<td>Non  Repeatable reads</td>
 * 	<td>Phantom reads</td>
 * 	<td>Performance impact</td>
 * </tr>
 * <tr>
 * 	<td>TRANSACTION_NONE</td>
 * 	<td>N/A</td>
 * 	<td>N/A</td>
 * 	<td>N/A</td>
 * 	<td>FASTEST</td>
 * </tr>
 * <tr>
 * 	<td>TRANSACTION_READ_UNCOMMITED</td>
 * 	<td>YES</td>
 * 	<td>YES</td>
 * 	<td>YES</td>
 * 	<td>FASTEST</td>
 * </tr>
 * <tr>
 * 	<td>TRANSACTION_READ_COMMITED</td>
 * 	<td>NO</td>
 * 	<td>YES</td>
 * 	<td>YES</td>
 * 	<td>FAST</td>
 * </tr>
 * <tr>
 * 	<td>TRANSACTION_REPEATABLE_READ</td>
 * 	<td>NO</td>
 * 	<td>NO</td>
 * 	<td>YES</td>
 * 	<td>MEDIUM</td>
 * </tr>
 * <tr>
 * 	<td>TRANSACTION_SERIALIZABLE</td>
 * 	<td>NO</td>
 * 	<td>NO</td>
 * 	<td>NO</td>
 * 	<td>SLOW</td>
 * </tr>
 * </table>
 * 
 * 
 * @author badnom
 *
 */
public class TransactionIsolationLevel {
	
	public static final int UNSET_ISOLATION_LEVEL = Integer.MIN_VALUE;
	
	private int isolationLevel = UNSET_ISOLATION_LEVEL;
	private int originalIsolationLevel = UNSET_ISOLATION_LEVEL;
	
	public TransactionIsolationLevel(int isolationLevel) {
		this.isolationLevel = isolationLevel;
	}
	
	public void applyIsolationLevel(Connection conn) throws SQLException {
		if (isolationLevel != UNSET_ISOLATION_LEVEL) {
			originalIsolationLevel = conn.getTransactionIsolation();
			if (isolationLevel != originalIsolationLevel) {
				conn.setTransactionIsolation(isolationLevel);
			}
		}
	}
	
	public void restoreIsolationLevel(Connection conn) throws SQLException {
		if (isolationLevel != originalIsolationLevel) {
			conn.setTransactionIsolation(originalIsolationLevel);
		}
	}
}
