/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.transaction;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Date;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.Session;
import kr.kangwoo.damo.engine.transaction.user.UserProvidedTransaction;
import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.logging.Logger;
import kr.kangwoo.util.logging.LoggerFactory;
// setRollbackOnly
public class TransactionManager {
	
	private Logger connectionLogger = LoggerFactory.getLogger(Connection.class);
	
	private volatile int maxActive = Integer.MAX_VALUE;
	private int usedTransactions;
	private final ReentrantLock takeLock = new ReentrantLock();
	private final Condition notEmpty = takeLock.newCondition();
	private long nanosTimeout = 0;
	
	private TransactionProvider transactionProvider;
	
	public TransactionManager() {
	}
	
	public TransactionManager(TransactionProvider transactionProvider) {
		this.transactionProvider = transactionProvider;
	}
	
	public void setTransactionProvider(TransactionProvider transactionProvider) {
		this.transactionProvider = transactionProvider;
	}

	public int getMaxActive() {
		return maxActive;
	}

	public void setMaxActive(int maxActive) {
		this.maxActive = maxActive;
		if (this.maxActive < 1) {
			this.maxActive = Integer.MAX_VALUE;
		}
	}
	
	public long getTimeout() {
		return TimeUnit.MICROSECONDS.toMillis(nanosTimeout);
	}

	public void setTimeout(long timeout) {
		setTimeout(timeout, TimeUnit.MILLISECONDS);
	}
	
	public void setTimeout(long timeout, TimeUnit timeUnit) {
		nanosTimeout = timeUnit.toNanos(timeout);
	}
	
	/**
	 * <p>새로운 트랜잭션을 생성한다.</p>
	 * 
	 * @param transactionIsolationLevel 트랜잭션 격리 수준
	 * @return
	 * @throws TransactionException
	 */
	protected Transaction newTransaction(int transactionIsolationLevel) throws TransactionException {
		if (transactionProvider == null) {
			throw new TransactionException(Messages.getString("TransactionManager.TRANSACTION_PROVIDER_NOT_DEFINE")); //$NON-NLS-1$
		}

		try {
			takeLock.lockInterruptibly();
			try {
				long nanosTimeout = this.nanosTimeout;
				while (usedTransactions >= maxActive && nanosTimeout > 0) {
					nanosTimeout = notEmpty.awaitNanos(nanosTimeout);
				}
				
				if (usedTransactions < maxActive) {
					Transaction transaction = transactionProvider.newTransaction(transactionIsolationLevel);
					++usedTransactions;
					return transaction;
				} else {
					throw new IllegalStateException(MessageUtils.format(Messages.getString("TransactionManager.NOT_AVAILABLE_TRANSACTIONS"),  usedTransactions, maxActive));	 //$NON-NLS-1$
				}
			} finally {
				takeLock.unlock();
			}
		} catch(InterruptedException ex) {
			throw new IllegalStateException(Messages.getString("TransactionManager.TRANSACTION_INTERRUPTED"), ex); //$NON-NLS-1$
		}
	}
	
	protected void closeTransaction(Transaction transaction) throws TransactionException {
		try {
			takeLock.lockInterruptibly();
			try {
				--usedTransactions;
				notEmpty.signal();
			} finally {
				takeLock.unlock();
			}
		} catch(InterruptedException ex) {
			throw new IllegalStateException(Messages.getString("TransactionManager.TRANSACTION_INTERRUPTED"), ex);	 //$NON-NLS-1$
		}
	}
	
	/**
	 * <p>새로운 트랜잭션을 시작한다.</p>
	 * 
	 * @param transactionContext
	 * @throws SQLException
	 * @throws TransactionException
	 */
	public void begin(TransactionContext transactionContext) throws SQLException, TransactionException {
		begin(transactionContext, TransactionIsolationLevel.UNSET_ISOLATION_LEVEL);
	}
	
	/**
	 * <p>새로운 트랜잭션을 시작한다.</p>
	 * 
	 * @param transactionContext 트랜잭션 컨텍스트
	 * @param transactionIsolationLevel 트랜잭션 격리 수준
	 * @throws SQLException
	 * @throws TransactionException
	 */
	public void begin(TransactionContext transactionContext, int transactionIsolationLevel) throws SQLException, TransactionException {
		TransactionStatus status = transactionContext.getTransactionStatus();
		if (status == TransactionStatus.USER_PROVIDED) {
			throw new TransactionException(Messages.getString("TransactionManager.FAIL_BEGIN_TRANSACTION_USER")); //$NON-NLS-1$
		}
		if (status != TransactionStatus.NO_TRANSACTION) {
			throw new TransactionException(Messages.getString("TransactionManager.FAIL_BEGIN_TRANSACTION_NOT_ENDED")); //$NON-NLS-1$
		}
		
		Transaction transaction = newTransaction(transactionIsolationLevel);
		transactionContext.setTransaction(transaction);
		transactionContext.setTransactionStatus(TransactionStatus.ACTIVE);
	}
	
	public void commit(TransactionContext transactionContext) throws SQLException, TransactionException {
		TransactionStatus status = transactionContext.getTransactionStatus();
		if (status == TransactionStatus.USER_PROVIDED) {
			throw new TransactionException(Messages.getString("TransactionManager.FAIL_COMMIT_TRANSACTION_USER")); //$NON-NLS-1$
		}
		if (status != TransactionStatus.ACTIVE && status != TransactionStatus.COMMITTED && status != TransactionStatus.ROLLEDBACK) {
			throw new TransactionException(Messages.getString("TransactionManager.FAIL_COMMIT_TRANSACTION_NOT_BEGIN")); //$NON-NLS-1$
		}
		
		Transaction transaction = transactionContext.getTransaction();
		transaction.commit();
		transactionContext.setCommitRequired(false);
		transactionContext.setTransactionStatus(TransactionStatus.COMMITTED);
	}
	
	public void rollback(TransactionContext transactionContext) throws SQLException, TransactionException {
		TransactionStatus status = transactionContext.getTransactionStatus();
		if (status == TransactionStatus.USER_PROVIDED) {
			throw new TransactionException(Messages.getString("TransactionManager.FAIL_ROLLBACK_TRANSACTION_USER")); //$NON-NLS-1$
		}
		if (status != TransactionStatus.ACTIVE && status != TransactionStatus.COMMITTED && status != TransactionStatus.ROLLEDBACK) {
			throw new TransactionException(Messages.getString("TransactionManager.FAIL_ROLLBACK_TRANSACTION_NOT_BEGIN")); //$NON-NLS-1$
		}
		
		Transaction transaction = transactionContext.getTransaction();
		transaction.rollback();
		transactionContext.setCommitRequired(false);
		transactionContext.setTransactionStatus(TransactionStatus.ROLLEDBACK);
	}

	public void end(TransactionContext transactionContext) throws SQLException, TransactionException {
		TransactionStatus status = transactionContext.getTransactionStatus();
		if (status == TransactionStatus.USER_PROVIDED) {
			throw new TransactionException(Messages.getString("TransactionManager.FAIL_END_TRANSACTION_USER")); //$NON-NLS-1$
		}
		
		Transaction transaction = transactionContext.getTransaction();
		try {
			if (transaction != null) {
				try {
					// state != TransactionState.COMMITTED && state != TransactionState.ROLLBACKED
					if (transactionContext.isCommitRequired()) {
						transaction.rollback();
					}
				} finally {
					transactionContext.closeStatements();
					transaction.close();
				}
			}
		} finally {
			transactionContext.setTransaction(null);
			transactionContext.setTransactionStatus(TransactionStatus.NO_TRANSACTION);
			
			closeTransaction(transaction);
		}
	}
	
	/**
	 * <p>사용사 커넥션(Connection)을 사용한다.</p>
	 * <p>트랜잭션(Transaction)이 이미 사용중이면, 에러가 발생한다.</p>
	 * 
	 * @param transactionContext
	 * @param conn
	 * @throws SQLException
	 */
	public void setUserProvidedConnection(TransactionContext transactionContext, Connection conn) throws SQLException {
		TransactionStatus status = transactionContext.getTransactionStatus();
		if (status != TransactionStatus.USER_PROVIDED && status != TransactionStatus.NO_TRANSACTION) {
			throw new TransactionException(Messages.getString("TransactionManager.FAIL_CUSTORM_TRANSACTION_ALWAYS"));	 //$NON-NLS-1$
		}
		if (conn != null) {
			Transaction transaction = new UserProvidedTransaction(conn);
			transactionContext.setTransaction(transaction);	
			transactionContext.setTransactionStatus(TransactionStatus.USER_PROVIDED);
		} else {
			transactionContext.setTransaction(null);
			transactionContext.setTransactionStatus(TransactionStatus.NO_TRANSACTION);
		}
	}
}
