/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.transaction.jdbc;

import java.sql.Connection;
import java.sql.SQLException;

import javax.sql.DataSource;

import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.engine.transaction.BaseTransaction;
import kr.kangwoo.damo.engine.transaction.TransactionException;
import kr.kangwoo.damo.engine.transaction.TransactionIsolationLevel;
import kr.kangwoo.damo.engine.transaction.TransactionProvider;

/**
 * JDBC를 사용하여 트랜잭션을 구현한다.
 * JDBC 트랜잭션을  이용하면 여러 개의 SQL 문장을 하나의 트랜잭션으로 결합할 수 있다. 
 * JDBC 트랜잭션은 JTA 트랜잭션 보다 단순하다는 장점이 있다.
 * 하지만, 트랙잭션 범위가 하나의 데이터베이스 연결로 제한되어 있다는 단점이 있다. 
 * 즉, JDBC 트랜잭션은 다중 데이터베이스로 연결할 수 없다는 것이다. 다중 연결이 필요할 경우는 JTA 트랙잭션을 이용하기 바란다. 
 * 
 * JDBC 트랜잭션은 java.sql.Connection 인퍼페이스를 이용한다. 
 * JDBC 트랜잭션을 시작할 때, java.sql.Connection 인터페이스의 setAutoCommit(false)을 호출하고, 끝낼때에는  commit() 또는 rollback()을 호출한다.
 * 
 * @author badnom
 *
 */
public class JdbcTransaction extends BaseTransaction {

	private DataSource dataSource;
	private TransactionIsolationLevel isolationLevel;
	
	private Connection connection;
	  
	public JdbcTransaction(TransactionProvider transactionProvider, DataSource dataSource, int isolationLevel) throws TransactionException {
		super(transactionProvider);
		if (dataSource == null) {
			throw new TransactionException(Messages.getString("JdbcTransaction.INIT_ERROR")); //$NON-NLS-1$
		}
		this.dataSource = dataSource;
		this.isolationLevel = new TransactionIsolationLevel(isolationLevel);
	}
	
	public Connection getConnection() throws SQLException, TransactionException {
		if (connection == null) {
			connection = dataSource.getConnection();
			if (connection == null) {
				throw new TransactionException(Messages.getString("JdbcTransaction.CONNECTION_IS_NULL"));	 //$NON-NLS-1$
			}
			isolationLevel.applyIsolationLevel(connection);
			if (connection.getAutoCommit()) {
				connection.setAutoCommit(false);
			}
			
			// 로그 출력을 위한 랩핑
			connection = wrapConnection(connection);
		}
		return connection;
	}
	

	public void close() throws SQLException, TransactionException {
		if (connection != null) {
			try {
				isolationLevel.restoreIsolationLevel(connection);	
			} finally {
				connection.close();
				connection = null;
			}
		}
	}

	public void commit() throws SQLException, TransactionException {
		if (connection != null) {
			connection.commit();	
		}
	}

	public void rollback() throws SQLException, TransactionException {
		if (connection != null) {
			connection.rollback();	
		}
	}
}
