/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.engine.transaction.jta;

import java.sql.Connection;
import java.sql.SQLException;

import javax.sql.DataSource;
import javax.transaction.Status;
import javax.transaction.UserTransaction;

import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.engine.transaction.BaseTransaction;
import kr.kangwoo.damo.engine.transaction.TransactionException;
import kr.kangwoo.damo.engine.transaction.TransactionIsolationLevel;
import kr.kangwoo.damo.engine.transaction.TransactionProvider;

/**
 * JTA(Java Transaction API)를 사용하여 트랜잭션을 구현한다. 
 * JTA 트랜잭션은 JDBC 트랜잭션 보다 강력하다. 
 * JDBC 트랜잭션이 하나의 데이터베이스 연결로 제한되어 있다면, JTA 트랜잭션은 다중 참여자들을 가질 수 있다.
 * 
 * JTA 트랜잭션을 시작할 때, javax.transaction.UserTransaction 인터페이스의 begin()을 호출하고, 끝낼때에는  commit() 또는 rollback()을 호출한다.
 * 
 * @author badnom
 *
 */
public class JtaTransaction extends BaseTransaction {

	private UserTransaction userTransaction;
	private DataSource dataSource;
	private TransactionIsolationLevel isolationLevel;
	
	private boolean newTransaction = false;
	private Connection connection;
	  
	public JtaTransaction(TransactionProvider transactionProvider, UserTransaction userTransaction, DataSource dataSource, int isolationLevel) throws TransactionException {
		super(transactionProvider);
		
		if (userTransaction == null) {
			throw new TransactionException(Messages.getString("JtaTransaction.INIT_ERROR_USER_TRANSACTION_IS_NULL")); //$NON-NLS-1$
		}
		if (dataSource == null) {
			throw new TransactionException(Messages.getString("JtaTransaction.INIT_ERROR_DATA_SOURCE_IS_NULL")); //$NON-NLS-1$
		}
		this.dataSource = dataSource;
		this.isolationLevel = new TransactionIsolationLevel(isolationLevel);
	}
	
	public Connection getConnection() throws SQLException, TransactionException {
		if (connection == null) {
			try {
				newTransaction = (userTransaction.getStatus() == Status.STATUS_NO_TRANSACTION);
				if (newTransaction) {
					userTransaction.begin();
				}
			} catch (Exception e) {
				throw new TransactionException(Messages.getString("JtaTransaction.INIT_ERROR"), e); //$NON-NLS-1$
			} 
			connection = dataSource.getConnection();
			if (connection == null) {
				throw new TransactionException(Messages.getString("JtaTransaction.CONNECTION_IS_NULL"));	 //$NON-NLS-1$
			}
			isolationLevel.applyIsolationLevel(connection);
			if (connection.getAutoCommit()) {
				connection.setAutoCommit(false);
			}
			
			// 로그 출력을 위한 랩핑
			connection = wrapConnection(connection);
		}
		return connection;
	}
	

	public void close() throws SQLException, TransactionException {
		if (connection != null) {
			try {
				isolationLevel.restoreIsolationLevel(connection);	
			} finally {
				connection.close();
				connection = null;
			}
		}
	}

	public void commit() throws SQLException, TransactionException {
		if (connection != null) {
			if (newTransaction) {
				try {
					userTransaction.commit();
				} catch (Exception e) {
					throw new TransactionException(Messages.getString("JtaTransaction.COMMIT_ERROR"), e); //$NON-NLS-1$
				}
			}

		}
	}

	public void rollback() throws SQLException, TransactionException {
		if (connection != null) {
			try {
				if (newTransaction) {
					userTransaction.rollback();
				} else {
					userTransaction.setRollbackOnly();
				}
			} catch (Exception e) {
				throw new TransactionException(Messages.getString("JtaTransaction.ROLLBACK_ERROR"), e); //$NON-NLS-1$
			}

		}
	}
}
