/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.external.ehcache;

import java.net.URL;
import java.sql.SQLException;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;

import kr.kangwoo.damo.DamoException;
import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.Session;
import kr.kangwoo.damo.engine.cache.CacheKey;
import kr.kangwoo.damo.engine.cache.CacheProvider;
import kr.kangwoo.damo.engine.query.Query;
import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.StringUtils;
import kr.kangwoo.util.logging.Logger;
import kr.kangwoo.util.logging.LoggerFactory;
import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Element;
import net.sf.ehcache.bootstrap.BootstrapCacheLoader;
import net.sf.ehcache.config.DiskStoreConfiguration;
import net.sf.ehcache.event.RegisteredEventListeners;
import net.sf.ehcache.store.MemoryStoreEvictionPolicy;

public class EhCacheProvider implements CacheProvider {

	private Logger logger = LoggerFactory.getLogger(getClass());
	private static final ReentrantLock lock = new ReentrantLock();
	
	private CacheManager cacheManager;
	
	public EhCacheProvider() {
		URL configFileUrl = EhCacheProvider.class.getResource("/damo-cache.xml"); //$NON-NLS-1$
		cacheManager = new CacheManager(configFileUrl);
	}

	
	private Cache getCache(Query query) {
		String cacheName = query.getCacheName();
		if (StringUtils.isBlank(cacheName)) {
			throw new DamoException(MessageUtils.format(Messages.getString("EhCacheProvider.CACHE_NOT_DEFINE"), query.getDescription())); //$NON-NLS-1$
		}
		Cache cache = cacheManager.getCache(cacheName);
		if (cache == null) {
			lock.lock();
			try {
				cache = cacheManager.getCache(cacheName);
				if (cache == null) {
					cache = createCache(cacheName);
					cacheManager.addCache(cache);
					logger.debug(MessageUtils.format(Messages.getString("EhCacheProvider.DEFAULT_CREATE"), cacheName)); //$NON-NLS-1$
				}
			} finally {
				lock.unlock();
			}
			
		}
		return cache;
	}
	
	private Cache createCache(String name) {
		int maxElementsInMemory = 10000;
        MemoryStoreEvictionPolicy memoryStoreEvictionPolicy = MemoryStoreEvictionPolicy.LRU;
        boolean overflowToDisk = true;
        String diskStorePath = DiskStoreConfiguration.getDefaultPath();
        boolean eternal = false;
        long timeToLiveSeconds = 120;
        long timeToIdleSeconds = 120;
        boolean diskPersistent = false;
        long diskExpiryThreadIntervalSeconds = 120;
        RegisteredEventListeners registeredEventListeners = null;
        BootstrapCacheLoader bootstrapCacheLoader = null;
        int maxElementsOnDisk = 10000000;
        int diskSpoolBufferSizeMB = 0;
        boolean clearOnFlush = true;
        Cache cache = new Cache(name, maxElementsInMemory, memoryStoreEvictionPolicy, overflowToDisk
				, diskStorePath, eternal, timeToLiveSeconds, timeToIdleSeconds, diskPersistent, diskExpiryThreadIntervalSeconds
				, registeredEventListeners, bootstrapCacheLoader, maxElementsOnDisk, diskSpoolBufferSizeMB, clearOnFlush);
		return cache;
	}
	
	public void clearCache(Query query, Object parameterObject) {
		Cache cache = getCache(query);
		cache.removeAll();
	}

	@SuppressWarnings("unchecked")
	public <E> List<E> getCachedList(Session session, Query query, CacheKey cacheKey, Class<E> resultClass) throws SQLException {
		Cache cache = getCache(query);
		CacheKey newCacheKey = new CacheKey(cacheKey, resultClass);
		Element element = cache.get(newCacheKey);
		if (element == null) {
			List<E> list = session.getList(query, cacheKey.getParameterObject(), resultClass);
			element = new Element(newCacheKey, list);
			cache.put(element);
		}
		return (List<E>)element.getObjectValue();
	}
	


	@SuppressWarnings("unchecked")
	public <E> List<E> getCachedList(Session session, Query query, CacheKey cacheKey, Class<E> resultClass, int pageNo, int pageSize) throws SQLException {
		Cache cache = getCache(query);
		CacheKey newCacheKey = new CacheKey(cacheKey, resultClass, pageNo, pageSize);
		Element element = cache.get(newCacheKey);
		if (element == null) {
			List<E> list = session.getList(query, cacheKey.getParameterObject(), resultClass, pageNo, pageSize);
			element = new Element(newCacheKey, list);
			cache.put(element);
		}
		return (List<E>)element.getObjectValue();
	}



}
