/*
 * Copyright 2002-2009 Team Jaru.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package kr.kangwoo.damo.external.el;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.Map;

import kr.kangwoo.damo.DamoException;
import kr.kangwoo.damo.Messages;
import kr.kangwoo.damo.bundled.org.apache.commons.el.Constants;
import kr.kangwoo.damo.bundled.org.apache.commons.el.ELEvaluator;
import kr.kangwoo.damo.bundled.org.apache.commons.el.ELException;
import kr.kangwoo.damo.bundled.org.apache.commons.el.functions.Functions;
import kr.kangwoo.util.MessageUtils;

public class ExpressionEvaluator {

	private static final String DEFAULT_PREFIX = "fn"; //$NON-NLS-1$
	private static final ExpressionEvaluator expressionEvaluator = new ExpressionEvaluator();

	private ELEvaluator sEvaluator = new ELEvaluator(
			new SimpleVariableResolver());
	
	private Map<String, Method> functions = new HashMap<String, Method>();

	private ExpressionEvaluator() {
		Method[] methods = Functions.class.getDeclaredMethods();
		for (Method method : methods) {
			if (Modifier.isPublic(method.getModifiers())
					&& Modifier.isStatic(method.getModifiers())) {
				functions.put(DEFAULT_PREFIX + ":" + method.getName(), method); //$NON-NLS-1$
			}
		}
	}

	
	/**
	 * <p>해당 표현식이 문법에 맞는지 검사합니다.</p>
	 * 
	 * @param pAttributeName 표현식의 이름
	 * @param pAttributeValue 표현식
	 * @return
	 */
	public static String validate(String pAttributeName, String pAttributeValue) {
		try {
			expressionEvaluator.sEvaluator
					.parseExpressionString(pAttributeValue);
			return null;
		} catch (ELException exc) {
			return MessageFormat.format(Constants.ATTRIBUTE_PARSE_EXCEPTION,
					new Object[] { "" + pAttributeName, "" + pAttributeValue, //$NON-NLS-1$ //$NON-NLS-2$
							exc.getMessage() });
		}
	}


	/**
	 * <p>표현식을 분석 후 처리한 값을 가져온다.</p>
	 * 
	 * @param name 표현식 이름
	 * @param expression 표현식
	 * @param expectedType 반환할 타입
	 * @param context 값을 가져올 대상
	 * @return
	 * @throws DamoException
	 */
	public static Object evaluate(String name,
			String expression, Class<?> expectedType, Object context)
			throws DamoException {
		return evaluate(name, expression, expectedType, context, expressionEvaluator.functions,
				DEFAULT_PREFIX);
	}

	/**
	 * <p>표현식을 분석 후 처리한 값을 가져온다.</p>
	 * 
	 * @param name 표현식 이름
	 * @param expression 표현식
	 * @param expectedType 반환할 타입
	 * @param context 값을 가져올 대상
	 * @param functions 함수들
	 * @param defaultPrefix 접두어
	 * @return
	 * @throws DamoException
	 */
	public static Object evaluate(String name, String expression,
			Class<?> expectedType, Object context,
			Map<String, Method> functions, String defaultPrefix)
			throws DamoException {
		try {
			return expressionEvaluator.sEvaluator.evaluate(expression, context,
					expectedType, functions, defaultPrefix);
		} catch (ELException ex) {
			throw new DamoException(MessageUtils.format(
					Constants.ATTRIBUTE_EVALUATION_EXCEPTION, "" //$NON-NLS-1$
							+ name, "" + expression, ex //$NON-NLS-1$
							.getMessage(), ex.getRootCause()), ex
					.getRootCause());
		}
	}

	/**
	 * <p>표현식을 분석한다.</p>
	 * 
	 * @param expression 표현식
	 * @return
	 */
	public static Object parseExpressionString(String expression) {
		try {
			return expressionEvaluator.sEvaluator.parseExpressionString(expression);
		} catch (ELException ex) {
			throw new DamoException(MessageUtils.format(Messages.getString("ExpressionEvaluator.PARSE_ERROR"), expression), ex); //$NON-NLS-1$
		}
	}
}
