package kr.kangwoo.damo.support.logging.simple;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.sql.CallableStatement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import kr.kangwoo.util.BeanUtils;
import kr.kangwoo.util.StopWatch;
import kr.kangwoo.util.SystemUtils;
import kr.kangwoo.util.logging.Logger;
import kr.kangwoo.util.logging.LoggerFactory;

public class PreparedStatementLogProxy implements InvocationHandler {

	private static final Set<String> EXECUTE_METHOD_NAMES = new HashSet<String>();
	private static final Set<String> SET_METHOD_NAMES = new HashSet<String>();
	
	static {
		EXECUTE_METHOD_NAMES.add("execute");
		EXECUTE_METHOD_NAMES.add("executeUpdate");
		EXECUTE_METHOD_NAMES.add("executeQuery");
		
		SET_METHOD_NAMES.add("setNull");
		SET_METHOD_NAMES.add("setBoolean");
		SET_METHOD_NAMES.add("setByte");
		SET_METHOD_NAMES.add("setShort");
		SET_METHOD_NAMES.add("setInt");
		SET_METHOD_NAMES.add("setLong");
		SET_METHOD_NAMES.add("setFloat");
		SET_METHOD_NAMES.add("setDouble");
		SET_METHOD_NAMES.add("setBigDecimal");
		SET_METHOD_NAMES.add("setString");
		SET_METHOD_NAMES.add("setBytes");
		SET_METHOD_NAMES.add("setDate");
		SET_METHOD_NAMES.add("setTime");
		SET_METHOD_NAMES.add("setTimestamp");
		SET_METHOD_NAMES.add("setAsciiStream");
		SET_METHOD_NAMES.add("setBinaryStream");
		SET_METHOD_NAMES.add("setObject");
		// JDBC 2.0
		SET_METHOD_NAMES.add("setCharacterStream");
		SET_METHOD_NAMES.add("setRef");
		SET_METHOD_NAMES.add("setBlob");
		SET_METHOD_NAMES.add("setClob");
		// JDBC 3.0
		SET_METHOD_NAMES.add("setURL");
	}
	
	private Logger logger = LoggerFactory.getLogger(PreparedStatement.class);
	private Logger rsLogger = LoggerFactory.getLogger(ResultSet.class);
	

	private PreparedStatement pstmt;
	private String sql;
	
	private Map<Object, Object> columnMap = new LinkedHashMap<Object, Object>();
	private int batchTotalCount;
	private int batchCount;
	
	private PreparedStatementLogProxy(PreparedStatement pstmt, String sql) {
		this.pstmt = pstmt;
		this.sql = sql;
	}
	
	public Object invoke(Object proxy, Method method, Object[] args)
			throws Throwable {
		try {
			if (EXECUTE_METHOD_NAMES.contains(method.getName())) {
				StopWatch stopWatch = null;
				if (logger.isDebugEnabled()) {
					StringBuilder msg = new StringBuilder();
//					msg.append(" Executing ").append(pstmt).append(SystemUtils.LINE_SEPARATOR);
//					msg.append(sql);
//					int keyLen = String.valueOf(columnMap.size()).length();
//					for (Object key : columnMap.keySet()) {
//						msg.append(SystemUtils.LINE_SEPARATOR);
//						msg.append("[").append(StringUtils.leftPad(key.toString(), keyLen, "0")).append("] ").append(columnMap.get(key));
//					}

					// '?'를 변수값으로 변경하기.
					int paramIndex = 1;
					int beginIndex = 0;
					int index = 0;
					while ( (index = sql.indexOf('?', beginIndex)) != -1) {
						msg.append(sql.substring(beginIndex, index));
						Object paramValue = columnMap.get(paramIndex);
						if (paramValue instanceof String) {
							msg.append("'");
							msg.append(paramValue);
							msg.append("'");
						} else {
							msg.append(paramValue);
						}
						beginIndex = index + 1;
						paramIndex++;
					}
					msg.append(sql.substring(beginIndex));
					
					logger.debug(method.getName() + " " + pstmt + SystemUtils.LINE_SEPARATOR + msg.toString());
					columnMap.clear();
					stopWatch = new StopWatch();
				}
				
				if ("executeQuery".equals(method.getName())) {
					ResultSet rs = (ResultSet) method.invoke(pstmt, args);
					if (logger.isDebugEnabled() && stopWatch != null) {
						stopWatch.stop();
						logger.debug(method.getName() + " " + pstmt + " Elapsed Time : " + stopWatch.getTotalElapsedTime() + "ms");
					}
					if (rs != null) {
						return rsLogger.isDebugEnabled() ? ResultSetLogProxy.newInstance(rs) : rs;
					} else {
						return null;
					}
				} else {
					Object obj = method.invoke(pstmt, args);
					if (logger.isDebugEnabled() && stopWatch != null) {
						stopWatch.stop();
						logger.debug(method.getName() + " " + pstmt + " Elapsed Time : " + stopWatch.getTotalElapsedTime() + "ms");
					}
					return obj;
				}
			} else if ("executeBatch".contains(method.getName())) {
				logger.debug(" Executing Batch " + pstmt + " : COUNT = " + batchCount + ", TOTAL = " + batchTotalCount);
				batchCount = 0;
			} else if ("addBatch".contains(method.getName())) {
				batchCount++;
				batchTotalCount++;
			} else if (SET_METHOD_NAMES.contains(method.getName())) {
				if ("setNull".equals(method.getName())) {
					columnMap.put(args[0], null);
				} else {
					columnMap.put(args[0], args[1]);
				}
			} else if ("getResultSet".equals(method.getName())) {
				ResultSet rs = (ResultSet) method.invoke(pstmt, args);
				if (rs != null) {
					return rsLogger.isDebugEnabled() ? ResultSetLogProxy.newInstance(rs) : rs;
				} else {
					return null;
				}
			} else if ("equals".equals(method.getName())) {
				Object obj = args[0];
				if (obj instanceof Proxy) {
					return new Boolean(proxy == obj);
				}
				return Boolean.FALSE;
			} else if ("hashCode".equals(method.getName())) {
				return new Integer(proxy.hashCode());
			} else if ("close".equals(method.getName())) {
				
			}
			return method.invoke(pstmt, args);
		} catch(Throwable t) {
			throw BeanUtils.unwrapThrowable(t);
		}
	}
	

	
	public PreparedStatement getPreparedStatement() {
		return pstmt;
	}
	
	public static PreparedStatement newInstance(PreparedStatement pstmt, String sql) {
		InvocationHandler handler = new PreparedStatementLogProxy(pstmt, sql);
		return (PreparedStatement)Proxy.newProxyInstance(PreparedStatement.class.getClassLoader(), new Class[]{PreparedStatement.class, CallableStatement.class}, handler);
	}
}
