package kr.kangwoo.damo.util.tool;

import java.util.HashSet;
import java.util.Set;

import kr.kangwoo.util.MessageUtils;
import kr.kangwoo.util.StringUtils;
import kr.kangwoo.util.sql.ColumnDefinition;
import kr.kangwoo.util.sql.JavaType;
import kr.kangwoo.util.sql.TableDefinition;

public class ModelCodeGenerator extends CodeGenerator {

	private boolean serializableFlag = true;
	private boolean toStringFlag = true;
	private boolean trimFlag = false; // setMethod 만들때 값을 trim할 여부
	private boolean finalFlag = false; // setMethod 만들때 값을 final할 여부

	private String getterCommnet = "{0}을;를 가져온다.";
	private String setterCommnet = "{0}을;를 설정한다.";
	
	public String generateModelCode(TableDefinition tableDefinition, String packageName, String modelName) {
		StringBuilder result = new StringBuilder();
		
		if (StringUtils.isNotBlank(packageName)) {
			result.append("package ").append(packageName).append(";");
			newLine(result);
			newLine(result);
		}
		
		
		Set<String> importSet = new HashSet<String>();
		
		if (serializableFlag) {
			importSet.add("java.io.Serializable");
		}
		
		importSet.add("kr.kangwoo.damo.annotation.PrimaryKey");
		importSet.add("kr.kangwoo.damo.annotation.Table");
		
		if (toStringFlag) {
			importSet.add("kr.kangwoo.util.ObjectUtils");
		}
		
		for (int i = 0; i < tableDefinition.size(); i++) {
			ColumnDefinition columnDefinition = tableDefinition.getColumn(i);
			JavaType javaType = columnDefinition.getJavaType();
			if (!"java.lang".equals(javaType.getPackageName())) {
				importSet.add(javaType.getFullyQualifiedName());
			}
		}
		// java.io.Serializable
		
		if (importSet.size() > 0) {
			for (String importElement : importSet) {
				result.append("import ").append(importElement).append(";");
				newLine(result);
			}
			newLine(result);
			newLine(result);		
		}

		result.append("@Table(tableName=\"").append(tableDefinition.getTableName().getTable()).append("\")");
		newLine(result);
		result.append("public class ").append(modelName);
		if (serializableFlag) {
			result.append(" implements Serializable");	
		}

		result.append(" {");
		newLine(result);
		
		// Field 생성
		newLine(result);
		for (int i = 0; i < tableDefinition.size(); i++) {
			ColumnDefinition columnDefinition = tableDefinition.getColumn(i);
			generateField(result, columnDefinition);
		}
		newLine(result);

		// Get/Set Method 생성
		newLine(result);
		for (int i = 0; i < tableDefinition.size(); i++) {
			ColumnDefinition columnDefinition = tableDefinition.getColumn(i);
			generateGetMethod(result, columnDefinition);
			generateSetMethod(result, columnDefinition);
		}
		
		// toString 생성
		if (toStringFlag) {
			generateToStringMethod(result);
		}
		result.append("}");
		
		
		return result.toString();
	}

	protected void generateField(StringBuilder sb, ColumnDefinition columnDefinition) {
		if (columnDefinition.isPrimaryKey()) {
			javaIndent(sb, 1);
			sb.append("@PrimaryKey");
			newLine(sb);
		}
		javaIndent(sb, 1);
		sb.append("private ");
		JavaType javaType = columnDefinition.getJavaType();
		sb.append(javaType.getShortName());
		sb.append(" ");
		sb.append(StringUtils.camelString(columnDefinition.getColumnName(), false));
		sb.append(";");
		newLine(sb);
	}
	
	protected void generateGetMethod(StringBuilder sb, ColumnDefinition columnDefinition) {
		// Comment
		if (StringUtils.isNotBlank(columnDefinition.getColumnComment())) {
			javaIndent(sb, 1);
			sb.append("/**");
			newLine(sb);
			
			javaIndent(sb, 1);
			sb.append(" * <p>").append(MessageUtils.format(getterCommnet, columnDefinition.getColumnComment())).append("</p>");
			newLine(sb);
			
			javaIndent(sb, 1);
			sb.append(" * ");
			newLine(sb);
			
			javaIndent(sb, 1);
			sb.append(" * @return ").append(columnDefinition.getColumnComment());
			newLine(sb);
			
			javaIndent(sb, 1);
			sb.append(" */");
			newLine(sb);
		}
		
		// Method
		javaIndent(sb, 1);
		sb.append("public ");
		JavaType javaType = columnDefinition.getJavaType();
		sb.append(javaType.getShortName());
		sb.append(" get");
		sb.append(StringUtils.camelString(columnDefinition.getColumnName(), true));
		sb.append("() {");
		newLine(sb);
		
		javaIndent(sb, 2);
		sb.append("return this.").append(StringUtils.camelString(columnDefinition.getColumnName(), false));
		sb.append(";");
		newLine(sb);
		
		javaIndent(sb, 1);
		sb.append("}");
		newLine(sb);
		
		newLine(sb);
	}
	
	protected void generateSetMethod(StringBuilder sb, ColumnDefinition columnDefinition) {
		// Comment
		if (StringUtils.isNotBlank(columnDefinition.getColumnComment())) {
			javaIndent(sb, 1);
			sb.append("/**");
			newLine(sb);
			
			javaIndent(sb, 1);
			sb.append(" * <p>").append(MessageUtils.format(setterCommnet, columnDefinition.getColumnComment())).append("</p>");
			newLine(sb);
			
			javaIndent(sb, 1);
			sb.append(" * ");
			newLine(sb);
			
			javaIndent(sb, 1);
			sb.append(" * @param ").append(StringUtils.camelString(columnDefinition.getColumnName(), false));
			sb.append(" ").append(columnDefinition.getColumnComment());
			newLine(sb);
			
			javaIndent(sb, 1);
			sb.append(" */");
			newLine(sb);
		}
		
		// Method
		javaIndent(sb, 1);
		sb.append("public ");
		sb.append("void set");
		sb.append(StringUtils.camelString(columnDefinition.getColumnName(), true));
		sb.append("(");
		if (finalFlag) {
			sb.append("final ");	
		}
		JavaType javaType = columnDefinition.getJavaType();
		sb.append(javaType.getShortName());
		sb.append(" ");
		sb.append(StringUtils.camelString(columnDefinition.getColumnName(), false));
		sb.append(") {");
		newLine(sb);
		
		javaIndent(sb, 2);
		sb.append("this.").append(StringUtils.camelString(columnDefinition.getColumnName(), false));
		sb.append(" = ");
		if (trimFlag && "String".equals(javaType.getShortName())) {
			sb.append("((");
			sb.append(StringUtils.camelString(columnDefinition.getColumnName(), false));
			sb.append(" != null) ? ").append(StringUtils.camelString(columnDefinition.getColumnName(), false)).append(".trim()");
			sb.append(" : ").append(StringUtils.camelString(columnDefinition.getColumnName(), false));
			sb.append(")");
		} else {
			sb.append(StringUtils.camelString(columnDefinition.getColumnName(), false));	
		}
		
		sb.append(";");
		newLine(sb);
		
		javaIndent(sb, 1);
		sb.append("}");
		newLine(sb);
		
		newLine(sb);
	}

	protected void generateToStringMethod(StringBuilder sb) {
		// Method
		javaIndent(sb, 1);
		sb.append("public String toString() {");
		newLine(sb);
		
		javaIndent(sb, 2);
		sb.append("return ObjectUtils.reflectionToString(this);");
		newLine(sb);
		
		javaIndent(sb, 1);
		sb.append("}");
		newLine(sb);

		newLine(sb);
	}

}
